@@ -1,8 +1,10 @@
 /*
- * Copyright (c) 1999-2018 Douglas Gilbert.
+ * Copyright (c) 1999-2021 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 /*
@@ -19,6 +21,7 @@
 #include <stdarg.h>
 #include <string.h>
 #include <unistd.h>
+#include <errno.h>
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -28,6 +31,7 @@
 #include "sg_cmds_basic.h"
 #include "sg_pt.h"
 #include "sg_unaligned.h"
+#include "sg_pr2serr.h"
 
 
 
@@ -69,26 +73,6 @@
 #define INQUIRY_RESP_INITIAL_LEN 36
 
 
-#if defined(__GNUC__) || defined(__clang__)
-static int pr2ws(const char * fmt, ...)
-        __attribute__ ((format (printf, 1, 2)));
-#else
-static int pr2ws(const char * fmt, ...);
-#endif
-
-
-static int
-pr2ws(const char * fmt, ...)
-{
-    va_list args;
-    int n;
-
-    va_start(args, fmt);
-    n = vfprintf(sg_warnings_strm ? sg_warnings_strm : stderr, fmt, args);
-    va_end(args);
-    return n;
-}
-
 static struct sg_pt_base *
 create_pt_obj(const char * cname)
 {
@@ -105,8 +89,8 @@ sg_ll_sync_cache_10(int sg_fd, bool sync_nv, bool immed, int group,
                     unsigned int lba, unsigned int count, bool noisy,
                     int verbose)
 {
-    static const char * const cdb_name_s = "synchronize cache(10)";
-    int res, ret, k, sense_cat;
+    static const char * const cdb_s = "synchronize cache(10)";
+    int res, ret, sense_cat;
     uint8_t sc_cdb[SYNCHRONIZE_CACHE_CMDLEN] =
                 {SYNCHRONIZE_CACHE_CMD, 0, 0, 0, 0, 0, 0, 0, 0, 0};
     uint8_t sense_b[SENSE_BUFF_LEN];
@@ -125,21 +109,24 @@ sg_ll_sync_cache_10(int sg_fd, bool sync_nv, bool immed, int group,
     sg_put_unaligned_be16((int16_t)count, sc_cdb + 7);
 
     if (verbose) {
-        pr2ws("    %s cdb: ", cdb_name_s);
-        for (k = 0; k < SYNCHRONIZE_CACHE_CMDLEN; ++k)
-            pr2ws("%02x ", sc_cdb[k]);
-        pr2ws("\n");
+        char b[128];
+
+        pr2ws("    %s cdb: %s\n", cdb_s,
+              sg_get_command_str(sc_cdb, SYNCHRONIZE_CACHE_CMDLEN, false,
+                                 sizeof(b), b));
     }
-    if (NULL == ((ptvp = create_pt_obj(cdb_name_s))))
+    if (NULL == ((ptvp = create_pt_obj(cdb_s))))
         return -1;
     set_scsi_pt_cdb(ptvp, sc_cdb, sizeof(sc_cdb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, cdb_name_s, res, SG_NO_DATA_IN, sense_b,
-                               noisy, verbose, &sense_cat);
-    if (-1 == ret)
-        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
-    else if (-2 == ret) {
+    ret = sg_cmds_process_resp(ptvp, cdb_s, res, noisy, verbose, &sense_cat);
+    if (-1 == ret) {
+        if (get_scsi_pt_transport_err(ptvp))
+            ret = SG_LIB_TRANSPORT_ERROR;
+        else
+            ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
+    } else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
         case SG_LIB_CAT_NO_SENSE:
@@ -162,8 +149,8 @@ int
 sg_ll_readcap_16(int sg_fd, bool pmi, uint64_t llba, void * resp,
                  int mx_resp_len, bool noisy, int verbose)
 {
-    static const char * const cdb_name_s = "read capacity(16)";
-    int k, ret, res, sense_cat;
+    static const char * const cdb_s = "read capacity(16)";
+    int ret, res, sense_cat;
     uint8_t rc_cdb[SERVICE_ACTION_IN_16_CMDLEN] =
                         {SERVICE_ACTION_IN_16_CMD, READ_CAPACITY_16_SA,
                          0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
@@ -177,22 +164,25 @@ sg_ll_readcap_16(int sg_fd, bool pmi, uint64_t llba, void * resp,
     /* Allocation length, no guidance in SBC-2 rev 15b */
     sg_put_unaligned_be32((uint32_t)mx_resp_len, rc_cdb + 10);
     if (verbose) {
-        pr2ws("    %s cdb: ", cdb_name_s);
-        for (k = 0; k < SERVICE_ACTION_IN_16_CMDLEN; ++k)
-            pr2ws("%02x ", rc_cdb[k]);
-        pr2ws("\n");
+        char b[128];
+
+        pr2ws("    %s cdb: %s\n", cdb_s,
+              sg_get_command_str(rc_cdb, SERVICE_ACTION_IN_16_CMDLEN, false,
+                                 sizeof(b), b));
     }
-    if (NULL == ((ptvp = create_pt_obj(cdb_name_s))))
+    if (NULL == ((ptvp = create_pt_obj(cdb_s))))
         return -1;
     set_scsi_pt_cdb(ptvp, rc_cdb, sizeof(rc_cdb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     set_scsi_pt_data_in(ptvp, (uint8_t *)resp, mx_resp_len);
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, cdb_name_s, res, mx_resp_len, sense_b,
-                               noisy, verbose, &sense_cat);
-    if (-1 == ret)
-        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
-    else if (-2 == ret) {
+    ret = sg_cmds_process_resp(ptvp, cdb_s, res, noisy, verbose, &sense_cat);
+    if (-1 == ret) {
+        if (get_scsi_pt_transport_err(ptvp))
+            ret = SG_LIB_TRANSPORT_ERROR;
+        else
+            ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
+    } else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
         case SG_LIB_CAT_NO_SENSE:
@@ -215,8 +205,8 @@ int
 sg_ll_readcap_10(int sg_fd, bool pmi, unsigned int lba, void * resp,
                  int mx_resp_len, bool noisy, int verbose)
 {
-    static const char * const cdb_name_s = "read capacity(10)";
-    int k, ret, res, sense_cat;
+    static const char * const cdb_s = "read capacity(10)";
+    int ret, res, sense_cat;
     uint8_t rc_cdb[READ_CAPACITY_10_CMDLEN] =
                          {READ_CAPACITY_10_CMD, 0, 0, 0, 0, 0, 0, 0, 0, 0};
     uint8_t sense_b[SENSE_BUFF_LEN];
@@ -227,22 +217,25 @@ sg_ll_readcap_10(int sg_fd, bool pmi, unsigned int lba, void * resp,
         sg_put_unaligned_be32((uint32_t)lba, rc_cdb + 2);
     }
     if (verbose) {
-        pr2ws("    %s cdb: ", cdb_name_s);
-        for (k = 0; k < READ_CAPACITY_10_CMDLEN; ++k)
-            pr2ws("%02x ", rc_cdb[k]);
-        pr2ws("\n");
+        char b[128];
+
+        pr2ws("    %s cdb: %s\n", cdb_s,
+              sg_get_command_str(rc_cdb, READ_CAPACITY_10_CMDLEN, false,
+                                 sizeof(b), b));
     }
-    if (NULL == ((ptvp = create_pt_obj(cdb_name_s))))
+    if (NULL == ((ptvp = create_pt_obj(cdb_s))))
         return -1;
     set_scsi_pt_cdb(ptvp, rc_cdb, sizeof(rc_cdb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     set_scsi_pt_data_in(ptvp, (uint8_t *)resp, mx_resp_len);
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, cdb_name_s, res, mx_resp_len, sense_b,
-                               noisy, verbose, &sense_cat);
-    if (-1 == ret)
-        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
-    else if (-2 == ret) {
+    ret = sg_cmds_process_resp(ptvp, cdb_s, res, noisy, verbose, &sense_cat);
+    if (-1 == ret) {
+        if (get_scsi_pt_transport_err(ptvp))
+            ret = SG_LIB_TRANSPORT_ERROR;
+        else
+            ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
+    } else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
         case SG_LIB_CAT_NO_SENSE:
@@ -265,8 +258,8 @@ int
 sg_ll_mode_sense6(int sg_fd, bool dbd, int pc, int pg_code, int sub_pg_code,
                   void * resp, int mx_resp_len, bool noisy, int verbose)
 {
-    static const char * const cdb_name_s = "mode sense(6)";
-    int res, ret, k, sense_cat, resid;
+    static const char * const cdb_s = "mode sense(6)";
+    int res, ret, sense_cat, resid;
     uint8_t modes_cdb[MODE_SENSE6_CMDLEN] =
         {MODE_SENSE6_CMD, 0, 0, 0, 0, 0};
     uint8_t sense_b[SENSE_BUFF_LEN];
@@ -281,23 +274,26 @@ sg_ll_mode_sense6(int sg_fd, bool dbd, int pc, int pg_code, int sub_pg_code,
         return -1;
     }
     if (verbose) {
-        pr2ws("    %s cdb: ", cdb_name_s);
-        for (k = 0; k < MODE_SENSE6_CMDLEN; ++k)
-            pr2ws("%02x ", modes_cdb[k]);
-        pr2ws("\n");
+        char b[128];
+
+        pr2ws("    %s cdb: %s\n", cdb_s,
+              sg_get_command_str(modes_cdb, MODE_SENSE6_CMDLEN, false,
+                                 sizeof(b), b));
     }
-    if (NULL == ((ptvp = create_pt_obj(cdb_name_s))))
+    if (NULL == ((ptvp = create_pt_obj(cdb_s))))
         return -1;
     set_scsi_pt_cdb(ptvp, modes_cdb, sizeof(modes_cdb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     set_scsi_pt_data_in(ptvp, (uint8_t *)resp, mx_resp_len);
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, cdb_name_s, res, mx_resp_len, sense_b,
-                               noisy, verbose, &sense_cat);
+    ret = sg_cmds_process_resp(ptvp, cdb_s, res, noisy, verbose, &sense_cat);
     resid = get_scsi_pt_resid(ptvp);
-    if (-1 == ret)
-        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
-    else if (-2 == ret) {
+    if (-1 == ret) {
+        if (get_scsi_pt_transport_err(ptvp))
+            ret = SG_LIB_TRANSPORT_ERROR;
+        else
+            ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
+    } else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
         case SG_LIB_CAT_NO_SENSE:
@@ -309,7 +305,7 @@ sg_ll_mode_sense6(int sg_fd, bool dbd, int pc, int pg_code, int sub_pg_code,
         }
     } else {
         if ((verbose > 2) && (ret > 0)) {
-            pr2ws("    %s: response", cdb_name_s);
+            pr2ws("    %s: response", cdb_s);
             if (3 == verbose) {
                 pr2ws("%s:\n", (ret > 256 ? ", first 256 bytes" : ""));
                 hex2stderr((const uint8_t *)resp, (ret > 256 ? 256 : ret),
@@ -326,7 +322,7 @@ sg_ll_mode_sense6(int sg_fd, bool dbd, int pc, int pg_code, int sub_pg_code,
     if (resid > 0) {
         if (resid > mx_resp_len) {
             pr2ws("%s: resid (%d) should never exceed requested len=%d\n",
-                  cdb_name_s, resid, mx_resp_len);
+                  cdb_s, resid, mx_resp_len);
             return ret ? ret : SG_LIB_CAT_MALFORMED;
         }
         /* zero unfilled section of response buffer */
@@ -360,8 +356,8 @@ sg_ll_mode_sense10_v2(int sg_fd, bool llbaa, bool dbd, int pc, int pg_code,
                       int sub_pg_code, void * resp, int mx_resp_len,
                       int timeout_secs, int * residp, bool noisy, int verbose)
 {
-    int res, ret, k, sense_cat, resid;
-    static const char * const cdb_name_s = "mode sense(10)";
+    int res, ret, sense_cat, resid;
+    static const char * const cdb_s = "mode sense(10)";
     struct sg_pt_base * ptvp;
     uint8_t modes_cdb[MODE_SENSE10_CMDLEN] =
         {MODE_SENSE10_CMD, 0, 0, 0, 0, 0, 0, 0, 0, 0};
@@ -376,28 +372,31 @@ sg_ll_mode_sense10_v2(int sg_fd, bool llbaa, bool dbd, int pc, int pg_code,
         goto gen_err;
     }
     if (verbose) {
-        pr2ws("    %s cdb: ", cdb_name_s);
-        for (k = 0; k < MODE_SENSE10_CMDLEN; ++k)
-            pr2ws("%02x ", modes_cdb[k]);
-        pr2ws("\n");
+        char b[128];
+
+        pr2ws("    %s cdb: %s\n", cdb_s,
+              sg_get_command_str(modes_cdb, MODE_SENSE10_CMDLEN, false,
+                                 sizeof(b), b));
     }
     if (timeout_secs <= 0)
         timeout_secs = DEF_PT_TIMEOUT;
 
-    if (NULL == ((ptvp = create_pt_obj(cdb_name_s))))
+    if (NULL == ((ptvp = create_pt_obj(cdb_s))))
         goto gen_err;
     set_scsi_pt_cdb(ptvp, modes_cdb, sizeof(modes_cdb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     set_scsi_pt_data_in(ptvp, (uint8_t *)resp, mx_resp_len);
     res = do_scsi_pt(ptvp, sg_fd, timeout_secs, verbose);
-    ret = sg_cmds_process_resp(ptvp, cdb_name_s, res, mx_resp_len, sense_b,
-                               noisy, verbose, &sense_cat);
+    ret = sg_cmds_process_resp(ptvp, cdb_s, res, noisy, verbose, &sense_cat);
     resid = get_scsi_pt_resid(ptvp);
     if (residp)
         *residp = resid;
-    if (-1 == ret)
-        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
-    else if (-2 == ret) {
+    if (-1 == ret) {
+        if (get_scsi_pt_transport_err(ptvp))
+            ret = SG_LIB_TRANSPORT_ERROR;
+        else
+            ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
+    } else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
         case SG_LIB_CAT_NO_SENSE:
@@ -409,7 +408,7 @@ sg_ll_mode_sense10_v2(int sg_fd, bool llbaa, bool dbd, int pc, int pg_code,
         }
     } else {
         if ((verbose > 2) && (ret > 0)) {
-            pr2ws("    %s: response", cdb_name_s);
+            pr2ws("    %s: response", cdb_s);
             if (3 == verbose) {
                 pr2ws("%s:\n", (ret > 256 ? ", first 256 bytes" : ""));
                 hex2stderr((const uint8_t *)resp, (ret > 256 ? 256 : ret),
@@ -426,7 +425,7 @@ sg_ll_mode_sense10_v2(int sg_fd, bool llbaa, bool dbd, int pc, int pg_code,
     if (resid > 0) {
         if (resid > mx_resp_len) {
             pr2ws("%s: resid (%d) should never exceed requested len=%d\n",
-                  cdb_name_s, resid, mx_resp_len);
+                  cdb_s, resid, mx_resp_len);
             return ret ? ret : SG_LIB_CAT_MALFORMED;
         }
         /* zero unfilled section of response buffer */
@@ -442,44 +441,49 @@ gen_err:
 /* Invokes a SCSI MODE SELECT (6) command.  Return of 0 -> success,
  * various SG_LIB_CAT_* positive values or -1 -> other errors */
 int
-sg_ll_mode_select6(int sg_fd, bool pf, bool sp, void * paramp, int param_len,
-                   bool noisy, int verbose)
+sg_ll_mode_select6_v2(int sg_fd, bool pf, bool rtd, bool sp, void * paramp,
+                      int param_len, bool noisy, int verbose)
 {
-    static const char * const cdb_name_s = "mode select(6)";
-    int res, ret, k, sense_cat;
+    static const char * const cdb_s = "mode select(6)";
+    int res, ret, sense_cat;
     uint8_t modes_cdb[MODE_SELECT6_CMDLEN] =
         {MODE_SELECT6_CMD, 0, 0, 0, 0, 0};
     uint8_t sense_b[SENSE_BUFF_LEN];
     struct sg_pt_base * ptvp;
 
     modes_cdb[1] = (uint8_t)((pf ? 0x10 : 0x0) | (sp ? 0x1 : 0x0));
+    if (rtd)
+        modes_cdb[1] |= 0x2;
     modes_cdb[4] = (uint8_t)(param_len & 0xff);
     if (param_len > 0xff) {
-        pr2ws("%s: param_len too big\n", cdb_name_s);
+        pr2ws("%s: param_len too big\n", cdb_s);
         return -1;
     }
     if (verbose) {
-        pr2ws("    %s cdb: ", cdb_name_s);
-        for (k = 0; k < MODE_SELECT6_CMDLEN; ++k)
-            pr2ws("%02x ", modes_cdb[k]);
-        pr2ws("\n");
+        char b[128];
+
+        pr2ws("    %s cdb: %s\n", cdb_s,
+              sg_get_command_str(modes_cdb, MODE_SELECT6_CMDLEN, false,
+                                 sizeof(b), b));
     }
     if (verbose > 1) {
-        pr2ws("    %s parameter list\n", cdb_name_s);
+        pr2ws("    %s parameter list\n", cdb_s);
         hex2stderr((const uint8_t *)paramp, param_len, -1);
     }
 
-    if (NULL == ((ptvp = create_pt_obj(cdb_name_s))))
+    if (NULL == ((ptvp = create_pt_obj(cdb_s))))
         return -1;
     set_scsi_pt_cdb(ptvp, modes_cdb, sizeof(modes_cdb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     set_scsi_pt_data_out(ptvp, (uint8_t *)paramp, param_len);
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, cdb_name_s, res, SG_NO_DATA_IN, sense_b,
-                               noisy, verbose, &sense_cat);
-    if (-1 == ret)
-        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
-    else if (-2 == ret) {
+    ret = sg_cmds_process_resp(ptvp, cdb_s, res, noisy, verbose, &sense_cat);
+    if (-1 == ret) {
+        if (get_scsi_pt_transport_err(ptvp))
+            ret = SG_LIB_TRANSPORT_ERROR;
+        else
+            ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
+    } else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
         case SG_LIB_CAT_NO_SENSE:
@@ -496,47 +500,61 @@ sg_ll_mode_select6(int sg_fd, bool pf, bool sp, void * paramp, int param_len,
     return ret;
 }
 
-/* Invokes a SCSI MODE SELECT (10) command.  Return of 0 -> success,
- * various SG_LIB_CAT_* positive values or -1 -> other errors */
 int
-sg_ll_mode_select10(int sg_fd, bool pf, bool sp, void * paramp, int param_len,
-                    bool noisy, int verbose)
+sg_ll_mode_select6(int sg_fd, bool pf, bool sp, void * paramp, int param_len,
+                   bool noisy, int verbose)
 {
-    static const char * const cdb_name_s = "mode select(10)";
-    int res, ret, k, sense_cat;
+    return sg_ll_mode_select6_v2(sg_fd, pf, false, sp, paramp, param_len,
+                                 noisy, verbose);
+}
+
+/* Invokes a SCSI MODE SELECT (10) command.  Return of 0 -> success,
+ * various SG_LIB_CAT_* positive values or -1 -> other errors,
+ * v2 adds rtd (revert to defaults) bit (spc5r11).  */
+int
+sg_ll_mode_select10_v2(int sg_fd, bool pf, bool rtd, bool sp, void * paramp,
+                       int param_len, bool noisy, int verbose)
+{
+    static const char * const cdb_s = "mode select(10)";
+    int res, ret, sense_cat;
     uint8_t modes_cdb[MODE_SELECT10_CMDLEN] =
         {MODE_SELECT10_CMD, 0, 0, 0, 0, 0, 0, 0, 0, 0};
     uint8_t sense_b[SENSE_BUFF_LEN];
     struct sg_pt_base * ptvp;
 
     modes_cdb[1] = (uint8_t)((pf ? 0x10 : 0x0) | (sp ? 0x1 : 0x0));
+    if (rtd)
+        modes_cdb[1] |= 0x2;
     sg_put_unaligned_be16((int16_t)param_len, modes_cdb + 7);
     if (param_len > 0xffff) {
-        pr2ws("%s: param_len too big\n", cdb_name_s);
+        pr2ws("%s: param_len too big\n", cdb_s);
         return -1;
     }
     if (verbose) {
-        pr2ws("    %s cdb: ", cdb_name_s);
-        for (k = 0; k < MODE_SELECT10_CMDLEN; ++k)
-            pr2ws("%02x ", modes_cdb[k]);
-        pr2ws("\n");
+        char b[128];
+
+        pr2ws("    %s cdb: %s\n", cdb_s,
+              sg_get_command_str(modes_cdb, MODE_SELECT10_CMDLEN, false,
+                                 sizeof(b), b));
     }
     if (verbose > 1) {
-        pr2ws("    %s parameter list\n", cdb_name_s);
+        pr2ws("    %s parameter list\n", cdb_s);
         hex2stderr((const uint8_t *)paramp, param_len, -1);
     }
 
-    if (NULL == ((ptvp = create_pt_obj(cdb_name_s))))
+    if (NULL == ((ptvp = create_pt_obj(cdb_s))))
         return -1;
     set_scsi_pt_cdb(ptvp, modes_cdb, sizeof(modes_cdb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     set_scsi_pt_data_out(ptvp, (uint8_t *)paramp, param_len);
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, cdb_name_s, res, SG_NO_DATA_IN, sense_b,
-                               noisy, verbose, &sense_cat);
-    if (-1 == ret)
-        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
-    else if (-2 == ret) {
+    ret = sg_cmds_process_resp(ptvp, cdb_s, res, noisy, verbose, &sense_cat);
+    if (-1 == ret) {
+        if (get_scsi_pt_transport_err(ptvp))
+            ret = SG_LIB_TRANSPORT_ERROR;
+        else
+            ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
+    } else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
         case SG_LIB_CAT_NO_SENSE:
@@ -553,6 +571,14 @@ sg_ll_mode_select10(int sg_fd, bool pf, bool sp, void * paramp, int param_len,
     return ret;
 }
 
+int
+sg_ll_mode_select10(int sg_fd, bool pf, bool sp, void * paramp,
+                       int param_len, bool noisy, int verbose)
+{
+    return sg_ll_mode_select10_v2(sg_fd, pf, false, sp, paramp, param_len,
+                                  noisy, verbose);
+}
+
 /* MODE SENSE commands yield a response that has header then zero or more
  * block descriptors followed by mode pages. In most cases users are
  * interested in the first mode page. This function returns the (byte)
@@ -788,8 +814,8 @@ sg_ll_log_sense_v2(int sg_fd, bool ppc, bool sp, int pc, int pg_code,
                    int mx_resp_len, int timeout_secs, int * residp,
                    bool noisy, int verbose)
 {
-    static const char * const cdb_name_s = "log sense";
-    int res, ret, k, sense_cat, resid;
+    static const char * const cdb_s = "log sense";
+    int res, ret, sense_cat, resid;
     uint8_t logs_cdb[LOG_SENSE_CMDLEN] =
         {LOG_SENSE_CMD, 0, 0, 0, 0, 0, 0, 0, 0, 0};
     uint8_t sense_b[SENSE_BUFF_LEN];
@@ -805,28 +831,31 @@ sg_ll_log_sense_v2(int sg_fd, bool ppc, bool sp, int pc, int pg_code,
     sg_put_unaligned_be16((int16_t)paramp, logs_cdb + 5);
     sg_put_unaligned_be16((int16_t)mx_resp_len, logs_cdb + 7);
     if (verbose) {
-        pr2ws("    %s cdb: ", cdb_name_s);
-        for (k = 0; k < LOG_SENSE_CMDLEN; ++k)
-            pr2ws("%02x ", logs_cdb[k]);
-        pr2ws("\n");
+        char b[128];
+
+        pr2ws("    %s cdb: %s\n", cdb_s,
+              sg_get_command_str(logs_cdb, LOG_SENSE_CMDLEN, false,
+                                 sizeof(b), b));
     }
     if (timeout_secs <= 0)
         timeout_secs = DEF_PT_TIMEOUT;
 
-    if (NULL == ((ptvp = create_pt_obj(cdb_name_s))))
+    if (NULL == ((ptvp = create_pt_obj(cdb_s))))
         goto gen_err;
     set_scsi_pt_cdb(ptvp, logs_cdb, sizeof(logs_cdb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     set_scsi_pt_data_in(ptvp, resp, mx_resp_len);
     res = do_scsi_pt(ptvp, sg_fd, timeout_secs, verbose);
-    ret = sg_cmds_process_resp(ptvp, cdb_name_s, res, mx_resp_len,
-                               sense_b, noisy, verbose, &sense_cat);
+    ret = sg_cmds_process_resp(ptvp, cdb_s, res, noisy, verbose, &sense_cat);
     resid = get_scsi_pt_resid(ptvp);
     if (residp)
         *residp = resid;
-    if (-1 == ret)
-        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
-    else if (-2 == ret) {
+    if (-1 == ret) {
+        if (get_scsi_pt_transport_err(ptvp))
+            ret = SG_LIB_TRANSPORT_ERROR;
+        else
+            ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
+    } else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
         case SG_LIB_CAT_NO_SENSE:
@@ -849,7 +878,7 @@ sg_ll_log_sense_v2(int sg_fd, bool ppc, bool sp, int pc, int pg_code,
     if (resid > 0) {
         if (resid > mx_resp_len) {
             pr2ws("%s: resid (%d) should never exceed requested len=%d\n",
-                  cdb_name_s, resid, mx_resp_len);
+                  cdb_s, resid, mx_resp_len);
             return ret ? ret : SG_LIB_CAT_MALFORMED;
         }
         /* zero unfilled section of response buffer */
@@ -869,15 +898,15 @@ sg_ll_log_select(int sg_fd, bool pcr, bool sp, int pc, int pg_code,
                  int subpg_code, uint8_t * paramp, int param_len,
                  bool noisy, int verbose)
 {
-    static const char * const cdb_name_s = "log select";
-    int res, ret, k, sense_cat;
+    static const char * const cdb_s = "log select";
+    int res, ret, sense_cat;
     uint8_t logs_cdb[LOG_SELECT_CMDLEN] =
         {LOG_SELECT_CMD, 0, 0, 0, 0, 0, 0, 0, 0, 0};
     uint8_t sense_b[SENSE_BUFF_LEN];
     struct sg_pt_base * ptvp;
 
     if (param_len > 0xffff) {
-        pr2ws("%s: param_len too big\n", cdb_name_s);
+        pr2ws("%s: param_len too big\n", cdb_s);
         return -1;
     }
     logs_cdb[1] = (uint8_t)((pcr ? 2 : 0) | (sp ? 1 : 0));
@@ -885,27 +914,30 @@ sg_ll_log_select(int sg_fd, bool pcr, bool sp, int pc, int pg_code,
     logs_cdb[3] = (uint8_t)(subpg_code & 0xff);
     sg_put_unaligned_be16((int16_t)param_len, logs_cdb + 7);
     if (verbose) {
-        pr2ws("    %s cdb: ", cdb_name_s);
-        for (k = 0; k < LOG_SELECT_CMDLEN; ++k)
-            pr2ws("%02x ", logs_cdb[k]);
-        pr2ws("\n");
+        char b[128];
+
+        pr2ws("    %s cdb: %s\n", cdb_s,
+              sg_get_command_str(logs_cdb, LOG_SELECT_CMDLEN, false,
+                                 sizeof(b), b));
     }
     if ((verbose > 1) && (param_len > 0)) {
-        pr2ws("    %s parameter list\n", cdb_name_s);
+        pr2ws("    %s parameter list\n", cdb_s);
         hex2stderr(paramp, param_len, -1);
     }
 
-    if (NULL == ((ptvp = create_pt_obj(cdb_name_s))))
+    if (NULL == ((ptvp = create_pt_obj(cdb_s))))
         return -1;
     set_scsi_pt_cdb(ptvp, logs_cdb, sizeof(logs_cdb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     set_scsi_pt_data_out(ptvp, paramp, param_len);
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, cdb_name_s, res, SG_NO_DATA_IN, sense_b,
-                               noisy, verbose, &sense_cat);
-    if (-1 == ret)
-        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
-    else if (-2 == ret) {
+    ret = sg_cmds_process_resp(ptvp, cdb_s, res, noisy, verbose, &sense_cat);
+    if (-1 == ret) {
+        if (get_scsi_pt_transport_err(ptvp))
+            ret = SG_LIB_TRANSPORT_ERROR;
+        else
+            ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
+    } else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
         case SG_LIB_CAT_NO_SENSE:
@@ -930,14 +962,16 @@ sg_ll_log_select(int sg_fd, bool pcr, bool sp, int pc, int pg_code,
  * and fl(mmc) one bit field. This is the cause of the awkardly named
  * pc_mod__fl_num and noflush__fl arguments to this function.
  *  */
-int
-sg_ll_start_stop_unit(int sg_fd, bool immed, int pc_mod__fl_num,
-                      int power_cond, bool noflush__fl, bool loej, bool start,
-                      bool noisy, int verbose)
+static int
+sg_ll_start_stop_unit_com(struct sg_pt_base * ptvp, int sg_fd, bool immed,
+                          int pc_mod__fl_num, int power_cond, bool noflush__fl,
+                          bool loej, bool start, bool noisy, int verbose)
 {
-    static const char * const cdb_name_s = "start stop unit";
-    int k, res, ret, sense_cat;
-    struct sg_pt_base * ptvp;
+    static const char * const cdb_s = "start stop unit";
+    bool ptvp_given = false;
+    bool local_sense = true;
+    bool local_cdb = true;
+    int res, ret, sense_cat;
     uint8_t ssuBlk[START_STOP_CMDLEN] = {START_STOP_CMD, 0, 0, 0, 0, 0};
     uint8_t sense_b[SENSE_BUFF_LEN];
 
@@ -952,22 +986,38 @@ sg_ll_start_stop_unit(int sg_fd, bool immed, int pc_mod__fl_num,
     if (start)
         ssuBlk[4] |= 0x1;
     if (verbose) {
-        pr2ws("    %s command:", cdb_name_s);
-        for (k = 0; k < (int)sizeof(ssuBlk); ++k)
-                pr2ws(" %02x", ssuBlk[k]);
-        pr2ws("\n");
-    }
+        char b[128];
 
-    if (NULL == ((ptvp = create_pt_obj(cdb_name_s))))
-        return -1;
-    set_scsi_pt_cdb(ptvp, ssuBlk, sizeof(ssuBlk));
-    set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
-    res = do_scsi_pt(ptvp, sg_fd, START_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, cdb_name_s, res, SG_NO_DATA_IN, sense_b,
-                               noisy, verbose, &sense_cat);
-    if (-1 == ret)
-        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
-    else if (-2 == ret) {
+        pr2ws("    %s cdb: %s\n", cdb_s,
+              sg_get_command_str(ssuBlk, sizeof(ssuBlk), false,
+                                 sizeof(b), b));
+    }
+    if (ptvp) {
+        ptvp_given = true;
+        partial_clear_scsi_pt_obj(ptvp);
+        if (get_scsi_pt_cdb_buf(ptvp))
+            local_cdb = false; /* N.B. Ignores locally built cdb */
+        else
+            set_scsi_pt_cdb(ptvp, ssuBlk, sizeof(ssuBlk));
+        if (get_scsi_pt_sense_buf(ptvp))
+            local_sense = false;
+        else
+            set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
+    } else {
+        ptvp = construct_scsi_pt_obj_with_fd(sg_fd, verbose);
+        if (NULL == ptvp)
+            return sg_convert_errno(ENOMEM);
+        set_scsi_pt_cdb(ptvp, ssuBlk, sizeof(ssuBlk));
+        set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
+    }
+    res = do_scsi_pt(ptvp, -1, START_PT_TIMEOUT, verbose);
+    ret = sg_cmds_process_resp(ptvp, cdb_s, res, noisy, verbose, &sense_cat);
+    if (-1 == ret) {
+        if (get_scsi_pt_transport_err(ptvp))
+            ret = SG_LIB_TRANSPORT_ERROR;
+        else
+            ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
+    } else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
         case SG_LIB_CAT_NO_SENSE:
@@ -979,10 +1029,38 @@ sg_ll_start_stop_unit(int sg_fd, bool immed, int pc_mod__fl_num,
         }
     } else
             ret = 0;
-    destruct_scsi_pt_obj(ptvp);
+    if (ptvp_given) {
+        if (local_sense)    /* stop caller trying to access local sense */
+            set_scsi_pt_sense(ptvp, NULL, 0);
+        if (local_cdb)
+            set_scsi_pt_cdb(ptvp, NULL, 0);
+    } else {
+        if (ptvp)
+            destruct_scsi_pt_obj(ptvp);
+    }
     return ret;
 }
 
+int
+sg_ll_start_stop_unit(int sg_fd, bool immed, int pc_mod__fl_num,
+                      int power_cond, bool noflush__fl, bool loej, bool start,
+                      bool noisy, int verbose)
+{
+    return sg_ll_start_stop_unit_com(NULL, sg_fd, immed, pc_mod__fl_num,
+                                     power_cond, noflush__fl, loej, start,
+                                     noisy, verbose);
+}
+
+int
+sg_ll_start_stop_unit_pt(struct sg_pt_base * ptvp, bool immed,
+                         int pc_mod__fl_num, int power_cond, bool noflush__fl,
+                         bool loej, bool start, bool noisy, int verbose)
+{
+    return sg_ll_start_stop_unit_com(ptvp, -1, immed, pc_mod__fl_num,
+                                     power_cond, noflush__fl, loej, start,
+                                     noisy, verbose);
+}
+
 /* Invokes a SCSI PREVENT ALLOW MEDIUM REMOVAL command
  * [was in SPC-3 but displaced from SPC-4 into SBC-3, MMC-5, SSC-3]
  * prevent==0 allows removal, prevent==1 prevents removal ...
@@ -991,8 +1069,8 @@ sg_ll_start_stop_unit(int sg_fd, bool immed, int pc_mod__fl_num,
 int
 sg_ll_prevent_allow(int sg_fd, int prevent, bool noisy, int verbose)
 {
-    static const char * const cdb_name_s = "prevent allow medium removal";
-    int k, res, ret, sense_cat;
+    static const char * const cdb_s = "prevent allow medium removal";
+    int res, ret, sense_cat;
     uint8_t p_cdb[PREVENT_ALLOW_CMDLEN] =
                 {PREVENT_ALLOW_CMD, 0, 0, 0, 0, 0};
     uint8_t sense_b[SENSE_BUFF_LEN];
@@ -1004,22 +1082,25 @@ sg_ll_prevent_allow(int sg_fd, int prevent, bool noisy, int verbose)
     }
     p_cdb[4] |= (prevent & 0x3);
     if (verbose) {
-        pr2ws("    %s cdb: ", cdb_name_s);
-        for (k = 0; k < PREVENT_ALLOW_CMDLEN; ++k)
-            pr2ws("%02x ", p_cdb[k]);
-        pr2ws("\n");
+        char b[128];
+
+        pr2ws("    %s cdb: %s\n", cdb_s,
+              sg_get_command_str(p_cdb, PREVENT_ALLOW_CMDLEN, false,
+                                 sizeof(b), b));
     }
 
-    if (NULL == ((ptvp = create_pt_obj(cdb_name_s))))
+    if (NULL == ((ptvp = create_pt_obj(cdb_s))))
         return -1;
     set_scsi_pt_cdb(ptvp, p_cdb, sizeof(p_cdb));
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
-    ret = sg_cmds_process_resp(ptvp, cdb_name_s, res, SG_NO_DATA_IN, sense_b,
-                               noisy, verbose, &sense_cat);
-    if (-1 == ret)
-        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
-    else if (-2 == ret) {
+    ret = sg_cmds_process_resp(ptvp, cdb_s, res, noisy, verbose, &sense_cat);
+    if (-1 == ret) {
+        if (get_scsi_pt_transport_err(ptvp))
+            ret = SG_LIB_TRANSPORT_ERROR;
+        else
+            ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
+    } else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
         case SG_LIB_CAT_NO_SENSE:
