@@ -1,8 +1,10 @@
 /*
- * Copyright (c) 2007-2018 Douglas Gilbert.
+ * Copyright (c) 2007-2021 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <stdlib.h>
@@ -17,7 +19,8 @@
 #include "sg_lib_data.h"
 
 
-const char * sg_lib_version_str = "2.41 20180223";/* spc5r17, sbc4r15 */
+const char * sg_lib_version_str = "2.83 20211007";
+/* spc6r05, sbc5r01, zbc2r10 */
 
 
 /* indexed by pdt; those that map to own index do not decay */
@@ -31,6 +34,22 @@ int sg_lib_pdt_decay_arr[32] = {
     0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, PDT_WLUN, PDT_UNKNOWN
 };
 
+/* SCSI Status values */
+struct sg_lib_simple_value_name_t sg_lib_sstatus_str_arr[] = {
+    {0x0,  "Good"},
+    {0x2,  "Check Condition"},
+    {0x4,  "Condition Met"},
+    {0x8,  "Busy"},
+    {0x10, "Intermediate (obsolete)"},
+    {0x14, "Intermediate-Condition Met (obsolete)"},
+    {0x18, "Reservation Conflict"},
+    {0x22, "Command terminated (obsolete)"},
+    {0x28, "Task Set Full"},
+    {0x30, "ACA Active"},
+    {0x40, "Task Aborted"},
+    {0xffff, NULL},
+};
+
 #ifdef SG_SCSI_STRINGS
 struct sg_lib_value_name_t sg_lib_normal_opcodes[] = {
     {0, 0, "Test Unit Ready"},
@@ -101,7 +120,8 @@ struct sg_lib_value_name_t sg_lib_normal_opcodes[] = {
     {0x37, 0, "Read defect data(10)"},
                         /* SBC-3 r31 recommends Read defect data(12) */
     {0x37, PDT_MCHANGER, "Initialize element status with range"},
-    {0x38, 0, "Medium scan"},
+    {0x38, 0, "Format with preset scan"},
+    {0x38, PDT_OCRW, "Medium scan"},
     {0x39, 0, "Compare"},               /* obsolete in SPC-4 r11 */
     {0x3a, 0, "Copy and verify"},       /* obsolete in SPC-4 r11 */
     {0x3b, 0, "Write buffer"},
@@ -273,6 +293,7 @@ struct sg_lib_value_name_t sg_lib_read_pos_arr[] = {
 /* Maintenance in [0xa3] service actions */
 struct sg_lib_value_name_t sg_lib_maint_in_arr[] = {
     {0x0, PDT_SAC, "Report assigned/unassigned p_extent"},
+    {0x0, PDT_ADC, "Report automation device attributes"},
     {0x1, PDT_SAC, "Report component device"},
     {0x2, PDT_SAC, "Report component device attachments"},
     {0x3, PDT_SAC, "Report peripheral device"},
@@ -302,6 +323,7 @@ struct sg_lib_value_name_t sg_lib_maint_in_arr[] = {
 /* Maintenance out [0xa4] service actions */
 struct sg_lib_value_name_t sg_lib_maint_out_arr[] = {
     {0x0, PDT_SAC, "Add peripheral device / component device"},
+    {0x0, PDT_ADC, "Set automation device attribute"},
     {0x1, PDT_SAC, "Attach to component device"},
     {0x2, PDT_SAC, "Exchange p_extent"},
     {0x3, PDT_SAC, "Exchange peripheral device / component device"},
@@ -340,6 +362,7 @@ struct sg_lib_value_name_t sg_lib_serv_in12_arr[] = {
 
 /* Service action out(12) [0xa9] service actions */
 struct sg_lib_value_name_t sg_lib_serv_out12_arr[] = {
+    {0x1f, PDT_ADC, "Set medium attribute"},
     {0xff, 0, "Impossible command name"},
     {0xffff, 0, NULL},
 };
@@ -356,6 +379,8 @@ struct sg_lib_value_name_t sg_lib_serv_in16_arr[] = {
     {0x16, 0, "Get stream status"},
     {0x17, 0, "Get physical element status"},   /* added sbc4r13 */
     {0x18, 0, "Remove element and truncate"},   /* added sbc4r13 */
+    {0x19, 0, "Restore elements and rebuild"},  /* added sbc4r19 */
+    {0x1a, 0, "Remove element and modify zones"},   /* added zbc2r07 */
     {0xffff, 0, NULL},
 };
 
@@ -517,12 +542,17 @@ struct sg_lib_value_name_t sg_lib_zoning_out_arr[] = {
     {0x2, PDT_ZBC, "Finish zone"},
     {0x3, PDT_ZBC, "Open zone"},
     {0x4, PDT_ZBC, "Reset write pointer"},
+    {0x10, PDT_ZBC, "Sequentialize zone"},      /* zbc2r01b */
     {0xffff, 0, NULL},
 };
 
 /* Zoning in [0x95] service actions */
 struct sg_lib_value_name_t sg_lib_zoning_in_arr[] = {
     {0x0, PDT_ZBC, "Report zones"},
+    {0x6, PDT_ZBC, "Report realms"},            /* zbc2r04 */
+    {0x7, PDT_ZBC, "Report zone domains"},      /* zbc2r04 */
+    {0x8, PDT_ZBC, "Zone activate"},            /* zbc2r04 */
+    {0x9, PDT_ZBC, "Zone query"},               /* zbc2r04 */
     {0xffff, 0, NULL},
 };
 
@@ -622,7 +652,7 @@ struct sg_lib_value_name_t sg_lib_read_attr_arr[] = {
 
 /* A conveniently formatted list of SCSI ASC/ASCQ codes and their
  * corresponding text can be found at: www.t10.org/lists/asc-num.txt
- * The following should match asc-num.txt dated 20150423 */
+ * The following should match asc-num.txt dated 20200817 */
 
 #ifdef SG_SCSI_STRINGS
 struct sg_lib_asc_ascq_range_t sg_lib_asc_ascq_range[] =
@@ -712,6 +742,8 @@ struct sg_lib_asc_ascq_t sg_lib_asc_ascq[] =
     {0x04,0x21,"Logical unit not ready, hard reset required"},
     {0x04,0x22,"Logical unit not ready, power cycle required"},
     {0x04,0x23,"Logical unit not ready, affiliation required"},
+    {0x04,0x24,"Depopulation in progress"},             /* spc5r15 */
+    {0x04,0x25,"Depopulation restoration in progress"}, /* spc6r02 */
     {0x05,0x00,"Logical unit does not respond to selection"},
     {0x06,0x00,"No reference position found"},
     {0x07,0x00,"Multiple peripheral devices selected"},
@@ -747,6 +779,7 @@ struct sg_lib_asc_ascq_t sg_lib_asc_ascq[] =
     {0x0B,0x11,"Warning - low operating humidity limit exceeded"},
     {0x0B,0x12,"Warning - microcode security at risk"},
     {0x0B,0x13,"Warning - microcode digital signature validation failure"},
+    {0x0B,0x14,"Warning - physical element status change"},     /* spc5r15 */
     {0x0C,0x00,"Write error"},
     {0x0C,0x01,"Write error - recovered with auto reallocation"},
     {0x0C,0x02,"Write error - auto reallocation failed"},
@@ -879,6 +912,7 @@ struct sg_lib_asc_ascq_t sg_lib_asc_ascq[] =
     {0x21,0x06,"Attempt to read invalid data"},
     {0x21,0x07,"Read boundary violation"},
     {0x21,0x08,"Misaligned write command"},
+    {0x21,0x09,"Attempt to access gap zone"},
     {0x22,0x00,"Illegal function (use 20 00, 24 00, or 26 00)"},
     {0x23,0x00,"Invalid token operation, cause not reportable"},
     {0x23,0x01,"Invalid token operation, unsupported token type"},
@@ -925,6 +959,7 @@ struct sg_lib_asc_ascq_t sg_lib_asc_ascq[] =
     {0x26,0x13,"Application tag mode page is invalid"},
     {0x26,0x14,"Tape stream mirroring prevented"},
     {0x26,0x15,"Copy source or copy destination not authorized"},
+    {0x26,0x16,"Fast copy not possible"},
     {0x27,0x00,"Write protected"},
     {0x27,0x01,"Hardware write protected"},
     {0x27,0x02,"Logical unit software write protected"},
@@ -986,6 +1021,8 @@ struct sg_lib_asc_ascq_t sg_lib_asc_ascq[] =
     {0x2C,0x0F,"Stream not open"},
     {0x2C,0x10,"Unwritten data in zone"},
     {0x2C,0x11,"Descriptor format sense data required"},
+    {0x2C,0x12,"Zone is inactive"},
+    {0x2C,0x13,"Well known logical unit access required"},      /* spc6r02 */
     {0x2D,0x00,"Overwrite error on update in place"},
     {0x2E,0x00,"Insufficient time for operation"},
     {0x2E,0x01,"Command timeout before processing"},
@@ -1017,6 +1054,8 @@ struct sg_lib_asc_ascq_t sg_lib_asc_ascq[] =
     {0x31,0x01,"Format command failed"},
     {0x31,0x02,"Zoned formatting failed due to spare linking"},
     {0x31,0x03,"Sanitize command failed"},
+    {0x31,0x04,"Depopulation failed"},               /* spc5r15 */
+    {0x31,0x05,"Depopulation restoration failed"},   /* spc6r02 */
     {0x32,0x00,"No defect spare location available"},
     {0x32,0x01,"Defect list update failure"},
     {0x33,0x00,"Tape length error"},
@@ -1034,6 +1073,7 @@ struct sg_lib_asc_ascq_t sg_lib_asc_ascq[] =
     {0x38,0x04,"Esn - media class event"},
     {0x38,0x06,"Esn - device busy class event"},
     {0x38,0x07,"Thin provisioning soft threshold reached"},
+    {0x38,0x08,"Depopulation interrupted"},     /* spc6r03 */
     {0x39,0x00,"Saving parameters not supported"},
     {0x3A,0x00,"Medium not present"},
     {0x3A,0x01,"Medium not present - tray closed"},
@@ -1068,6 +1108,7 @@ struct sg_lib_asc_ascq_t sg_lib_asc_ascq[] =
     {0x3B,0x1a,"Data transfer device removed"},
     {0x3B,0x1b,"Data transfer device inserted"},
     {0x3B,0x1c,"Too many logical objects on partition to support operation"},
+    {0x3B,0x20,"Element static information changed"},
     {0x3D,0x00,"Invalid bits in identify message"},
     {0x3E,0x00,"Logical unit has not self-configured yet"},
     {0x3E,0x01,"Logical unit failure"},
@@ -1479,16 +1520,17 @@ const char * sg_lib_pdt_strs[32] = {    /* should have 2**5 elements */
     "object based storage",
     "automation/driver interface",
     "security manager device",  /* obsolete, spc5r01 */
-    "zoned block commands",
+    "host managed zoned block",
     "0x15", "0x16", "0x17", "0x18",
     "0x19", "0x1a", "0x1b", "0x1c", "0x1d",
     "well known logical unit",
-    "no physical device on this lu",
+    "unknown or no device type", /* coupled with PQ=3 for not accessible
+                                    via this lu's port (try the other) */
 };
 
 const char * sg_lib_transport_proto_strs[] =
 {
-    "Fibre Channel Protocol for SCSI (FCP-4)",
+    "Fibre Channel Protocol for SCSI (FCP-5)",  /* now at fcp5r01 */
     "SCSI Parallel Interface (SPI-5)",  /* obsolete in spc5r01 */
     "Serial Storage Architecture SCSI-3 Protocol (SSA-S3P)",
     "Serial Bus Protocol for IEEE 1394 (SBP-3)",
@@ -1512,6 +1554,9 @@ struct sg_lib_value_name_t sg_lib_scsi_feature_sets[] =
     {SCSI_FS_SBC_BASE_2016, PDT_DISK, "SBC Base 2016"},
     {SCSI_FS_SBC_BASIC_PROV_2016, PDT_DISK, "Basic provisioning 2016"},
     {SCSI_FS_SBC_DRIVE_MAINT_2016, PDT_DISK, "Drive maintenance 2016"},
+    {SCSI_FS_ZBC_HOST_AWARE_2020, PDT_ZBC, "Host Aware 2020"},
+    {SCSI_FS_ZBC_HOST_MANAGED_2020, PDT_ZBC, "Host Managed 2020"},
+    {SCSI_FS_ZBC_DOMAINS_REALMS_2020, PDT_ZBC, "Domains and Realms 2020"},
     {0x0, 0, NULL},     /* 0x0 is reserved sfs; trailing sentinel */
 };
 
@@ -1550,6 +1595,7 @@ struct sg_lib_simple_value_name_t sg_lib_nvme_admin_cmd_arr[] =
     {0x81, "Security Send"},
     {0x82, "Security Receive"},
     {0x84, "Sanitize"},                 /* last NVM specific in 1.3a */
+    {0x86, "Get LBA status"},           /* NVM specific, new in 1.4 */
     /* Vendor specific 0xc0 to 0xff */
     {0xffff, NULL},                     /* Sentinel */
 };
@@ -1625,7 +1671,11 @@ struct sg_lib_value_name_t sg_lib_nvme_cmd_status_arr[] =
     {0x1d, 11, "Sanitize in progress"},
     {0x1e,  5, "SGL data block granularity invalid"},
     {0x1f,  5, "Command not supported for queue in CMB"},
+    {0x20,  18, "Namespace is write protected"},        /* NVMe 1.4 */
+    {0x21,  6, "Command interrupted"},                  /* NVMe 1.4 */
+    {0x22,  5, "Transient transport error"},            /* NVMe 1.4 */
 
+    /* 0x80 - 0xbf: I/O command set specific */
     /* Generic command status values, NVM (I/O) Command Set */
     {0x80, 12, "LBA out of range"},
     {0x81,  3, "Capacity exceeded"},
@@ -1668,8 +1718,11 @@ struct sg_lib_value_name_t sg_lib_nvme_cmd_status_arr[] =
     {0x11e,18, "Boot partition write prohibited"},
     {0x11f, 5, "Invalid controller identifier"},
     {0x120, 5, "Invalid secondary controller state"},
-    {0x121, 5, "Invalid number of controller resorces"},
-    {0x122, 5, "Invalid resorce identifier"},
+    {0x121, 5, "Invalid number of controller resources"},
+    {0x122, 5, "Invalid resource identifier"},
+    {0x123, 5, "Sanitize prohibited while PM enabled"},         /* NVMe 1.4 */
+    {0x124, 5, "ANA group identifier invalid"},                 /* NVMe 1.4 */
+    {0x125, 5, "ANA attach failed"},                            /* NVMe 1.4 */
 
     /* Command specific status values, Status Code Type (SCT): 1h
      * for NVM (I/O) Command Set */
@@ -1732,6 +1785,88 @@ struct sg_lib_4tuple_u8 sg_lib_scsi_status_sense_arr[] =
     {0xff, 0xff, 0xff, 0xff},
 };
 
+/* These are the error (or warning) exit status values and their associated
+ * strings. They combine utility input syntax errors, SCSI status and sense
+ * key categories, OS errors (e.g. ENODEV for device not found), one that
+ * indicates NVMe non-zero status plus listing those that a Unix OS generates
+ * for any executable (that fails). The convention is 0 means no error and
+ * that in Unix the exit status is an (unsigned) 8 bit value. */
+struct sg_value_2names_t sg_exit_str_arr[] = {
+    {0,  "No errors", "may also convey true"},
+    {1,  "Syntax error", "command line options (usually)"},
+    {2,  "Device not ready", "type: sense key"},
+    {3,  "Medium or hardware error", "type: sense key (plus blank check for "
+         "tape)"},
+    {5,  "Illegal request", "type: sense key, apart from Invalid opcode"},
+    {6,  "Unit attention", "type: sense key"},
+    {7,  "Data protect", "type: sense key; write protected media?"},
+    {9,  "Illegal request, Invalid opcode", "type: sense key + asc,ascq"},
+    {10, "Copy aborted", "type: sense key"},
+    {11, "Aborted command",
+         "type: sense key, other than protection related (asc=0x10)"},
+    {14, "Miscompare", "type: sense key"},
+    {15, "File error", NULL},
+    {17, "Illegal request with Info field", NULL},
+    {18, "Medium or hardware error with Info", NULL},
+    {20, "No sense key", "type: probably additional sense code"},
+    {21, "Recovered error (warning)", "tye: sense key"},
+         /* N.B. this is a warning not error */
+    {22, "LBA out of range", NULL},
+    {24, "Reservation conflict", "type: SCSI status"},
+    {25, "Condition met", "type: SCSI status"}, /* from PRE-FETCH command */
+    {26, "Busy", "type: SCSI status"},   /* could be transport issue */
+    {27, "Task set full", "type: SCSI status"},
+    {28, "ACA aactive", "type: SCSI status"},
+    {29, "Task aborted", "type: SCSI status"},
+    {31, "Contradict", "command line options contradict or select bad mode"},
+    {32, "Logic error", "unexpected situation, contact author"},
+    {33, "SCSI command timeout", NULL},         /* OS timed out command */
+    {34, "Windows error number", "doesn't fit in 7 bits"},
+    {35, "Transport error", "driver or interconnect error"},
+    {36, "No errors (false)", NULL},
+    {40, "Aborted command, protection error", NULL},
+    {41, "Aborted command, protection error with Info field", NULL},
+    {47, "flock (Unix system call) error", NULL},       /* ddpt */
+    {48, "NVMe command with non-zero status", NULL},
+    {50, "An OS error occurred", "(errno > 46 or negative)"},
+    /* OS errors (errno in Unix) from 1 to 46 mapped into this range */
+    {97, "Malformed SCSI command", "trouble building command"},
+    {98, "Some other sense error", "try '-v' option for more information"},
+    {99, "Some other error", "possible transport of driver issue"},
+    {100, "Parameter list length error", NULL}, /* these for ddpt, xcopy */
+    {101, "Invalid field in parameter", NULL},
+    {102, "Too many segments in parameters", NULL},
+    {103, "Target underrun", NULL},
+    {104, "Target overrun", NULL},
+    {105, "Operation in progress", NULL},
+    {106, "Insufficient resources to create ROD", NULL},
+    {107, "Insufficient resources to create ROD token", NULL},
+    {108, "Commands cleared by device server", NULL},
+    {109, "See leave_reason for error", NULL},        /* internal error */
+    /* DDPT_CAT_TOKOP_BASE: asc=0x23, ascq=110 follow */
+    {110, "Invalid token operation, cause not reportable", NULL},
+    {111, "Invalid token operation, unsupported token type", NULL},
+    {112, "Invalid token operation, remote token usage not supported", NULL},
+    {113, "Invalid token operation, remote token creation not supported",
+          NULL},
+    {114, "Invalid token operation, token unknown", NULL},
+    {115, "Invalid token operation, token corrupt", NULL},
+    {116, "Invalid token operation, token revoked", NULL},
+    {117, "Invalid token operation, token expired", NULL},
+    {118, "Invalid token operation, token cancelled", NULL},
+    {119, "Invalid token operation, token deleted", NULL},
+    {120, "Invalid token operation, invalid token length", NULL},
+
+    /* The following error codes are generated by a Unix OS */
+    {126, "Utility found but did not have execute permissions", NULL},
+    {127, "Utility to be executed was not found", NULL},
+    {128, "Utility stopped/aborted by signal number: 0", "signal # 0 ??"},
+    /* 128 + <signal_number>: signal number that aborted the utility.
+                              real time signals start at offset SIGRTMIN */
+    /* OS signals from 1 to 126 mapped into this range (129 to 254) */
+    {255, "Utility returned 255 or higher", "Windows error number?"},
+    {0xffff, NULL, NULL},       /* end marking sentinel */
+};
 
 #else           /* (SG_SCSI_STRINGS && HAVE_NVME && (! IGNORE_NVME)) */
 
@@ -1763,4 +1898,8 @@ struct sg_lib_4tuple_u8 sg_lib_scsi_status_sense_arr[] =
     {0xff, 0xff, 0xff, 0xff},
 };
 
+struct sg_value_2names_t sg_exit_str_arr[] = {
+    {0xffff, NULL, NULL},       /* end marking sentinel */
+};
+
 #endif           /* (SG_SCSI_STRINGS && HAVE_NVME && (! IGNORE_NVME)) */
