@@ -1,8 +1,10 @@
 /*
- * Copyright (c) 2009-2018 Douglas Gilbert.
+ * Copyright (c) 2009-2021 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <stdio.h>
@@ -23,15 +25,16 @@
 #include "sg_pt.h"
 #include "sg_unaligned.h"
 #include "sg_pr2serr.h"
+#include "sg_pr2serr.h"
 
 #if (HAVE_NVME && (! IGNORE_NVME))
 #include "sg_pt_nvme.h"
 #endif
 
+static const char * scsi_pt_version_str = "3.18 20210617";
 
-static const char * scsi_pt_version_str = "3.05 20180309";
-
-
+/* List of external functions that need to be defined for each OS are
+ * listed at the top of sg_pt_dummy.c   */
 
 const char *
 scsi_pt_version()
@@ -45,6 +48,7 @@ sg_pt_version()
     return scsi_pt_version_str;
 }
 
+
 #if (HAVE_NVME && (! IGNORE_NVME))
 /* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */
 
@@ -62,7 +66,7 @@ static const char * nvme_scsi_vendor_str = "NVMe    ";
 #define F_INV_OP                0x200
 
 /* Table of SCSI operation code (opcodes) supported by SNTL */
-struct sg_opcode_info_t sg_opcode_info_arr[] =
+static struct sg_opcode_info_t sg_opcode_info_arr[] =
 {
     {0x0, 0, 0, {6,              /* TEST UNIT READY */
       0, 0, 0, 0, 0xc7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0} },
@@ -70,14 +74,51 @@ struct sg_opcode_info_t sg_opcode_info_arr[] =
       0xe1, 0, 0, 0xff, 0xc7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0} },
     {0x12, 0, 0, {6,            /* INQUIRY */
       0xe3, 0xff, 0xff, 0xff, 0xc7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0} },
+    {0x1b, 0, 0, {6,            /* START STOP UNIT */
+      0x1, 0, 0xf, 0xf7, 0xc7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0} },
     {0x1c, 0, 0, {6,            /* RECEIVE DIAGNOSTIC RESULTS */
       0x1, 0xff, 0xff, 0xff, 0xc7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0} },
     {0x1d, 0, 0, {6,            /* SEND DIAGNOSTIC */
       0xf7, 0x0, 0xff, 0xff, 0xc7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0} },
+    {0x25, 0, 0, {10,            /* READ CAPACITY(10) */
+      0x1, 0xff, 0xff, 0xff, 0xff, 0, 0, 0x1, 0xc7, 0, 0, 0, 0, 0, 0} },
+    {0x28, 0, 0, {10,            /* READ(10) */
+      0xff, 0xff, 0xff, 0xff, 0xff, 0x3f, 0xff, 0xff, 0xc7, 0, 0, 0, 0,
+      0, 0} },
+    {0x2a, 0, 0, {10,            /* WRITE(10) */
+      0xfb, 0xff, 0xff, 0xff, 0xff, 0x3f, 0xff, 0xff, 0xc7, 0, 0, 0, 0,
+      0, 0} },
+    {0x2f, 0, 0, {10,            /* VERIFY(10) */
+      0xf6, 0xff, 0xff, 0xff, 0xff, 0x3f, 0xff, 0xff, 0xc7, 0, 0, 0, 0,
+      0, 0} },
+    {0x35, 0, 0, {10,            /* SYNCHRONIZE CACHE(10) */
+      0x7, 0xff, 0xff, 0xff, 0xff, 0x3f, 0xff, 0xff, 0xc7, 0, 0, 0, 0,
+      0, 0} },
+    {0x41, 0, 0, {10,            /* WRITE SAME(10) */
+      0xff, 0xff, 0xff, 0xff, 0xff, 0x3f, 0xff, 0xff, 0xc7, 0, 0, 0, 0,
+      0, 0} },
     {0x55, 0, 0, {10,           /* MODE SELECT(10) */
       0x13, 0x0, 0x0, 0x0, 0x0, 0x0, 0xff, 0xff, 0xc7, 0, 0, 0, 0, 0, 0} },
     {0x5a, 0, 0, {10,           /* MODE SENSE(10) */
       0x18, 0xff, 0xff, 0x0, 0x0, 0x0, 0xff, 0xff, 0xc7, 0, 0, 0, 0, 0, 0} },
+    {0x88, 0, 0, {16,            /* READ(16) */
+      0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+      0xff, 0xff, 0xff, 0xc7} },
+    {0x8a, 0, 0, {16,            /* WRITE(16) */
+      0xfb, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+      0xff, 0xff, 0xff, 0xc7} },
+    {0x8f, 0, 0, {16,            /* VERIFY(16) */
+      0xf6, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+      0xff, 0xff, 0x3f, 0xc7} },
+    {0x91, 0, 0, {16,            /* SYNCHRONIZE CACHE(16) */
+      0x7, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+      0xff, 0xff, 0x3f, 0xc7} },
+    {0x93, 0, 0, {16,            /* WRITE SAME(16) */
+      0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+      0xff, 0xff, 0x3f, 0xc7} },
+    {0x9e, 0x10, F_SA_LOW, {16,  /* READ CAPACITY(16) [service action in] */
+      0x10, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
+      0xff, 0xff, 0x1, 0xc7} },
     {0xa0, 0, 0, {12,           /* REPORT LUNS */
       0xe3, 0xff, 0, 0, 0, 0xff, 0xff, 0xff, 0xff, 0, 0xc7, 0, 0, 0, 0} },
     {0xa3, 0xc, F_SA_LOW, {12,  /* REPORT SUPPORTED OPERATION CODES */
@@ -90,7 +131,13 @@ struct sg_opcode_info_t sg_opcode_info_arr[] =
       0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0} },
 };
 
-
+/* Returns pointer to array of struct sg_opcode_info_t of SCSI commands
+ * translated to NVMe. */
+const struct sg_opcode_info_t *
+sg_get_opcode_translation(void)
+{
+    return sg_opcode_info_arr;
+}
 
 /* Given the NVMe Identify controller response and optionally the NVMe
  * Identify namespace response (NULL otherwise), generate the SCSI VPD
@@ -210,6 +257,45 @@ resp_disconnect_pg(uint8_t * p, int pcontrol)
     return sizeof(disconnect_pg);
 }
 
+static uint8_t caching_m_pg[] = {0x8, 18, 0x14, 0, 0xff, 0xff, 0, 0,
+                                 0xff, 0xff, 0xff, 0xff, 0x80, 0x14, 0, 0,
+                                 0, 0, 0, 0};
+
+/* Control mode page (SBC) for mode_sense */
+static int
+resp_caching_m_pg(unsigned char *p, int pcontrol, bool wce)
+{       /* Caching page for mode_sense */
+        uint8_t ch_caching_m_pg[] = {/* 0x8, 18, */ 0x4, 0, 0, 0, 0, 0,
+                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+        uint8_t d_caching_m_pg[] = {0x8, 18, 0x14, 0, 0xff, 0xff, 0, 0,
+                0xff, 0xff, 0xff, 0xff, 0x80, 0x14, 0, 0,     0, 0, 0, 0};
+
+        // if (SDEBUG_OPT_N_WCE & sdebug_opts)
+                caching_m_pg[2] &= ~0x4;  /* set WCE=0 (default WCE=1) */
+        if ((0 == pcontrol) || (3 == pcontrol)) {
+            if (wce)
+                caching_m_pg[2] |= 0x4;
+            else
+                caching_m_pg[2] &= ~0x4;
+        }
+        memcpy(p, caching_m_pg, sizeof(caching_m_pg));
+        if (1 == pcontrol) {
+            if (wce)
+                ch_caching_m_pg[2] |= 0x4;
+            else
+                ch_caching_m_pg[2] &= ~0x4;
+            memcpy(p + 2, ch_caching_m_pg, sizeof(ch_caching_m_pg));
+        }
+        else if (2 == pcontrol) {
+            if (wce)
+                d_caching_m_pg[2] |= 0x4;
+            else
+                d_caching_m_pg[2] &= ~0x4;
+            memcpy(p, d_caching_m_pg, sizeof(d_caching_m_pg));
+        }
+        return sizeof(caching_m_pg);
+}
+
 static uint8_t ctrl_m_pg[] = {0xa, 10, 2, 0, 0, 0, 0, 0,
                               0, 0, 0x2, 0x4b};
 
@@ -230,6 +316,32 @@ resp_ctrl_m_pg(uint8_t *p, int pcontrol)
     return sizeof(ctrl_m_pg);
 }
 
+static uint8_t ctrl_ext_m_pg[] = {0x4a, 0x1, 0, 0x1c,  0, 0, 0x40, 0,
+                                  0, 0, 0, 0,  0, 0, 0, 0,
+                                  0, 0, 0, 0,  0, 0, 0, 0,
+                                  0, 0, 0, 0,  0, 0, 0, 0, };
+
+/* Control Extension mode page [0xa,0x1] for mode_sense */
+static int
+resp_ctrl_ext_m_pg(uint8_t *p, int pcontrol)
+{
+    uint8_t ch_ctrl_ext_m_pg[] = {/* 0x4a, 0x1, 0, 0x1c, */ 0, 0, 0, 0,
+                         0, 0, 0, 0,  0, 0, 0, 0,
+                         0, 0, 0, 0,  0, 0, 0, 0,
+                         0, 0, 0, 0,  0, 0, 0, 0, };
+    uint8_t d_ctrl_ext_m_pg[] = {0x4a, 0x1, 0, 0x1c,  0, 0, 0x40, 0,
+                         0, 0, 0, 0,  0, 0, 0, 0,
+                         0, 0, 0, 0,  0, 0, 0, 0,
+                         0, 0, 0, 0,  0, 0, 0, 0, };
+
+    memcpy(p, ctrl_ext_m_pg, sizeof(ctrl_ext_m_pg));
+    if (1 == pcontrol)
+        memcpy(p + 4, ch_ctrl_ext_m_pg, sizeof(ch_ctrl_ext_m_pg));
+    else if (2 == pcontrol)
+        memcpy(p, d_ctrl_ext_m_pg, sizeof(d_ctrl_ext_m_pg));
+    return sizeof(ctrl_ext_m_pg);
+}
+
 static uint8_t iec_m_pg[] = {0x1c, 0xa, 0x08, 0, 0, 0, 0, 0, 0, 0, 0x0, 0x0};
 
 /* Informational Exceptions control mode page for mode_sense */
@@ -271,8 +383,10 @@ resp_vs_ua_m_pg(uint8_t *p, int pcontrol)
 void
 sntl_init_dev_stat(struct sg_sntl_dev_state_t * dsp)
 {
-    dsp->descriptor_sense = !! (0x4 & ctrl_m_pg[2]);
-    dsp->enclosure_override = vs_ua_m_pg[2];
+    if (dsp) {
+        dsp->scsi_dsense = !! (0x4 & ctrl_m_pg[2]);
+        dsp->enclosure_override = vs_ua_m_pg[2];
+    }
 }
 
 
@@ -299,7 +413,7 @@ sntl_resp_mode_sense10(const struct sg_sntl_dev_state_t * dsp,
     pcode = cdbp[2] & 0x3f;
     subpcode = cdbp[3];
     llbaa = !!(cdbp[1] & 0x10);
-    is_disk = (dsp->pdt == PDT_DISK);
+    is_disk = ((dsp->pdt == PDT_DISK) || (dsp->pdt == PDT_ZBC));
     if (is_disk && !dbd)
         bd_len = llbaa ? 16 : 8;
     else
@@ -333,33 +447,59 @@ sntl_resp_mode_sense10(const struct sg_sntl_dev_state_t * dsp,
         offset += bd_len;
         ap = arr + offset;
     }
-
-    if ((subpcode > 0x0) && (subpcode < 0xff)) {
-        resp->asc = INVALID_FIELD_IN_CDB;
-        resp->in_byte = 3;
-        resp->in_bit = 255;
-        goto err_out;
-    }
     bad_pcode = false;
 
     switch (pcode) {
     case 0x2:       /* Disconnect-Reconnect page, all devices */
-        len = resp_disconnect_pg(ap, pcontrol);
+        if (0x0 == subpcode)
+            len = resp_disconnect_pg(ap, pcontrol);
+        else {
+            len = 0;
+            bad_pcode = true;
+        }
+        offset += len;
+        break;
+    case 0x8:       /* Caching Mode page, disk (like) devices */
+        if (! is_disk) {
+            len = 0;
+            bad_pcode = true;
+        } else if (0x0 == subpcode)
+            len = resp_caching_m_pg(ap, pcontrol, dsp->wce);
+        else {
+            len = 0;
+            bad_pcode = true;
+        }
         offset += len;
         break;
     case 0xa:       /* Control Mode page, all devices */
-        len = resp_ctrl_m_pg(ap, pcontrol);
+        if (0x0 == subpcode)
+            len = resp_ctrl_m_pg(ap, pcontrol);
+        else if (0x1 == subpcode)
+            len = resp_ctrl_ext_m_pg(ap, pcontrol);
+        else {
+            len = 0;
+            bad_pcode = true;
+        }
         offset += len;
         break;
     case 0x1c:      /* Informational Exceptions Mode page, all devices */
-        len = resp_iec_m_pg(ap, pcontrol);
+        if (0x0 == subpcode)
+            len = resp_iec_m_pg(ap, pcontrol);
+        else {
+            len = 0;
+            bad_pcode = true;
+        }
         offset += len;
         break;
     case 0x3f:      /* Read all Mode pages */
         if ((0 == subpcode) || (0xff == subpcode)) {
             len = 0;
             len = resp_disconnect_pg(ap + len, pcontrol);
+            if (is_disk)
+                len += resp_caching_m_pg(ap + len, pcontrol, dsp->wce);
             len += resp_ctrl_m_pg(ap + len, pcontrol);
+            if (0xff == subpcode)
+                len += resp_ctrl_ext_m_pg(ap + len, pcontrol);
             len += resp_iec_m_pg(ap + len, pcontrol);
             len += resp_vs_ua_m_pg(ap + len, pcontrol);
             offset += len;
@@ -371,6 +511,7 @@ sntl_resp_mode_sense10(const struct sg_sntl_dev_state_t * dsp,
         }
         break;
     case 0x0:       /* Vendor specific "Unit Attention" mode page */
+        /* all sub-page codes ?? */
         len = resp_vs_ua_m_pg(ap, pcontrol);
         offset += len;
         break;      /* vendor is "NVMe    " (from INQUIRY field) */
@@ -406,6 +547,7 @@ sntl_resp_mode_select10(struct sg_sntl_dev_state_t * dsp,
                         struct sg_sntl_result_t * resp)
 {
     int pf, sp, ps, md_len, bd_len, off, spf, pg_len, rlen, param_len, mpage;
+    int sub_mpage;
     uint8_t arr[SDEBUG_MAX_MSELECT_SZ];
 
     memset(resp, 0, sizeof(*resp));
@@ -448,24 +590,42 @@ sntl_resp_mode_select10(struct sg_sntl_dev_state_t * dsp,
     spf = !!(arr[off] & 0x40);
     pg_len = spf ? (sg_get_unaligned_be16(arr + off + 2) + 4) :
                    (arr[off + 1] + 2);
+    sub_mpage = spf ? arr[off + 1] : 0;
     if ((pg_len + off) > param_len) {
         resp->asc = PARAMETER_LIST_LENGTH_ERR;
         goto err_out;
     }
     switch (mpage) {
+    case 0x8:      /* Caching Mode page */
+        if (0x0 == sub_mpage) {
+            if (caching_m_pg[1] == arr[off + 1]) {
+                memcpy(caching_m_pg + 2, arr + off + 2,
+                       sizeof(caching_m_pg) - 2);
+                dsp->wce = !!(caching_m_pg[2] & 0x4);
+                dsp->wce_changed = true;
+                break;
+            }
+        }
+        goto def_case;
     case 0xa:      /* Control Mode page */
-        if (ctrl_m_pg[1] == arr[off + 1]) {
-            memcpy(ctrl_m_pg + 2, arr + off + 2,
-                   sizeof(ctrl_m_pg) - 2);
-            dsp->descriptor_sense = !!(ctrl_m_pg[2] & 0x4);
+        if (0x0 == sub_mpage) {
+            if (ctrl_m_pg[1] == arr[off + 1]) {
+                memcpy(ctrl_m_pg + 2, arr + off + 2,
+                       sizeof(ctrl_m_pg) - 2);
+                dsp->scsi_dsense = !!(ctrl_m_pg[2] & 0x4);
+                break;
+            }
         }
-        break;
-    case 0x1c:      /* Informational Exceptions Mode page */
-        if (iec_m_pg[1] == arr[off + 1]) {
-            memcpy(iec_m_pg + 2, arr + off + 2,
-                   sizeof(iec_m_pg) - 2);
+        goto def_case;
+    case 0x1c:      /* Informational Exceptions Mode page (SBC) */
+        if (0x0 == sub_mpage) {
+            if (iec_m_pg[1] == arr[off + 1]) {
+                memcpy(iec_m_pg + 2, arr + off + 2,
+                       sizeof(iec_m_pg) - 2);
+                break;
+            }
         }
-        break;
+        goto def_case;
     case 0x0:       /* Vendor specific "Unit Attention" mode page */
         if (vs_ua_m_pg[1] == arr[off + 1]) {
             memcpy(vs_ua_m_pg + 2, arr + off + 2,
@@ -474,6 +634,7 @@ sntl_resp_mode_select10(struct sg_sntl_dev_state_t * dsp,
         }
         break;
     default:
+def_case:
         resp->asc = INVALID_FIELD_IN_PARAM_LIST;
         resp->in_byte = off;
         resp->in_bit = 5;
