@@ -1,11 +1,13 @@
 /*
- * Copyright (c) 2005-2018 Douglas Gilbert.
+ * Copyright (c) 2005-2021 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
-/* sg_pt_linux version 1.40 20180309 */
+/* sg_pt_linux version 1.54 20210923 */
 
 
 #include <stdio.h>
@@ -35,6 +37,7 @@
 #include "sg_lib.h"
 #include "sg_linux_inc.h"
 #include "sg_pt_linux.h"
+#include "sg_pr2serr.h"
 
 
 #ifdef major
@@ -52,6 +55,12 @@
 
 #define DEF_TIMEOUT 60000       /* 60,000 millisecs (60 seconds) */
 
+/* sg driver displayed format: [x]xyyzz --> [x]x.[y]y.zz */
+#define SG_LINUX_SG_VER_V4_BASE 40000   /* lowest sg driver version with
+                                         * v4 interface */
+#define SG_LINUX_SG_VER_V4_FULL 40030   /* lowest version with full v4
+                                         * interface */
+
 static const char * linux_host_bytes[] = {
     "DID_OK", "DID_NO_CONNECT", "DID_BUS_BUSY", "DID_TIME_OUT",
     "DID_BAD_TARGET", "DID_ABORT", "DID_PARITY", "DID_ERROR",
@@ -62,8 +71,11 @@ static const char * linux_host_bytes[] = {
     "DID_NEXUS_FAILURE (reservation conflict)",
     "DID_ALLOC_FAILURE",
     "DID_MEDIUM_ERROR",
+    "DID_TRANSPORT_MARGINAL",   /*0x14 */
 };
 
+/* These where made obsolete around lk 5.12.0 . Only DRIVER_SENSE [0x8] is
+ * defined in scsi/sg.h for backward comaptibility */
 static const char * linux_driver_bytes[] = {
     "DRIVER_OK", "DRIVER_BUSY", "DRIVER_SOFT", "DRIVER_MEDIA",
     "DRIVER_ERROR", "DRIVER_INVALID", "DRIVER_TIMEOUT", "DRIVER_HARD",
@@ -103,29 +115,13 @@ bool sg_bsg_nvme_char_major_checked = false;
 int sg_bsg_major = 0;
 volatile int sg_nvme_char_major = 0;
 
+bool sg_checked_version_num = false;
+int sg_driver_version_num = 0;
+bool sg_duration_set_nano = false;
+
 long sg_lin_page_size = 4096;   /* default, overridden with correct value */
 
 
-#if defined(__GNUC__) || defined(__clang__)
-static int pr2ws(const char * fmt, ...)
-        __attribute__ ((format (printf, 1, 2)));
-#else
-static int pr2ws(const char * fmt, ...);
-#endif
-
-
-static int
-pr2ws(const char * fmt, ...)
-{
-    va_list args;
-    int n;
-
-    va_start(args, fmt);
-    n = vfprintf(sg_warnings_strm ? sg_warnings_strm : stderr, fmt, args);
-    va_end(args);
-    return n;
-}
-
 /* This function only needs to be called once (unless a NVMe controller
  * can be hot-plugged into system in which case it should be called
  * (again) after that event). */
@@ -335,16 +331,6 @@ check_pt_file_handle(int dev_fd, const char * device_name, int verbose)
 #endif
 
 
-/* Returns >= 0 if successful. If error in Unix returns negated errno. */
-int
-scsi_pt_open_device(const char * device_name, bool read_only, int verbose)
-{
-    int oflags = O_NONBLOCK;
-
-    oflags |= (read_only ? O_RDONLY : O_RDWR);
-    return scsi_pt_open_flags(device_name, oflags, verbose);
-}
-
 /* Similar to scsi_pt_open_device() but takes Unix style open flags OR-ed */
 /* together. The 'flags' argument is advisory and may be ignored. */
 /* Returns >= 0 if successful, otherwise returns negated errno. */
@@ -370,6 +356,16 @@ scsi_pt_open_flags(const char * device_name, int flags, int verbose)
     return fd;
 }
 
+/* Returns >= 0 if successful. If error in Unix returns negated errno. */
+int
+scsi_pt_open_device(const char * device_name, bool read_only, int verbose)
+{
+    int oflags = O_NONBLOCK;
+
+    oflags |= (read_only ? O_RDONLY : O_RDWR);
+    return scsi_pt_open_flags(device_name, oflags, verbose);
+}
+
 /* Returns 0 if successful. If error in Unix returns negated errno. */
 int
 scsi_pt_close_device(int device_fd)
@@ -382,25 +378,30 @@ scsi_pt_close_device(int device_fd)
     return res;
 }
 
+#if (HAVE_NVME && (! IGNORE_NVME))
+static bool checked_ev_dsense = false;
+static bool ev_dsense = false;
+#endif
+
 
 /* Caller should additionally call get_scsi_pt_os_err() after this call */
 struct sg_pt_base *
 construct_scsi_pt_obj_with_fd(int dev_fd, int verbose)
 {
-    int err;
     struct sg_pt_linux_scsi * ptp;
 
-    /* The following 2 lines are temporary. It is to avoid a NULL pointer
-     * crash when an old utility is used with a newer library built after
-     * the sg_warnings_strm cleanup */
-    if (NULL == sg_warnings_strm)
-        sg_warnings_strm = stderr;
-
     ptp = (struct sg_pt_linux_scsi *)
           calloc(1, sizeof(struct sg_pt_linux_scsi));
     if (ptp) {
+        int err;
+
 #if (HAVE_NVME && (! IGNORE_NVME))
         sntl_init_dev_stat(&ptp->dev_stat);
+        if (! checked_ev_dsense) {
+            ev_dsense = sg_get_initial_dsense();
+            checked_ev_dsense = true;
+        }
+        ptp->dev_stat.scsi_dsense = ev_dsense;
 #endif
         err = set_pt_file_handle((struct sg_pt_base *)ptp, dev_fd, verbose);
         if ((0 == err) && (! ptp->is_nvme)) {
@@ -438,8 +439,8 @@ destruct_scsi_pt_obj(struct sg_pt_base * vp)
             ptp->free_nvme_id_ctlp = NULL;
             ptp->nvme_id_ctlp = NULL;
         }
-        if (ptp)
-            free(ptp);
+        if (vp)
+            free(vp);
     }
 }
 
@@ -447,17 +448,20 @@ destruct_scsi_pt_obj(struct sg_pt_base * vp)
 void
 clear_scsi_pt_obj(struct sg_pt_base * vp)
 {
-    bool is_sg, is_bsg, is_nvme;
-    int fd;
-    uint32_t nvme_nsid;
     struct sg_pt_linux_scsi * ptp = &vp->impl;
 
     if (ptp) {
+        bool is_sg, is_bsg, is_nvme;
+        int fd;
+        uint32_t nvme_nsid;
+        struct sg_sntl_dev_state_t dev_stat;
+
         fd = ptp->dev_fd;
         is_sg = ptp->is_sg;
         is_bsg = ptp->is_bsg;
         is_nvme = ptp->is_nvme;
         nvme_nsid = ptp->nvme_nsid;
+        dev_stat = ptp->dev_stat;
         if (ptp->free_nvme_id_ctlp)
             free(ptp->free_nvme_id_ctlp);
         memset(ptp, 0, sizeof(struct sg_pt_linux_scsi));
@@ -472,11 +476,61 @@ clear_scsi_pt_obj(struct sg_pt_base * vp)
         ptp->is_sg = is_sg;
         ptp->is_bsg = is_bsg;
         ptp->is_nvme = is_nvme;
-        ptp->nvme_direct = false;
+        ptp->nvme_our_sntl = false;
         ptp->nvme_nsid = nvme_nsid;
+        ptp->dev_stat = dev_stat;
     }
 }
 
+void
+partial_clear_scsi_pt_obj(struct sg_pt_base * vp)
+{
+    struct sg_pt_linux_scsi * ptp = &vp->impl;
+
+    if (NULL == ptp)
+        return;
+    ptp->in_err = 0;
+    ptp->os_err = 0;
+    ptp->io_hdr.device_status = 0;
+    ptp->io_hdr.transport_status = 0;
+    ptp->io_hdr.driver_status = 0;
+    ptp->io_hdr.din_xferp = 0;
+    ptp->io_hdr.din_xfer_len = 0;
+    ptp->io_hdr.dout_xferp = 0;
+    ptp->io_hdr.dout_xfer_len = 0;
+    ptp->nvme_result = 0;
+}
+
+#ifndef SG_SET_GET_EXTENDED
+
+/* If both sei_wr_mask and sei_rd_mask are 0, this ioctl does nothing */
+struct sg_extended_info {
+    uint32_t   sei_wr_mask;    /* OR-ed SG_SEIM_* user->driver values */
+    uint32_t   sei_rd_mask;    /* OR-ed SG_SEIM_* driver->user values */
+    uint32_t   ctl_flags_wr_mask;      /* OR-ed SG_CTL_FLAGM_* values */
+    uint32_t   ctl_flags_rd_mask;      /* OR-ed SG_CTL_FLAGM_* values */
+    uint32_t   ctl_flags;      /* bit values OR-ed, see SG_CTL_FLAGM_* */
+    uint32_t   read_value;     /* write SG_SEIRV_*, read back related */
+
+    uint32_t   reserved_sz;    /* data/sgl size of pre-allocated request */
+    uint32_t   tot_fd_thresh;  /* total data/sgat for this fd, 0: no limit */
+    uint32_t   minor_index;    /* rd: kernel's sg device minor number */
+    uint32_t   share_fd;       /* SHARE_FD and CHG_SHARE_FD use this */
+    uint32_t   sgat_elem_sz;   /* sgat element size (must be power of 2) */
+    uint8_t    pad_to_96[52];  /* pad so struct is 96 bytes long */
+};
+
+#define SG_IOCTL_MAGIC_NUM 0x22
+
+#define SG_SET_GET_EXTENDED _IOWR(SG_IOCTL_MAGIC_NUM, 0x51,     \
+                                  struct sg_extended_info)
+
+#define SG_SEIM_CTL_FLAGS       0x1
+
+#define SG_CTL_FLAGM_TIME_IN_NS 0x1
+
+#endif
+
 /* Forget any previous dev_fd and install the one given. May attempt to
  * find file type (e.g. if pass-though) from OS so there could be an error.
  * Returns 0 for success or the same value as get_scsi_pt_os_err()
@@ -492,21 +546,87 @@ set_pt_file_handle(struct sg_pt_base * vp, int dev_fd, int verbose)
         sg_find_bsg_nvme_char_major(verbose);
     }
     ptp->dev_fd = dev_fd;
-    if (dev_fd >= 0)
+    if (dev_fd >= 0) {
         ptp->is_sg = check_file_type(dev_fd, &a_stat, &ptp->is_bsg,
                                      &ptp->is_nvme, &ptp->nvme_nsid,
                                      &ptp->os_err, verbose);
-    else {
+        if (ptp->is_sg && (! sg_checked_version_num)) {
+            if (ioctl(dev_fd, SG_GET_VERSION_NUM, &ptp->sg_version) < 0) {
+                ptp->sg_version = 0;
+                if (verbose > 3)
+                    pr2ws("%s: ioctl(SG_GET_VERSION_NUM) failed: errno: %d "
+                          "[%s]\n", __func__, errno, safe_strerror(errno));
+            } else {    /* got version number */
+                sg_driver_version_num = ptp->sg_version;
+                sg_checked_version_num = true;
+            }
+            if (verbose > 4) {
+                int ver = ptp->sg_version;
+
+                if (ptp->sg_version >= SG_LINUX_SG_VER_V4_BASE) {
+#ifdef IGNORE_LINUX_SGV4
+                    pr2ws("%s: sg driver version %d.%02d.%02d but config "
+                          "override back to v3\n", __func__, ver / 10000,
+                          (ver / 100) % 100, ver % 100);
+#else
+                    pr2ws("%s: sg driver version %d.%02d.%02d so choose v4\n",
+                          __func__, ver / 10000, (ver / 100) % 100,
+                          ver % 100);
+#endif
+                } else if (verbose > 5)
+                    pr2ws("%s: sg driver version %d.%02d.%02d so choose v3\n",
+                          __func__, ver / 10000, (ver / 100) % 100,
+                          ver % 100);
+            }
+        } else if (ptp->is_sg)
+            ptp->sg_version = sg_driver_version_num;
+
+        if (ptp->is_sg && (ptp->sg_version >= SG_LINUX_SG_VER_V4_FULL) &&
+            getenv("SG3_UTILS_LINUX_NANO")) {
+            struct sg_extended_info sei;
+            struct sg_extended_info * seip = &sei;
+
+            memset(seip, 0, sizeof(*seip));
+            /* try to override default of milliseconds */
+            seip->sei_wr_mask |= SG_SEIM_CTL_FLAGS;
+            seip->ctl_flags_wr_mask |= SG_CTL_FLAGM_TIME_IN_NS;
+            seip->ctl_flags |= SG_CTL_FLAGM_TIME_IN_NS;
+            if (ioctl(dev_fd, SG_SET_GET_EXTENDED, seip) < 0) {
+                if (verbose > 2)
+                    pr2ws("%s: unable to override milli --> nanoseconds: "
+                          "%s\n", __func__, safe_strerror(errno));
+            } else {
+                if (! sg_duration_set_nano)
+                    sg_duration_set_nano = true;
+                if (verbose > 5)
+                    pr2ws("%s: dev_fd=%d, succeeding in setting durations "
+                          "to nanoseconds\n", __func__, dev_fd);
+            }
+        } else if (ptp->is_sg && (ptp->sg_version >= SG_LINUX_SG_VER_V4_BASE)
+                   && getenv("SG3_UTILS_LINUX_NANO")) {
+            if (verbose > 2)
+                pr2ws("%s: dev_fd=%d, ignored SG3_UTILS_LINUX_NANO\nbecause "
+                      "base version sg version 4 driver\n", __func__, dev_fd);
+        }
+    } else {
         ptp->is_sg = false;
         ptp->is_bsg = false;
         ptp->is_nvme = false;
-        ptp->nvme_direct = false;
+        ptp->nvme_our_sntl = false;
         ptp->nvme_nsid = 0;
         ptp->os_err = 0;
     }
     return ptp->os_err;
 }
 
+int
+sg_linux_get_sg_version(const struct sg_pt_base * vp)
+{
+    const struct sg_pt_linux_scsi * ptp = &vp->impl;
+
+    return ptp->sg_version;
+}
+
 /* Valid file handles (which is the return value) are >= 0 . Returns -1
  * if there is no valid file handle. */
 int
@@ -523,21 +643,36 @@ set_scsi_pt_cdb(struct sg_pt_base * vp, const uint8_t * cdb,
 {
     struct sg_pt_linux_scsi * ptp = &vp->impl;
 
-    if (ptp->io_hdr.request)
-        ++ptp->in_err;
     ptp->io_hdr.request = (__u64)(sg_uintptr_t)cdb;
     ptp->io_hdr.request_len = cdb_len;
 }
 
+int
+get_scsi_pt_cdb_len(const struct sg_pt_base * vp)
+{
+    const struct sg_pt_linux_scsi * ptp = &vp->impl;
+
+    return ptp->io_hdr.request_len;
+}
+
+uint8_t *
+get_scsi_pt_cdb_buf(const struct sg_pt_base * vp)
+{
+    const struct sg_pt_linux_scsi * ptp = &vp->impl;
+
+    return (uint8_t *)(sg_uintptr_t)ptp->io_hdr.request;
+}
+
 void
 set_scsi_pt_sense(struct sg_pt_base * vp, uint8_t * sense,
                   int max_sense_len)
 {
     struct sg_pt_linux_scsi * ptp = &vp->impl;
 
-    if (ptp->io_hdr.response)
-        ++ptp->in_err;
-    memset(sense, 0, max_sense_len);
+    if (sense) {
+        if (max_sense_len > 0)
+            memset(sense, 0, max_sense_len);
+    }
     ptp->io_hdr.response = (__u64)(sg_uintptr_t)sense;
     ptp->io_hdr.max_response_len = max_sense_len;
 }
@@ -590,7 +725,7 @@ set_scsi_pt_packet_id(struct sg_pt_base * vp, int pack_id)
 {
     struct sg_pt_linux_scsi * ptp = &vp->impl;
 
-    ptp->io_hdr.spare_in = pack_id;
+    ptp->io_hdr.request_extra = pack_id;        /* was placed in spare_in */
 }
 
 void
@@ -653,15 +788,67 @@ set_scsi_pt_flags(struct sg_pt_base * vp, int flags)
     }
 }
 
+/* If supported it is the number of bytes requested to transfer less the
+ * number actually transferred. This it typically important for data-in
+ * transfers. For data-out (only) transfers, the 'dout_req_len -
+ * dout_act_len' is returned. For bidi transfer the "din" residual is
+ * returned. */
 /* N.B. Returns din_resid and ignores dout_resid */
 int
 get_scsi_pt_resid(const struct sg_pt_base * vp)
 {
     const struct sg_pt_linux_scsi * ptp = &vp->impl;
 
-    if (NULL == ptp)
+    if ((NULL == ptp) || (ptp->is_nvme && ! ptp->nvme_our_sntl))
         return 0;
-    return ptp->nvme_direct ? 0 : ptp->io_hdr.din_resid;
+    else if ((ptp->io_hdr.din_xfer_len > 0) &&
+             (ptp->io_hdr.dout_xfer_len > 0))
+        return ptp->io_hdr.din_resid;
+    else if (ptp->io_hdr.dout_xfer_len > 0)
+        return ptp->io_hdr.dout_resid;
+    return ptp->io_hdr.din_resid;
+}
+
+void
+get_pt_req_lengths(const struct sg_pt_base * vp, int * req_dinp,
+                   int * req_doutp)
+{
+    const struct sg_pt_linux_scsi * ptp = &vp->impl;
+
+    if (req_dinp) {
+        if (ptp->io_hdr.din_xfer_len > 0)
+            *req_dinp = ptp->io_hdr.din_xfer_len;
+        else
+            *req_dinp = 0;
+    }
+    if (req_doutp) {
+        if (ptp->io_hdr.dout_xfer_len > 0)
+            *req_doutp = ptp->io_hdr.dout_xfer_len;
+        else
+            *req_doutp = 0;
+    }
+}
+
+void
+get_pt_actual_lengths(const struct sg_pt_base * vp, int * act_dinp,
+                      int * act_doutp)
+{
+    const struct sg_pt_linux_scsi * ptp = &vp->impl;
+
+    if (act_dinp) {
+        if (ptp->io_hdr.din_xfer_len > 0) {
+            int res = ptp->io_hdr.din_xfer_len - ptp->io_hdr.din_resid;
+
+            *act_dinp = (res > 0) ? res : 0;
+        } else
+            *act_dinp = 0;
+    }
+    if (act_doutp) {
+        if (ptp->io_hdr.dout_xfer_len > 0)
+            *act_doutp = ptp->io_hdr.dout_xfer_len - ptp->io_hdr.dout_resid;
+        else
+            *act_doutp = 0;
+    }
 }
 
 int
@@ -671,8 +858,8 @@ get_scsi_pt_status_response(const struct sg_pt_base * vp)
 
     if (NULL == ptp)
         return 0;
-    return (int)(ptp->nvme_direct ? ptp->nvme_status :
-                                    ptp->io_hdr.device_status);
+    return (int)((ptp->is_nvme && ! ptp->nvme_our_sntl) ?
+                         ptp->nvme_status : ptp->io_hdr.device_status);
 }
 
 uint32_t
@@ -682,8 +869,8 @@ get_pt_result(const struct sg_pt_base * vp)
 
     if (NULL == ptp)
         return 0;
-    return ptp->nvme_direct ? ptp->nvme_result :
-                              ptp->io_hdr.device_status;
+    return (ptp->is_nvme && ! ptp->nvme_our_sntl) ?
+                        ptp->nvme_result : ptp->io_hdr.device_status;
 }
 
 int
@@ -694,12 +881,31 @@ get_scsi_pt_sense_len(const struct sg_pt_base * vp)
     return ptp->io_hdr.response_len;
 }
 
+uint8_t *
+get_scsi_pt_sense_buf(const struct sg_pt_base * vp)
+{
+    const struct sg_pt_linux_scsi * ptp = &vp->impl;
+
+    return (uint8_t *)(sg_uintptr_t)ptp->io_hdr.response;
+}
+
 int
 get_scsi_pt_duration_ms(const struct sg_pt_base * vp)
 {
     const struct sg_pt_linux_scsi * ptp = &vp->impl;
 
-    return ptp->io_hdr.duration;
+    return sg_duration_set_nano ? (ptp->io_hdr.duration / 1000) :
+                                  ptp->io_hdr.duration;
+}
+
+/* If not available return 0 otherwise return number of nanoseconds that the
+ * lower layers (and hardware) took to execute the command just completed. */
+uint64_t
+get_pt_duration_ns(const struct sg_pt_base * vp)
+{
+    const struct sg_pt_linux_scsi * ptp = &vp->impl;
+
+    return sg_duration_set_nano ? (uint32_t)ptp->io_hdr.duration : 0;
 }
 
 int
@@ -750,16 +956,13 @@ get_scsi_pt_transport_err_str(const struct sg_pt_base * vp, int max_b_len,
         return b;
     }
     cp += n;
-    driv = ds & SG_LIB_DRIVER_MASK;
-    if (driv < (int)SG_ARRAY_SIZE(linux_driver_bytes))
-        driv_cp = linux_driver_bytes[driv];
-#if 0
-    sugg = (ds & SG_LIB_SUGGEST_MASK) >> 4;
-    if (sugg < SG_ARRAY_SIZE(linux_driver_suggests)
-        sugg_cp = linux_driver_suggests[sugg];
-#endif
-    n = snprintf(cp, m, "Driver_status=0x%02x [%s]\n", ds, driv_cp);
-    m -= n;
+    if (ds) {
+        driv = ds & SG_LIB_DRIVER_MASK;
+        if (driv < (int)SG_ARRAY_SIZE(linux_driver_bytes))
+            driv_cp = linux_driver_bytes[driv];
+        n = snprintf(cp, m, "Driver_status=0x%02x [%s]\n", ds, driv_cp);
+        m -= n;
+    }
     if (m < 1)
         b[max_b_len - 1] = '\0';
     return b;
@@ -859,7 +1062,7 @@ do_scsi_pt_v3(struct sg_pt_linux_scsi * ptp, int fd, int time_secs,
         v3_hdr.sbp = (uint8_t *)(sg_uintptr_t)ptp->io_hdr.response;
         v3_hdr.mx_sb_len = (uint8_t)ptp->io_hdr.max_response_len;
     }
-    v3_hdr.pack_id = (int)ptp->io_hdr.spare_in;
+    v3_hdr.pack_id = (int)ptp->io_hdr.request_extra;
     if (BSG_FLAG_Q_AT_HEAD & ptp->io_hdr.flags)
         v3_hdr.flags |= SG_FLAG_Q_AT_HEAD;      /* favour AT_HEAD */
     else if (BSG_FLAG_Q_AT_TAIL & ptp->io_hdr.flags)
@@ -867,7 +1070,7 @@ do_scsi_pt_v3(struct sg_pt_linux_scsi * ptp, int fd, int time_secs,
 
     if (NULL == v3_hdr.cmdp) {
         if (verbose)
-            pr2ws("No SCSI command (cdb) given\n");
+            pr2ws("No SCSI command (cdb) given [v3]\n");
         return SCSI_PT_DO_BAD_PARAMS;
     }
     /* io_hdr.timeout is in milliseconds, if greater than zero */
@@ -890,13 +1093,34 @@ do_scsi_pt_v3(struct sg_pt_linux_scsi * ptp, int fd, int time_secs,
     return 0;
 }
 
+/* Executes SCSI command using sg v4 interface */
+static int
+do_scsi_pt_v4(struct sg_pt_linux_scsi * ptp, int fd, int time_secs,
+              int verbose)
+{
+    if (0 == ptp->io_hdr.request) {
+        if (verbose)
+            pr2ws("No SCSI command (cdb) given [v4]\n");
+        return SCSI_PT_DO_BAD_PARAMS;
+    }
+    /* io_hdr.timeout is in milliseconds, if greater than zero */
+    ptp->io_hdr.timeout = ((time_secs > 0) ? (time_secs * 1000) : DEF_TIMEOUT);
+    if (ioctl(fd, SG_IO, &ptp->io_hdr) < 0) {
+        ptp->os_err = errno;
+        if (verbose > 1)
+            pr2ws("ioctl(SG_IO v4) failed: %s (errno=%d)\n",
+                  safe_strerror(ptp->os_err), ptp->os_err);
+        return -ptp->os_err;
+    }
+    return 0;
+}
+
 /* Executes SCSI command (or at least forwards it to lower layers).
  * Returns 0 for success, negative numbers are negated 'errno' values from
  * OS system calls. Positive return values are errors from this package. */
 int
 do_scsi_pt(struct sg_pt_base * vp, int fd, int time_secs, int verbose)
 {
-    int err;
     struct sg_pt_linux_scsi * ptp = &vp->impl;
     bool have_checked_for_type = (ptp->dev_fd >= 0);
 
@@ -924,44 +1148,34 @@ do_scsi_pt(struct sg_pt_base * vp, int fd, int time_secs, int verbose)
     } else
         fd = ptp->dev_fd;
     if (! have_checked_for_type) {
-        err = set_pt_file_handle(vp, ptp->dev_fd, verbose);
+        int err = set_pt_file_handle(vp, ptp->dev_fd, verbose);
+
         if (err)
             return -ptp->os_err;
     }
     if (ptp->os_err)
         return -ptp->os_err;
+    if (verbose > 5)
+        pr2ws("%s:  is_nvme=%d, is_sg=%d, is_bsg=%d\n", __func__,
+              (int)ptp->is_nvme, (int)ptp->is_sg, (int)ptp->is_bsg);
     if (ptp->is_nvme)
         return sg_do_nvme_pt(vp, -1, time_secs, verbose);
-    else if (sg_bsg_major <= 0)
+    else if (ptp->is_sg) {
+#ifdef IGNORE_LINUX_SGV4
+        return do_scsi_pt_v3(ptp, fd, time_secs, verbose);
+#else
+        if (ptp->sg_version >= SG_LINUX_SG_VER_V4_BASE)
+            return do_scsi_pt_v4(ptp, fd, time_secs, verbose);
+        else
+            return do_scsi_pt_v3(ptp, fd, time_secs, verbose);
+#endif
+    } else if (sg_bsg_major <= 0)
         return do_scsi_pt_v3(ptp, fd, time_secs, verbose);
     else if (ptp->is_bsg)
-        ; /* drop through to sg v4 implementation */
+        return do_scsi_pt_v4(ptp, fd, time_secs, verbose);
     else
         return do_scsi_pt_v3(ptp, fd, time_secs, verbose);
 
-    if (! ptp->io_hdr.request) {
-        if (verbose)
-            pr2ws("No SCSI command (cdb) given (v4)\n");
-        return SCSI_PT_DO_BAD_PARAMS;
-    }
-    /* io_hdr.timeout is in milliseconds */
-    ptp->io_hdr.timeout = ((time_secs > 0) ? (time_secs * 1000) :
-                                             DEF_TIMEOUT);
-#if 0
-    /* sense buffer already zeroed */
-    if (ptp->io_hdr.response && (ptp->io_hdr.max_response_len > 0)) {
-        void * p;
-
-        p = (void *)(sg_uintptr_t)ptp->io_hdr.response;
-        memset(p, 0, ptp->io_hdr.max_response_len);
-    }
-#endif
-    if (ioctl(fd, SG_IO, &ptp->io_hdr) < 0) {
-        ptp->os_err = errno;
-        if (verbose > 1)
-            pr2ws("ioctl(SG_IO v4) failed: %s (errno=%d)\n",
-                  safe_strerror(ptp->os_err), ptp->os_err);
-        return -ptp->os_err;
-    }
+    pr2ws("%s: Should never reach this point\n", __func__);
     return 0;
 }
