@@ -1,5 +1,5 @@
 /*
-*  Copyright (c) 2012-2018, Kaminario Technologies LTD
+*  Copyright (c) 2012-2021, Kaminario Technologies LTD
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
@@ -22,9 +22,11 @@
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ *
+ * SPDX-License-Identifier: BSD-3-Clause
  *
  * This command performs a SCSI COMPARE AND WRITE. See SBC-3 at
- * http://www.t10.org
+ * https://www.t10.org
  *
  */
 
@@ -54,7 +56,7 @@
 #include "sg_unaligned.h"
 #include "sg_pr2serr.h"
 
-static const char * version_str = "1.22 20180219";
+static const char * version_str = "1.30 20210830";
 
 #define DEF_BLOCK_SIZE 512
 #define DEF_NUM_BLOCKS (1)
@@ -72,6 +74,7 @@ static struct option long_options[] = {
         {"dpo", no_argument, 0, 'd'},
         {"fua", no_argument, 0, 'f'},
         {"fua_nv", no_argument, 0, 'F'},
+        {"fua-nv", no_argument, 0, 'F'},
         {"group", required_argument, 0, 'g'},
         {"grpnum", required_argument, 0, 'g'},
         {"help", no_argument, 0, 'h'},
@@ -99,6 +102,8 @@ struct caw_flags {
 
 struct opts_t {
         bool quiet;
+        bool verbose_given;
+        bool version_given;
         bool wfn_given;
         int numblocks;
         int verbose;
@@ -161,7 +166,11 @@ usage()
                 "                            (2 * NUM * 512) or 1024 when "
                 "NUM is 1\n"
                 "\n"
-                "Performs a SCSI COMPARE AND WRITE operation.\n");
+                "Performs a SCSI COMPARE AND WRITE operation. Sends a double "
+                "size\nbuffer, the first half is used to compare what is at "
+                "LBA for NUM\nblocks. If and only if the comparison is "
+                "equal, then the second\nhalf of the buffer is written to "
+                "LBA for NUM blocks.\n");
 }
 
 static int
@@ -245,11 +254,12 @@ parse_args(int argc, char* argv[], struct opts_t * op)
                         }
                         break;
                 case 'v':
+                        op->verbose_given = true;
                         ++op->verbose;
                         break;
                 case 'V':
-                        pr2serr(ME "version: %s\n", version_str);
-                        exit(0);
+                        op->version_given = true;
+                        break;
                 case 'w':
                         op->flags.wrprotect = sg_get_num(optarg);
                         if (op->flags.wrprotect >> 3) {
@@ -282,6 +292,8 @@ parse_args(int argc, char* argv[], struct opts_t * op)
                         goto out_err;
                 }
         }
+        if (op->version_given && (! op->verbose_given))
+                return 0;
         if (NULL == op->device_name) {
                 pr2serr("missing device name!\n");
                 goto out_err;
@@ -339,7 +351,7 @@ sg_ll_compare_and_write(int sg_fd, uint8_t * buff, int blocks,
                         bool noisy, int verbose)
 {
         bool valid;
-        int k, sense_cat, slen, res, ret;
+        int sense_cat, slen, res, ret;
         uint64_t ull = 0;
         struct sg_pt_base * ptvp;
         uint8_t cawCmd[COMPARE_AND_WRITE_CDB_SIZE];
@@ -360,10 +372,11 @@ sg_ll_compare_and_write(int sg_fd, uint8_t * buff, int blocks,
         set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
         set_scsi_pt_data_out(ptvp, buff, xfer_len);
         if (verbose > 1) {
-                pr2serr("    Compare and write cdb: ");
-                for (k = 0; k < COMPARE_AND_WRITE_CDB_SIZE; ++k)
-                        pr2serr("%02x ", cawCmd[k]);
-                pr2serr("\n");
+                char b[128];
+
+                pr2serr("    Compare and write cdb: %s\n",
+                sg_get_command_str(cawCmd, COMPARE_AND_WRITE_CDB_SIZE, false,
+                                   sizeof(b), b));
         }
         if ((verbose > 2) && (xfer_len > 0)) {
                 pr2serr("    Data-out buffer contents:\n");
@@ -371,11 +384,13 @@ sg_ll_compare_and_write(int sg_fd, uint8_t * buff, int blocks,
         }
         res = do_scsi_pt(ptvp, sg_fd, DEF_TIMEOUT_SECS, verbose);
         ret = sg_cmds_process_resp(ptvp, "COMPARE AND WRITE", res,
-                                   SG_NO_DATA_IN, sense_b, noisy, verbose,
-                                   &sense_cat);
-        if (-1 == ret)
+                                   noisy, verbose, &sense_cat);
+        if (-1 == ret) {
+            if (get_scsi_pt_transport_err(ptvp))
+                ret = SG_LIB_TRANSPORT_ERROR;
+            else
                 ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
-        else if (-2 == ret) {
+        } else if (-2 == ret) {
                 switch (sense_cat) {
                 case SG_LIB_CAT_RECOVERED:
                 case SG_LIB_CAT_NO_SENSE:
@@ -405,6 +420,11 @@ sg_ll_compare_and_write(int sg_fd, uint8_t * buff, int blocks,
                         else
                                 pr2serr("Miscompare reported\n");
                         break;
+                case SG_LIB_CAT_ILLEGAL_REQ:
+                        if (verbose)
+                                sg_print_command_len(cawCmd,
+                                             COMPARE_AND_WRITE_CDB_SIZE);
+                        /* FALL THROUGH */
                 default:
                         ret = sense_cat;
                         break;
@@ -443,12 +463,9 @@ open_dev(const char * outf, int verbose)
 {
         int sg_fd = sg_cmds_open_device(outf, false /* rw */, verbose);
 
-        if (sg_fd < 0) {
+        if ((sg_fd < 0) && verbose)
                 pr2serr(ME "open error: %s: %s\n", outf,
                         safe_strerror(-sg_fd));
-                return -SG_LIB_FILE_ERROR;
-        }
-
         return sg_fd;
 }
 
@@ -457,11 +474,12 @@ int
 main(int argc, char * argv[])
 {
         bool ifn_stdin;
-        int res, half_xlen;
+        int res, half_xlen, vb;
         int infd = -1;
         int wfd = -1;
         int devfd = -1;
         uint8_t * wrkBuff = NULL;
+        uint8_t * free_wrkBuff = NULL;
         struct opts_t * op;
         struct opts_t opts;
 
@@ -473,7 +491,31 @@ main(int argc, char * argv[])
                 goto out;
         }
 
-        if (op->verbose) {
+#ifdef DEBUG
+        pr2serr("In DEBUG mode, ");
+        if (op->verbose_given && op->version_given) {
+                pr2serr("but override: '-vV' given, zero verbose and "
+                        "continue\n");
+                op->verbose_given = false;
+                op->version_given = false;
+                op->verbose = 0;
+        } else if (! op->verbose_given) {
+                pr2serr("set '-vv'\n");
+                op->verbose = 2;
+        } else
+                pr2serr("keep verbose=%d\n", op->verbose);
+#else
+        if (op->verbose_given && op->version_given)
+                pr2serr("Not in DEBUG mode, so '-vV' has no special "
+                        "action\n");
+#endif
+        if (op->version_given) {
+                pr2serr(ME "version: %s\n", version_str);
+                return 0;
+        }
+        vb = op->verbose;
+
+        if (vb) {
                 pr2serr("Running COMPARE AND WRITE command with the "
                         "following options:\n  in=%s ", op->ifn);
                 if (op->wfn_given)
@@ -501,16 +543,17 @@ main(int argc, char * argv[])
                 }
         }
 
-        devfd = open_dev(op->device_name, op->verbose);
+        devfd = open_dev(op->device_name, vb);
         if (devfd < 0) {
-                res = -devfd;
+                res = sg_convert_errno(-devfd);
                 goto out;
         }
 
-        wrkBuff = (uint8_t *)malloc(op->xfer_len);
-        if (0 == wrkBuff) {
+        wrkBuff = (uint8_t *)sg_memalign(op->xfer_len, 0, &free_wrkBuff,
+                                         vb > 3);
+        if (NULL == wrkBuff) {
                 pr2serr("Not enough user memory\n");
-                res = SG_LIB_CAT_OTHER;
+                res = sg_convert_errno(ENOMEM);
                 goto out;
         }
 
@@ -547,9 +590,7 @@ main(int argc, char * argv[])
         }
         res = sg_ll_compare_and_write(devfd, wrkBuff, op->numblocks, op->lba,
                                       op->xfer_len, op->flags, ! op->quiet,
-                                      op->verbose);
-
-out:
+                                      vb);
         if (0 != res) {
                 char b[80];
 
@@ -559,20 +600,24 @@ out:
                 case SG_LIB_FILE_ERROR:
                         break;  /* already reported */
                 default:
-                        sg_get_category_sense_str(res, sizeof(b), b,
-                                                  op->verbose);
+                        sg_get_category_sense_str(res, sizeof(b), b, vb);
                         pr2serr(ME "SCSI COMPARE AND WRITE: %s\n", b);
                         break;
                 }
         }
-
-        if (wrkBuff)
-                free(wrkBuff);
+out:
+        if (free_wrkBuff)
+                free(free_wrkBuff);
         if ((infd >= 0) && (! ifn_stdin))
                 close(infd);
         if (wfd >= 0)
                 close(wfd);
         if (devfd >= 0)
                 close(devfd);
+        if (0 == op->verbose) {
+                if (! sg_if_can2stderr("sg_compare_and_write failed: ", res))
+                        pr2serr("Some error occurred, try again with '-v' "
+                                "or '-vv' for more information\n");
+        }
         return res;
 }
