@@ -3,6 +3,8 @@
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <unistd.h>
@@ -31,7 +33,7 @@
  * DEVICE IDENTIFIER and SET DEVICE IDENTIFIER prior to spc4r07.
  */
 
-static const char * version_str = "1.19 20180219";
+static const char * version_str = "1.23 20180814";
 
 #define ME "sg_ident: "
 
@@ -109,22 +111,26 @@ usage(void)
             "    --verbose|-v    increase verbosity of output\n"
             "    --version|-V    print version string and exit\n\n"
             "Performs a SCSI REPORT (or SET) IDENTIFYING INFORMATION "
-            "command\n");
+            "command. When no\noptions are given then REPORT IDENTIFYING "
+            "INFORMATION is sent and the\nresponse is output in "
+            "hexadecimal with ASCII to the right.\n");
 }
 
 int
 main(int argc, char * argv[])
 {
+    bool ascii = false;
+    bool do_clear = false;
+    bool raw = false;
+    bool do_set = false;
+    bool verbose_given = false;
+    bool version_given = false;
     int sg_fd, res, c, ii_len;
     uint8_t rdi_buff[REPORT_ID_INFO_SANITY_LEN + 4];
     char b[80];
     uint8_t * bp = NULL;
     int itype = 0;
     int verbose = 0;
-    bool ascii = false;
-    bool do_clear = false;
-    bool raw = false;
-    bool do_set = false;
     const char * device_name = NULL;
     int ret = 0;
 
@@ -161,11 +167,12 @@ main(int argc, char * argv[])
             do_set = true;
             break;
         case 'v':
+            verbose_given = true;
             ++verbose;
             break;
         case 'V':
-            pr2serr(ME "version: %s\n", version_str);
-            return 0;
+            version_given = true;
+            break;
         default:
             pr2serr("unrecognised option code 0x%x ??\n", c);
             usage();
@@ -185,30 +192,51 @@ main(int argc, char * argv[])
         }
     }
 
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr(ME "version: %s\n", version_str);
+        return 0;
+    }
+
     if (NULL == device_name) {
-        pr2serr("missing device name!\n");
+        pr2serr("missing device name!\n\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
     if (do_set && do_clear) {
         pr2serr("only one of '--clear' and '--set' can be given\n");
         usage();
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
     if (ascii && raw) {
         pr2serr("only one of '--ascii' and '--raw' can be given\n");
         usage();
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
     if ((do_set || do_clear) && (raw || ascii)) {
         pr2serr("'--set' cannot be used with either '--ascii' or '--raw'\n");
         usage();
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
     sg_fd = sg_cmds_open_device(device_name, false /* rw=false */, verbose);
     if (sg_fd < 0) {
         pr2serr(ME "open error: %s: %s\n", device_name, safe_strerror(-sg_fd));
-        return SG_LIB_FILE_ERROR;
+        return sg_convert_errno(-sg_fd);
     }
 
     memset(rdi_buff, 0x0, sizeof(rdi_buff));
@@ -279,7 +307,12 @@ err_out:
     if (res < 0) {
         pr2serr("close error: %s\n", safe_strerror(-res));
         if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+            ret = sg_convert_errno(-res);
+    }
+    if (0 == verbose) {
+        if (! sg_if_can2stderr("sg_ident failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' or '-vv' for "
+                    "more information\n");
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
