@@ -1,8 +1,10 @@
 /*
- * Copyright (c) 2004-2018 Douglas Gilbert.
+ * Copyright (c) 2004-2021 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <unistd.h>
@@ -11,6 +13,7 @@
 #include <stdlib.h>
 #include <stdbool.h>
 #include <string.h>
+#include <errno.h>
 #include <ctype.h>
 #include <getopt.h>
 #define __STDC_FORMAT_MACROS 1
@@ -31,7 +34,7 @@
  * and decodes the response.
  */
 
-static const char * version_str = "1.37 20180219";
+static const char * version_str = "1.48 20210804";      /* spc6r05 */
 
 #define MAX_RLUNS_BUFF_LEN (1024 * 1024)
 #define DEF_RLUNS_BUFF_LEN (1024 * 8)
@@ -249,6 +252,9 @@ decode_lun(const char * leadin, const uint8_t * lunp, bool lu_cong,
                 case 5:
                     printf("%sMANAGEMENT PROTOCOL %s\n", l_leadin, b);
                     break;
+                case 6:
+                    printf("%sTARGET COMMANDS %s\n", l_leadin, b);
+                    break;
                 default:
                     if (do_hex)
                         printf("%s%s 0x%02x\n", l_leadin, b, x);
@@ -305,9 +311,6 @@ decode_lun(const char * leadin, const uint8_t * lunp, bool lu_cong,
                 }
             }
             break;
-        default:
-            printf("%s<<%s: faulty logic>>\n", l_leadin, __func__);
-            break;
         }
         if (next_level)
             continue;
@@ -367,6 +370,8 @@ main(int argc, char * argv[])
     bool test_linux_out = false;
 #endif
     bool trunc;
+    bool verbose_given = false;
+    bool version_given = false;
     int sg_fd, k, m, off, res, c, list_len, len_cap, luns;
     int decode_arg = 0;
     int do_hex = 0;
@@ -380,6 +385,7 @@ main(int argc, char * argv[])
     const char * device_name = NULL;
     const char * cp;
     uint8_t * reportLunsBuff = NULL;
+    uint8_t * free_reportLunsBuff = NULL;
     uint8_t lun_arr[8];
     struct sg_simple_inquiry_resp sir;
 
@@ -422,6 +428,9 @@ main(int argc, char * argv[])
                 pr2serr("argument to '--maxlen' should be %d or less\n",
                         MAX_RLUNS_BUFF_LEN);
                 return SG_LIB_SYNTAX_ERROR;
+            } else if (maxlen < 4) {
+                pr2serr("Warning: setting '--maxlen' to 4\n");
+                maxlen = 4;
             }
             break;
         case 'q':
@@ -444,11 +453,12 @@ main(int argc, char * argv[])
             test_arg = optarg;
             break;
         case 'v':
+            verbose_given = true;
             ++verbose;
             break;
         case 'V':
-            pr2serr("version: %s\n", version_str);
-            return 0;
+            version_given = true;
+            break;
         default:
             pr2serr("unrecognised option code 0x%x ??\n", c);
             usage();
@@ -467,6 +477,26 @@ main(int argc, char * argv[])
             return SG_LIB_SYNTAX_ERROR;
         }
     }
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr("version: %s\n", version_str);
+        return 0;
+    }
 
     if (test_arg) {
         memset(lun_arr, 0, sizeof(lun_arr));
@@ -476,7 +506,7 @@ main(int argc, char * argv[])
         if ('L' == toupper(cp[0])) {
             uint64_t ull;
 
-            if (('0' == cp[1]) && ('X' == toupper(cp[2])))
+            if (('0' == cp[1]) && ('X' == toupper((uint8_t)cp[2])))
                 k = sscanf(cp + 3, " %" SCNx64, &ull);
             else
                 k = sscanf(cp + 1, " %" SCNu64, &ull);
@@ -572,15 +602,20 @@ main(int argc, char * argv[])
 
     sg_fd = sg_cmds_open_device(device_name, o_readonly, verbose);
     if (sg_fd < 0) {
-        pr2serr("open error: %s: %s\n", device_name, safe_strerror(-sg_fd));
-        return SG_LIB_FILE_ERROR;
+        int err = -sg_fd;
+
+        pr2serr("open error: %s: %s\n", device_name, safe_strerror(err));
+        if ((! o_readonly) && ((err == EACCES) || (err == EROFS)))
+            pr2serr("Perhaps try again with --readonly option or with root "
+                    "permissions\n");
+        return sg_convert_errno(-sg_fd);
     }
     if (decode_arg && (! lu_cong_arg_given)) {
         if (verbose > 1)
             pr2serr("in order to decode LUN and since --lu_cong not given, "
                     "do standard\nINQUIRY to find LU_CONG bit\n");
         /* check if LU_CONG set in standard INQUIRY response */
-        res = sg_simple_inquiry(sg_fd, &sir, 0, verbose);
+        res = sg_simple_inquiry(sg_fd, &sir, false, verbose);
         ret = res;
         if (res) {
             pr2serr("fetching standard INQUIRY response failed\n");
@@ -593,10 +628,11 @@ main(int argc, char * argv[])
 
     if (0 == maxlen)
         maxlen = DEF_RLUNS_BUFF_LEN;
-    reportLunsBuff = (uint8_t *)calloc(1, maxlen);
+    reportLunsBuff = (uint8_t *)sg_memalign(maxlen, 0, &free_reportLunsBuff,
+                                            verbose > 3);
     if (NULL == reportLunsBuff) {
-        pr2serr("unable to malloc %d bytes\n", maxlen);
-        return SG_LIB_CAT_OTHER;
+        pr2serr("unable to sg_memalign %d bytes\n", maxlen);
+        return sg_convert_errno(ENOMEM);
     }
     trunc = false;
 
@@ -669,13 +705,18 @@ main(int argc, char * argv[])
     }
 
 the_end:
-    if (reportLunsBuff)
-        free(reportLunsBuff);
+    if (free_reportLunsBuff)
+        free(free_reportLunsBuff);
     res = sg_cmds_close_device(sg_fd);
     if (res < 0) {
         pr2serr("close error: %s\n", safe_strerror(-res));
         if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+            return sg_convert_errno(-res);
+    }
+    if (0 == verbose) {
+        if (! sg_if_can2stderr("sg_luns failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' or '-vv' for "
+                    "more information\n");
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
