@@ -1,10 +1,12 @@
 /* A utility program originally written for the Linux OS SCSI subsystem.
- *  Copyright (C) 2004-2018 D. Gilbert
+ *  Copyright (C) 2004-2021 D. Gilbert
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 2, or (at your option)
  *  any later version.
  *
+ * SPDX-License-Identifier: GPL-2.0-or-later
+ *
  *  This program issues the SCSI PERSISTENT IN and OUT commands.
  */
 
@@ -33,7 +35,7 @@
 #include "sg_unaligned.h"
 #include "sg_pr2serr.h"
 
-static const char * version_str = "0.62 20180226";
+static const char * version_str = "0.68 20210610";
 
 
 #define PRIN_RKEY_SA     0x0
@@ -66,14 +68,16 @@ struct opts_t {
     bool readonly;
     bool readwrite_force;/* set when '-yy' given. Ooverrides environment
                             variable SG_PERSIST_IN_RDONLY and opens RW */
+    bool verbose_given;
+    bool version_given;
     int hex;
     int num_transportids;
     int prin_sa;
     int prout_sa;
     int verbose;
-    unsigned int alloc_len;
-    unsigned int param_rtp;
-    unsigned int prout_type;
+    uint32_t alloc_len;
+    uint32_t param_rtp;
+    uint32_t prout_type;
     uint64_t param_rk;
     uint64_t param_sark;
     uint8_t transportid_arr[MX_TIDS * MX_TID_LEN];
@@ -147,8 +151,7 @@ static const char * prin_sa_strs[] = {
     "[reserved 0x6]",
     "[reserved 0x7]",
 };
-static const int num_prin_sa_strs = sizeof(prin_sa_strs) /
-                                    sizeof(prin_sa_strs[0]);
+static const int num_prin_sa_strs = SG_ARRAY_SIZE(prin_sa_strs);
 
 static const char * prout_sa_strs[] = {
     "Register",
@@ -162,8 +165,7 @@ static const char * prout_sa_strs[] = {
     "Replace lost reservation",
     "[reserved 0x9]",
 };
-static const int num_prout_sa_strs = sizeof(prout_sa_strs) /
-                                     sizeof(prout_sa_strs[0]);
+static const int num_prout_sa_strs = SG_ARRAY_SIZE(prout_sa_strs);
 
 static const char * pr_type_strs[] = {
     "obsolete [0]",
@@ -278,13 +280,12 @@ prin_work(int sg_fd, const struct opts_t * op)
     uint8_t * bp;
     uint8_t * pr_buff = NULL;
     uint8_t * free_pr_buff = NULL;
-    int pr_buff_sz = MX_ALLOC_LEN;
 
-    pr_buff = sg_memalign(pr_buff_sz, 0 /* page aligned */, &free_pr_buff,
+    pr_buff = sg_memalign(op->alloc_len, 0 /* page aligned */, &free_pr_buff,
                           false);
     if (NULL == pr_buff) {
         pr2serr("%s: unable to allocate %d bytes on heap\n", __func__,
-                pr_buff_sz);
+                op->alloc_len);
         return sg_convert_errno(ENOMEM);
     }
     res = sg_ll_persistent_reserve_in(sg_fd, op->prin_sa, pr_buff,
@@ -321,6 +322,8 @@ prin_work(int sg_fd, const struct opts_t * op)
             hex2stdout(pr_buff, 8, 1);
         else {
             printf("Report capabilities response:\n");
+            printf("  Replace Lost Reservation Capable(RLR_C): %d\n",
+                   !!(pr_buff[2] & 0x80));      /* added spc4r26 */
             printf("  Compatible Reservation Handling(CRH): %d\n",
                    !!(pr_buff[2] & 0x10));
             printf("  Specify Initiator Ports Capable(SIP_C): %d\n",
@@ -336,17 +339,17 @@ prin_work(int sg_fd, const struct opts_t * op)
             if (pr_buff[3] & 0x80) {
                 printf("    Support indicated in Type mask:\n");
                 printf("      %s: %d\n", pr_type_strs[7],
-                       !!(pr_buff[4] & 0x80));
+                       !!(pr_buff[4] & 0x80));  /* WR_EX_AR */
                 printf("      %s: %d\n", pr_type_strs[6],
-                       !!(pr_buff[4] & 0x40));
+                       !!(pr_buff[4] & 0x40));  /* EX_AC_RO */
                 printf("      %s: %d\n", pr_type_strs[5],
-                       !!(pr_buff[4] & 0x20));
+                       !!(pr_buff[4] & 0x20));  /* WR_EX_RO */
                 printf("      %s: %d\n", pr_type_strs[3],
-                       !!(pr_buff[4] & 0x8));
+                       !!(pr_buff[4] & 0x8));   /* EX_AC */
                 printf("      %s: %d\n", pr_type_strs[1],
-                       !!(pr_buff[4] & 0x2));
+                       !!(pr_buff[4] & 0x2));   /* WR_EX */
                 printf("      %s: %d\n", pr_type_strs[8],
-                       !!(pr_buff[5] & 0x1));
+                       !!(pr_buff[5] & 0x1));   /* EX_AC_AR */
             }
         }
     } else {
@@ -359,10 +362,10 @@ prin_work(int sg_fd, const struct opts_t * op)
                 printf("  PR generation=0x%x, ", pr_gen);
                 if (add_len <= 0)
                     printf("Additional length=%d\n", add_len);
-                if (add_len > (pr_buff_sz - 8)) {
+                if ((uint32_t)add_len > (op->alloc_len - 8)) {
                     printf("Additional length too large=%d, truncate\n",
                            add_len);
-                    hex2stdout((pr_buff + 8), pr_buff_sz - 8, 1);
+                    hex2stdout((pr_buff + 8), op->alloc_len - 8, 1);
                 } else {
                     printf("Additional length=%d\n", add_len);
                     hex2stdout((pr_buff + 8), add_len, 1);
@@ -479,16 +482,15 @@ prout_work(int sg_fd, struct opts_t * op)
     int res = 0;
     uint8_t * pr_buff = NULL;
     uint8_t * free_pr_buff = NULL;
-    int pr_buff_sz = MX_ALLOC_LEN;
     char b[64];
     char bb[80];
 
     t_arr_len = compact_transportid_array(op);
-    pr_buff = sg_memalign(pr_buff_sz, 0 /* page aligned */, &free_pr_buff,
+    pr_buff = sg_memalign(op->alloc_len, 0 /* page aligned */, &free_pr_buff,
                           false);
     if (NULL == pr_buff) {
         pr2serr("%s: unable to allocate %d bytes on heap\n", __func__,
-                pr_buff_sz);
+                op->alloc_len);
         return sg_convert_errno(ENOMEM);
     }
     sg_put_unaligned_be64(op->param_rk, pr_buff + 0);
@@ -539,14 +541,13 @@ prout_reg_move_work(int sg_fd, struct opts_t * op)
     int res = 0;
     uint8_t * pr_buff = NULL;
     uint8_t * free_pr_buff = NULL;
-    int pr_buff_sz = MX_ALLOC_LEN;
 
     t_arr_len = compact_transportid_array(op);
-    pr_buff = sg_memalign(pr_buff_sz, 0 /* page aligned */, &free_pr_buff,
+    pr_buff = sg_memalign(op->alloc_len, 0 /* page aligned */, &free_pr_buff,
                           false);
     if (NULL == pr_buff) {
         pr2serr("%s: unable to allocate %d bytes on heap\n", __func__,
-                pr_buff_sz);
+                op->alloc_len);
         return sg_convert_errno(ENOMEM);
     }
     sg_put_unaligned_be64(op->param_rk, pr_buff + 0);
@@ -778,7 +779,7 @@ decode_file_tids(const char * fnp, struct opts_t * op)
             continue;
         }
         if (carry_over[0]) {
-            if (isxdigit(line[0])) {
+            if (isxdigit((uint8_t)line[0])) {
                 carry_over[1] = line[0];
                 carry_over[2] = '\0';
                 if (1 == sscanf(carry_over, "%x", &h))
@@ -825,7 +826,7 @@ decode_file_tids(const char * fnp, struct opts_t * op)
                     pr2serr("%s: array length exceeded\n", __func__);
                     goto bad;
                 }
-                tid_arr[off + k] = h;
+                op->transportid_arr[off + k] = h;/* keep code checker happy */
                 lcp = strpbrk(lcp, " ,\t");
                 if (NULL == lcp)
                     break;
@@ -898,14 +899,14 @@ build_transportid(const char * inp, struct opts_t * op)
             goto my_cont_b;
         k = strspn(inp, "0123456789aAbBcCdDeEfF, ");
         if (in_len != k) {
-            pr2serr("build_transportid: error at pos %d\n", k + 1);
+            pr2serr("%s: error at pos %d\n", __func__, k + 1);
             return 1;
         }
         for (k = 0; k < (int)sizeof(op->transportid_arr); ++k) {
             if (1 == sscanf(lcp, "%x", &h)) {
                 if (h > 0xff) {
-                    pr2serr("build_transportid: hex number larger than 0xff "
-                            "at pos %d\n", (int)(lcp - inp + 1));
+                    pr2serr("%s: hex number larger than 0xff at pos %d\n",
+                            __func__, (int)(lcp - inp + 1));
                     return 1;
                 }
                 tid_arr[k] = h;
@@ -919,7 +920,7 @@ build_transportid(const char * inp, struct opts_t * op)
                     cp = c2p;
                 lcp = cp + 1;
             } else {
-                pr2serr("build_transportid: error at pos %d\n",
+                pr2serr("%s: error at pos %d\n", __func__,
                         (int)(lcp - inp + 1));
                 return 1;
             }
@@ -927,7 +928,7 @@ build_transportid(const char * inp, struct opts_t * op)
 my_cont_b:
         op->num_transportids = 1;
         if (k >= (int)sizeof(op->transportid_arr)) {
-            pr2serr("build_transportid: array length exceeded\n");
+            pr2serr("%s: array length exceeded\n", __func__);
             return 1;
         }
     }
@@ -1097,11 +1098,12 @@ main(int argc, char * argv[])
             op->param_unreg = true;
             break;
         case 'v':
+            op->verbose_given = true;
             ++op->verbose;
             break;
         case 'V':
-            pr2serr("version: %s\n", version_str);
-            return 0;
+            op->version_given = true;
+            break;
         case 'X':
             if (0 != build_transportid(optarg, op)) {
                 pr2serr("bad argument to '--transport-id'\n");
@@ -1155,6 +1157,27 @@ main(int argc, char * argv[])
         usage(help);
         return 0;
     }
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (op->verbose_given && op->version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and "
+                "continue\n");
+        op->verbose_given = false;
+        op->version_given = false;
+        op->verbose = 0;
+    } else if (! op->verbose_given) {
+        pr2serr("set '-vv'\n");
+        op->verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", op->verbose);
+#else
+    if (op->verbose_given && op->version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (op->version_given) {
+        pr2serr("version: %s\n", version_str);
+        return 0;
+    }
 
     if (NULL == device_name) {
         pr2serr("No device name given\n");
@@ -1164,21 +1187,21 @@ main(int argc, char * argv[])
     if (want_prout && want_prin) {
         pr2serr("choose '--in' _or_ '--out' (not both)\n");
         usage(1);
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     } else if (want_prout) { /* syntax check on PROUT arguments */
         op->pr_in = false;
         if ((1 != num_prout_sa) || (0 != num_prin_sa)) {
             pr2serr(">> For Persistent Reserve Out one and only one "
                     "appropriate\n>> service action must be chosen (e.g. "
                     "'--register')\n");
-            return SG_LIB_SYNTAX_ERROR;
+            return SG_LIB_CONTRADICT;
         }
     } else { /* syntax check on PRIN arguments */
         if (num_prout_sa > 0) {
             pr2serr(">> When a service action for Persistent Reserve Out "
                     "is chosen the\n>> '--out' option must be given (as a "
                     "safeguard)\n");
-            return SG_LIB_SYNTAX_ERROR;
+            return SG_LIB_CONTRADICT;
         }
         if (0 == num_prin_sa) {
             pr2serr(">> No service action given; assume Persistent Reserve "
@@ -1188,7 +1211,7 @@ main(int argc, char * argv[])
         } else if (num_prin_sa > 1)  {
             pr2serr("Too many service actions given; choose one only\n");
             usage(1);
-            return SG_LIB_SYNTAX_ERROR;
+            return SG_LIB_CONTRADICT;
         }
     }
     if ((op->param_unreg || op->param_rtp) &&
@@ -1196,14 +1219,14 @@ main(int argc, char * argv[])
         pr2serr("--unreg or --relative-target-port only useful with "
                 "--register-move\n");
         usage(1);
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
     if ((PROUT_REG_MOVE_SA == op->prout_sa) &&
         (1 != op->num_transportids)) {
         pr2serr("with --register-move one (and only one) --transport-id "
                 "should be given\n");
         usage(1);
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
     if (((PROUT_RES_SA == op->prout_sa) ||
          (PROUT_REL_SA == op->prout_sa) ||
@@ -1284,7 +1307,7 @@ main(int argc, char * argv[])
 
 fini:
     if (ret && (0 == op->verbose) && (! flagged)) {
-        if (! sg_if_can2stderr("", ret))
+        if (! sg_if_can2stderr("sg_persist failed: ", ret))
             pr2serr("Some error occurred [%d]\n", ret);
     }
     if (sg_fd >= 0) {
@@ -1292,7 +1315,7 @@ fini:
         if (res < 0) {
             pr2serr("close error: %s\n", safe_strerror(-res));
             if (0 == ret)
-                ret = SG_LIB_FILE_ERROR;
+                ret = sg_convert_errno(-res);
         }
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
