@@ -1,13 +1,15 @@
 /*
  * A utility program originally written for the Linux OS SCSI subsystem.
  *
- * Copyright (C) 2000-2018 Ingo van Lil <inguin@gmx.de>
+ * Copyright (C) 2000-2021 Ingo van Lil <inguin@gmx.de>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2, or (at your option)
  * any later version.
  *
+ * SPDX-License-Identifier: GPL-2.0-or-later
+ *
  * This program can be used to send raw SCSI commands (with an optional
  * data phase) through a Generic SCSI interface.
  */
@@ -37,12 +39,12 @@
 #include "sg_pr2serr.h"
 #include "sg_unaligned.h"
 
-#define SG_RAW_VERSION "0.4.24 (2018-03-02)"
+#define SG_RAW_VERSION "0.4.37 (2021-06-01)"
 
 #define DEFAULT_TIMEOUT 20
 #define MIN_SCSI_CDBSZ 6
 #define MAX_SCSI_CDBSZ 260
-#define MAX_SCSI_DXLEN (64 * 1024)
+#define MAX_SCSI_DXLEN (1024 * 1024)
 
 #define NVME_ADDR_DATA_IN  0xfffffffffffffffe
 #define NVME_ADDR_DATA_OUT 0xfffffffffffffffd
@@ -52,15 +54,18 @@
 static struct option long_options[] = {
     { "binary",  no_argument,       NULL, 'b' },
     { "cmdfile", required_argument, NULL, 'c' },
+    { "cmdset",  required_argument, NULL, 'C' },
     { "enumerate", no_argument,     NULL, 'e' },
     { "help",    no_argument,       NULL, 'h' },
     { "infile",  required_argument, NULL, 'i' },
     { "skip",    required_argument, NULL, 'k' },
     { "nosense", no_argument,       NULL, 'n' },
+    { "nvm",     no_argument,       NULL, 'N' },
     { "outfile", required_argument, NULL, 'o' },
     { "raw",     no_argument,       NULL, 'w' },
     { "request", required_argument, NULL, 'r' },
     { "readonly", no_argument,      NULL, 'R' },
+    { "scan",    required_argument, NULL, 'Q' },
     { "send",    required_argument, NULL, 's' },
     { "timeout", required_argument, NULL, 't' },
     { "verbose", no_argument,       NULL, 'v' },
@@ -75,14 +80,19 @@ struct opts_t {
     bool do_dataout;
     bool do_enumerate;
     bool no_sense;
+    bool do_nvm;     /* the NVMe command set: NVM containing its READ+WRITE */
     bool do_help;
-    bool do_version;
+    bool verbose_given;
+    bool version_given;
     int cdb_length;
+    int cmdset;
     int datain_len;
     int dataout_len;
     int timeout;
     int raw;
     int readonly;
+    int scan_first;
+    int scan_last;
     int verbose;
     off_t dataout_offset;
     uint8_t cdb[MAX_SCSI_CDBSZ];        /* might be NVMe command (64 byte) */
@@ -97,11 +107,11 @@ static void
 pr_version()
 {
     pr2serr("sg_raw " SG_RAW_VERSION "\n"
-            "Copyright (C) 2007-2018 Ingo van Lil <inguin@gmx.de>\n"
+            "Copyright (C) 2007-2021 Ingo van Lil <inguin@gmx.de>\n"
             "This is free software.  You may redistribute copies of it "
             "under the terms of\n"
             "the GNU General Public License "
-            "<http://www.gnu.org/licenses/gpl.html>.\n"
+            "<https://www.gnu.org/licenses/gpl.html>.\n"
             "There is NO WARRANTY, to the extent permitted by law.\n");
 }
 
@@ -116,17 +126,25 @@ usage()
             "                         stdout\n"
             "  --cmdfile=CF|-c CF     CF is file containing command in hex "
             "bytes\n"
+            "  --cmdset=CS|-C CS      CS is 0 (def) heuristic chooses "
+            "command set;\n"
+            "                         1: force SCSI; 2: force NVMe\n"
             "  --enumerate|-e         Decodes cdb name then exits; requires "
             "DEVICE but\n"
             "                         ignores it\n"
             "  --help|-h              Show this message and exit\n"
-            "  --infile=IFILE|-i IFILE    Read data to send from IFILE "
-            "(default:\n"
-            "                             stdin)\n"
+            "  --infile=IFILE|-i IFILE    Read binary data to send (i.e. "
+            "data-out)\n"
+            "                             from IFILE (default: stdin)\n"
             "  --nosense|-n           Don't display sense information\n"
-            "  --outfile=OFILE|-o OFILE    Write binary data to OFILE (def: "
-            "hexdump\n"
-            "                              to stdout)\n"
+            "  --nvm|-N               command is for NVM command set (e.g. "
+            "Read);\n"
+            "                         default, if NVMe fd, Admin command "
+            "set\n"
+            "  --outfile=OFILE|-o OFILE    Write binary data from device "
+            "(i.e. data-in)\n"
+            "                              to OFILE (def: hexdump to "
+            "stdout)\n"
             "  --raw|-w               interpret CF (command file) as "
             "binary (def:\n"
             "                         interpret as ASCII hex)\n"
@@ -134,11 +152,16 @@ usage()
             "read-write)\n"
             "  --request=RLEN|-r RLEN    Request up to RLEN bytes of data "
             "(data-in)\n"
+            "  --scan=FO,LO|-Q FO,LO    scan command set from FO (first "
+            "opcode)\n"
+            "                           to LO (last opcode) inclusive. Uses "
+            "given\n"
+            "                           command bytes, varying the opcode\n"
             "  --send=SLEN|-s SLEN    Send SLEN bytes of data (data-out)\n"
             "  --skip=KLEN|-k KLEN    Skip the first KLEN bytes when "
             "reading\n"
             "                         data to send (default: 0)\n"
-            "  --timeout=SEC|-t SEC    Timeout in seconds (default: 20)\n"
+            "  --timeout=SECS|-t SECS    Timeout in seconds (default: 20)\n"
             "  --verbose|-v           Increase verbosity\n"
             "  --version|-V           Show version information and exit\n"
             "\n"
@@ -149,207 +172,15 @@ usage()
             "  sg_raw -r 1k /dev/sg0 12 00 00 00 60 00\n");
 }
 
-/* Read ASCII hex bytes or binary from fname (a file named '-' taken as
- * stdin). If reading ASCII hex then there should be either one entry per
- * line or a comma, space or tab separated list of bytes. If no_space is
- * true then a string of ACSII hex digits is expected, 2 per byte. Everything
- * from and including a '#' on a line is ignored. Returns true if ok, or
- * false if error. */
-static bool
-f2hex_arr(const char * fname, bool as_binary, bool no_space,
-          uint8_t * mp_arr, int * mp_arr_len, int max_arr_len)
-{
-    int fn_len, in_len, k, j, m, fd;
-    bool has_stdin, split_line;
-    unsigned int h;
-    const char * lcp;
-    FILE * fp;
-    char line[512];
-    char carry_over[4];
-    int off = 0;
-    struct stat a_stat;
-
-    if ((NULL == fname) || (NULL == mp_arr) || (NULL == mp_arr_len))
-        return false;
-    fn_len = strlen(fname);
-    if (0 == fn_len)
-        return false;
-    has_stdin = ((1 == fn_len) && ('-' == fname[0]));   /* read from stdin */
-    if (as_binary) {
-        if (has_stdin)
-            fd = STDIN_FILENO;
-        else {
-            fd = open(fname, O_RDONLY);
-            if (fd < 0) {
-                pr2serr("unable to open binary file %s: %s\n", fname,
-                         safe_strerror(errno));
-                return false;
-            }
-        }
-        k = read(fd, mp_arr, max_arr_len);
-        if (k <= 0) {
-            if (0 == k)
-                pr2serr("read 0 bytes from binary file %s\n", fname);
-            else
-                pr2serr("read from binary file %s: %s\n", fname,
-                        safe_strerror(errno));
-            if (! has_stdin)
-                close(fd);
-            return false;
-        }
-        if ((0 == fstat(fd, &a_stat)) && S_ISFIFO(a_stat.st_mode)) {
-            /* pipe; keep reading till error or 0 read */
-            while (k < max_arr_len) {
-                m = read(fd, mp_arr + k, max_arr_len - k);
-                if (0 == m)
-                   break;
-                if (m < 0) {
-                    pr2serr("read from binary pipe %s: %s\n", fname,
-                            safe_strerror(errno));
-                    if (! has_stdin)
-                        close(fd);
-                    return false;
-                }
-                k += m;
-            }
-        }
-        *mp_arr_len = k;
-        if (! has_stdin)
-            close(fd);
-        return true;
-    } else {    /* So read the file as ASCII hex */
-        if (has_stdin)
-            fp = stdin;
-        else {
-            fp = fopen(fname, "r");
-            if (NULL == fp) {
-                pr2serr("Unable to open %s for reading\n", fname);
-                return false;
-            }
-        }
-    }
-
-    carry_over[0] = 0;
-    for (j = 0; j < 512; ++j) {
-        if (NULL == fgets(line, sizeof(line), fp))
-            break;
-        in_len = strlen(line);
-        if (in_len > 0) {
-            if ('\n' == line[in_len - 1]) {
-                --in_len;
-                line[in_len] = '\0';
-                split_line = false;
-            } else
-                split_line = true;
-        }
-        if (in_len < 1) {
-            carry_over[0] = 0;
-            continue;
-        }
-        if (carry_over[0]) {
-            if (isxdigit(line[0])) {
-                carry_over[1] = line[0];
-                carry_over[2] = '\0';
-                if (1 == sscanf(carry_over, "%4x", &h))
-                    mp_arr[off - 1] = h;       /* back up and overwrite */
-                else {
-                    pr2serr("%s: carry_over error ['%s'] around line %d\n",
-                            __func__, carry_over, j + 1);
-                    goto bad;
-                }
-                lcp = line + 1;
-                --in_len;
-            } else
-                lcp = line;
-            carry_over[0] = 0;
-        } else
-            lcp = line;
-
-        m = strspn(lcp, " \t");
-        if (m == in_len)
-            continue;
-        lcp += m;
-        in_len -= m;
-        if ('#' == *lcp)
-            continue;
-        k = strspn(lcp, "0123456789aAbBcCdDeEfF ,\t");
-        if ((k < in_len) && ('#' != lcp[k]) && ('\r' != lcp[k])) {
-            pr2serr("%s: syntax error at line %d, pos %d\n", __func__,
-                    j + 1, m + k + 1);
-            goto bad;
-        }
-        if (no_space) {
-            for (k = 0; isxdigit(*lcp) && isxdigit(*(lcp + 1));
-                 ++k, lcp += 2) {
-                if (1 != sscanf(lcp, "%2x", &h)) {
-                    pr2serr("%s: bad hex number in line %d, pos %d\n",
-                            __func__, j + 1, (int)(lcp - line + 1));
-                    goto bad;
-                }
-                if ((off + k) >= max_arr_len) {
-                    pr2serr("%s: array length exceeded\n", __func__);
-                    goto bad;
-                }
-                mp_arr[off + k] = h;
-            }
-            if (isxdigit(*lcp) && (! isxdigit(*(lcp + 1))))
-                carry_over[0] = *lcp;
-            off += k;
-        } else {
-            for (k = 0; k < 1024; ++k) {
-                if (1 == sscanf(lcp, "%10x", &h)) {
-                    if (h > 0xff) {
-                        pr2serr("%s: hex number larger than 0xff in line "
-                                "%d, pos %d\n", __func__, j + 1,
-                                (int)(lcp - line + 1));
-                        goto bad;
-                    }
-                    if (split_line && (1 == strlen(lcp))) {
-                        /* single trailing hex digit might be a split pair */
-                        carry_over[0] = *lcp;
-                    }
-                    if ((off + k) >= max_arr_len) {
-                        pr2serr("%s: array length exceeded\n", __func__);
-                        goto bad;
-                    }
-                    mp_arr[off + k] = h;
-                    lcp = strpbrk(lcp, " ,\t");
-                    if (NULL == lcp)
-                        break;
-                    lcp += strspn(lcp, " ,\t");
-                    if ('\0' == *lcp)
-                        break;
-                } else {
-                    if (('#' == *lcp) || ('\r' == *lcp)) {
-                        --k;
-                        break;
-                    }
-                    pr2serr("%s: error in line %d, at pos %d\n", __func__,
-                            j + 1, (int)(lcp - line + 1));
-                    goto bad;
-                }
-            }
-            off += (k + 1);
-        }
-    }
-    *mp_arr_len = off;
-    if (stdin != fp)
-        fclose(fp);
-    return true;
-bad:
-    if (stdin != fp)
-        fclose(fp);
-    return false;
-}
-
 static int
 parse_cmd_line(struct opts_t * op, int argc, char *argv[])
 {
     while (1) {
         int c, n;
+        const char * cp;
 
-        c = getopt_long(argc, argv, "bc:ehi:k:no:r:Rs:t:vVw", long_options,
-                        NULL);
+        c = getopt_long(argc, argv, "bc:C:ehi:k:nNo:Q:r:Rs:t:vVw",
+                        long_options, NULL);
         if (c == -1)
             break;
 
@@ -361,6 +192,14 @@ parse_cmd_line(struct opts_t * op, int argc, char *argv[])
             op->cmd_file = optarg;
             op->cmdfile_given = true;
             break;
+        case 'C':
+            n = sg_get_num(optarg);
+            if ((n < 0) || (n > 2)) {
+                pr2serr("Invalid argument to --cmdset= expect 0, 1 or 2\n");
+                return SG_LIB_SYNTAX_ERROR;
+            }
+            op->cmdset = n;
+            break;
         case 'e':
             op->do_enumerate = true;
             break;
@@ -371,7 +210,7 @@ parse_cmd_line(struct opts_t * op, int argc, char *argv[])
         case 'i':
             if (op->dataout_file) {
                 pr2serr("Too many '--infile=' options\n");
-                return SG_LIB_SYNTAX_ERROR;
+                return SG_LIB_CONTRADICT;
             }
             op->dataout_file = optarg;
             break;
@@ -386,13 +225,37 @@ parse_cmd_line(struct opts_t * op, int argc, char *argv[])
         case 'n':
             op->no_sense = true;
             break;
+        case 'N':
+            op->do_nvm = true;
+            break;
         case 'o':
             if (op->datain_file) {
                 pr2serr("Too many '--outfile=' options\n");
-                return SG_LIB_SYNTAX_ERROR;
+                return SG_LIB_CONTRADICT;
             }
             op->datain_file = optarg;
             break;
+        case 'Q':       /* --scan=FO,LO */
+            cp = strchr(optarg, ',');
+            if (NULL == cp) {
+                pr2serr("--scan= expects two numbers, comma separated\n");
+                return SG_LIB_SYNTAX_ERROR;
+            }
+            n = sg_get_num(optarg);
+            if ((n < 0) || (n > 255)) {
+                pr2serr("Invalid first number to --scan= expect 0 to 255\n");
+                return SG_LIB_SYNTAX_ERROR;
+            }
+            op->scan_first = n;
+            n = sg_get_num(cp + 1);
+            if ((n < 0) || (n > 255)) {
+                pr2serr("Invalid second number to --scan= expect 0 to 255\n");
+                return SG_LIB_SYNTAX_ERROR;
+            }
+            op->scan_last = n;
+            if (op->scan_first >= n)
+                pr2serr("Warning: scan range degenerate, ignore\n");
+            break;
         case 'r':
             op->do_datain = true;
             n = sg_get_num(optarg);
@@ -423,10 +286,11 @@ parse_cmd_line(struct opts_t * op, int argc, char *argv[])
             op->timeout = n;
             break;
         case 'v':
+            op->verbose_given = true;
             ++op->verbose;
             break;
         case 'V':
-            op->do_version = true;
+            op->version_given = true;
             break;
         case 'w':       /* -r and -R already in use, this is --raw */
             ++op->raw;
@@ -436,6 +300,13 @@ parse_cmd_line(struct opts_t * op, int argc, char *argv[])
         }
     }
 
+    if (op->version_given
+#ifdef DEBUG
+        && ! op->verbose_given
+#endif
+       )
+        return 0;
+
     if (optind >= argc) {
         pr2serr("No device specified\n");
         return SG_LIB_SYNTAX_ERROR;
@@ -461,13 +332,16 @@ parse_cmd_line(struct opts_t * op, int argc, char *argv[])
     }
 
     if (op->cmdfile_given) {
-        bool ok;
+        int err;
 
-        ok = f2hex_arr(op->cmd_file, (op->raw > 0) /* as_binary */,
-                       false /* no_space */, op->cdb, &op->cdb_length,
-                       MAX_SCSI_CDBSZ);
-        if (! ok)
+        err = sg_f2hex_arr(op->cmd_file, (op->raw > 0) /* as_binary */,
+                           false /* no_space */, op->cdb, &op->cdb_length,
+                           MAX_SCSI_CDBSZ);
+        if (err) {
+            pr2serr("Unable to parse: %s as %s\n", op->cmd_file,
+                    (op->raw > 0) ? "binary" : "hex");
             return SG_LIB_SYNTAX_ERROR;
+        }
         if (op->verbose > 2) {
             pr2serr("Read %d from %s . They are in hex:\n", op->cdb_length,
                     op->cmd_file);
@@ -483,6 +357,16 @@ parse_cmd_line(struct opts_t * op, int argc, char *argv[])
         int sa;
         char b[80];
 
+        if ((1 == op->cmdset) && !is_scsi_cdb) {
+            is_scsi_cdb = true;
+            if (op->verbose > 3)
+                printf(">>> overriding cmdset guess to SCSI\n");
+        }
+        if ((2 == op->cmdset) && is_scsi_cdb) {
+            is_scsi_cdb = false;
+            if (op->verbose > 3)
+                printf(">>> overriding cmdset guess to NVMe\n");
+        }
         if (is_scsi_cdb) {
             if (op->cdb_length > 16) {
                 sa = sg_get_unaligned_be16(op->cdb + 8);
@@ -496,7 +380,7 @@ parse_cmd_line(struct opts_t * op, int argc, char *argv[])
             printf("Attempt to decode cdb name: %s\n", b);
         } else
             printf(">>> Seems to be NVMe %s command\n",
-                   sg_get_nvme_opcode_name(op->cdb[0], true /* admin */,
+                   sg_get_nvme_opcode_name(op->cdb[0], ! op->do_nvm,
                                            sizeof(b), b));
     }
     return 0;
@@ -536,7 +420,7 @@ static uint8_t *
 fetch_dataout(struct opts_t * op, uint8_t ** free_buf, int * errp)
 {
     bool ok = false;
-    int fd, len, err;
+    int fd, len, tot_len, boff, err;
     uint8_t *buf = NULL;
 
     *free_buf = NULL;
@@ -570,28 +454,29 @@ fetch_dataout(struct opts_t * op, uint8_t ** free_buf, int * errp)
         }
     }
 
-    buf = sg_memalign(op->dataout_len, 0 /* page_size */, free_buf,
-                      op->verbose > 3);
+    tot_len = op->dataout_len;
+    buf = sg_memalign(tot_len, 0 /* page_size */, free_buf, false);
     if (buf == NULL) {
-        pr2serr("sg_memalign: failed to get %d bytes of memory\n",
-                op->dataout_len);
+        pr2serr("sg_memalign: failed to get %d bytes of memory\n", tot_len);
         if (errp)
-            *errp = SG_LIB_OS_BASE_ERR + ENOMEM;
+            *errp = sg_convert_errno(ENOMEM);
         goto bail;
     }
 
-    len = read(fd, buf, op->dataout_len);
-    if (len < 0) {
-        err = errno;
-        if (errp)
-            *errp = sg_convert_errno(err);
-        perror("Failed to read input data");
-        goto bail;
-    } else if (len < op->dataout_len) {
-        if (errp)
-            *errp = SG_LIB_FILE_ERROR;
-        pr2serr("EOF on input file/stream\n");
-        goto bail;
+    for (boff = 0; boff < tot_len; boff += len) {
+        len = read(fd, buf + boff , tot_len - boff);
+        if (len < 0) {
+            err = errno;
+            if (errp)
+                *errp = sg_convert_errno(err);
+            perror("Failed to read input data");
+            goto bail;
+        } else if (0 == len) {
+            if (errp)
+                *errp = SG_LIB_FILE_ERROR;
+            pr2serr("EOF on input file/stream at buffer offset %d\n", boff);
+            goto bail;
+        }
     }
     ok = true;
 
@@ -620,6 +505,7 @@ write_dataout(const char *filename, uint8_t *buf, int len)
     else {
         fd = creat(filename, 0666);
         if (fd < 0) {
+            ret = sg_convert_errno(errno);
             perror(filename);
             goto bail;
         }
@@ -630,6 +516,7 @@ write_dataout(const char *filename, uint8_t *buf, int len)
     }
 
     if (write(fd, buf, len) != len) {
+        ret = sg_convert_errno(errno);
         perror(filename ? filename : "stdout");
         goto bail;
     }
@@ -647,6 +534,7 @@ int
 main(int argc, char *argv[])
 {
     bool is_scsi_cdb = true;
+    bool do_scan = false;
     int ret = 0;
     int err = 0;
     int res_cat, status, s_len, k, ret2;
@@ -668,15 +556,34 @@ main(int argc, char *argv[])
     memset(op, 0, sizeof(opts));
     op->timeout = DEFAULT_TIMEOUT;
     ret = parse_cmd_line(op, argc, argv);
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (op->verbose_given && op->version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        op->verbose_given = false;
+        op->version_given = false;
+        op->verbose = 0;
+    } else if (! op->verbose_given) {
+        pr2serr("set '-vv'\n");
+        op->verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", op->verbose);
+#else
+    if (op->verbose_given && op->version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (op->version_given) {
+        pr_version();
+        goto done;
+    }
+
     if (ret != 0) {
+        pr2serr("\n");  /* blank line before outputting usage */
         usage();
         goto done;
     } else if (op->do_help) {
         usage();
         goto done;
-    } else if (op->do_version) {
-        pr_version();
-        goto done;
     } else if (op->do_enumerate)
         goto done;
 
@@ -684,18 +591,35 @@ main(int argc, char *argv[])
                                 op->verbose);
     if (sg_fd < 0) {
         pr2serr("%s: %s\n", op->device_name, safe_strerror(-sg_fd));
-        ret = SG_LIB_FILE_ERROR;
+        ret = sg_convert_errno(-sg_fd);
         goto done;
     }
 
-    ptvp = construct_scsi_pt_obj();
+    ptvp = construct_scsi_pt_obj_with_fd(sg_fd, op->verbose);
     if (ptvp == NULL) {
-        pr2serr("out of memory\n");
+        pr2serr("construct_scsi_pt_obj_with_fd() failed\n");
         ret = SG_LIB_CAT_OTHER;
         goto done;
     }
 
+    if (op->scan_first < op->scan_last)
+        do_scan = true;
+
+and_again:
+    if (do_scan) {
+        op->cdb[0] = op->scan_first;
+        printf("Command bytes in hex:");
+        for (k = 0; k < op->cdb_length; ++k)
+            printf(" %02x", op->cdb[k]);
+        printf("\n");
+    }
+
     is_scsi_cdb = sg_is_scsi_cdb(op->cdb, op->cdb_length);
+    if ((1 == op->cmdset) && !is_scsi_cdb)
+        is_scsi_cdb = true;
+    else if ((2 == op->cmdset) && is_scsi_cdb)
+        is_scsi_cdb = false;
+
     if (op->do_dataout) {
         uint32_t dout_len;
 
@@ -723,12 +647,11 @@ main(int argc, char *argv[])
     if (op->do_datain) {
         uint32_t din_len = op->datain_len;
 
-        dinp = sg_memalign(din_len, 0 /* page_size */, &wrkBuf,
-                           op->verbose > 3);
+        dinp = sg_memalign(din_len, 0 /* page_size */, &wrkBuf, false);
         if (dinp == NULL) {
             pr2serr("sg_memalign: failed to get %d bytes of memory\n",
                     din_len);
-            ret = SG_LIB_OS_BASE_ERR + ENOMEM;
+            ret = sg_convert_errno(ENOMEM);
             goto done;
         }
         if (op->verbose > 2)
@@ -746,22 +669,19 @@ main(int argc, char *argv[])
         }
     }
     if (op->verbose) {
+        char d[128];
+
         pr2serr("    %s to send: ", is_scsi_cdb ? "cdb" : "cmd");
         if (is_scsi_cdb) {
-            for (k = 0; k < op->cdb_length; ++k)
-                pr2serr("%02x ", op->cdb[k]);
-            pr2serr("\n");
-            if (op->verbose > 1) {
-                sg_get_command_name(op->cdb, 0, b_len - 1, b);
-                b[b_len - 1] = '\0';
-                pr2serr("    Command name: %s\n", b);
-            }
+            pr2serr("%s\n", sg_get_command_str(op->cdb, op->cdb_length,
+                                               op->verbose > 1,
+                                               sizeof(d), d));
         } else {        /* If not SCSI cdb then treat as NVMe command */
             pr2serr("\n");
             hex2stderr(op->cdb, op->cdb_length, -1);
             if (op->verbose > 1)
                 pr2serr("  Command name: %s\n",
-                        sg_get_nvme_opcode_name(op->cdb[0], true /* admin */,
+                        sg_get_nvme_opcode_name(op->cdb[0], ! op->do_nvm,
                                                 b_len, b));
         }
     }
@@ -771,7 +691,10 @@ main(int argc, char *argv[])
                 (int)sizeof(sense_buffer));
     set_scsi_pt_sense(ptvp, sense_buffer, sizeof(sense_buffer));
 
-    ret = do_scsi_pt(ptvp, sg_fd, op->timeout, op->verbose);
+    if (op->do_nvm)
+        ret = do_nvm_pt(ptvp, 0, op->timeout, op->verbose);
+    else
+        ret = do_scsi_pt(ptvp, -1, op->timeout, op->verbose);
     if (ret > 0) {
         switch (ret) {
         case SCSI_PT_DO_BAD_PARAMS:
@@ -797,6 +720,10 @@ main(int argc, char *argv[])
                 }
             }
             break;
+        case SCSI_PT_DO_NOT_SUPPORTED:
+            pr2serr("do_scsi_pt: not supported\n");
+            ret = SG_LIB_CAT_TIMEOUT;
+            break;
         default:
             pr2serr("do_scsi_pt: unknown error: %d\n", ret);
             ret = SG_LIB_CAT_OTHER;
@@ -807,7 +734,7 @@ main(int argc, char *argv[])
         k = -ret;
         pr2serr("do_scsi_pt: %s\n", safe_strerror(k));
         err = get_scsi_pt_os_err(ptvp);
-        if (err != k)
+        if ((err != 0) && (err != k))
             pr2serr("    ... or perhaps: %s\n", safe_strerror(err));
         ret = sg_convert_errno(err);
         goto done;
@@ -897,6 +824,14 @@ main(int argc, char *argv[])
     }
 
 done:
+    if (do_scan) {
+        ++op->scan_first;
+        if (op->scan_first <= op->scan_last) {
+            clear_scsi_pt_obj(ptvp);
+            goto and_again;
+        }
+    }
+
     if (op->verbose && is_scsi_cdb) {
         sg_get_category_sense_str(ret, b_len, b, op->verbose - 1);
         pr2serr("%s\n", b);
