@@ -1,8 +1,10 @@
 /*
- * Copyright (c) 2016-2018 Douglas Gilbert.
+ * Copyright (c) 2016-2021 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <unistd.h>
@@ -12,6 +14,7 @@
 #include <stdarg.h>
 #include <stdbool.h>
 #include <string.h>
+#include <errno.h>
 #include <ctype.h>
 #include <getopt.h>
 #define __STDC_FORMAT_MACROS 1
@@ -36,7 +39,7 @@
  * and decodes the response. Based on spc5r08.pdf
  */
 
-static const char * version_str = "1.08 20180219";
+static const char * version_str = "1.15 20210830";
 
 #define MAX_RATTR_BUFF_LEN (1024 * 1024)
 #define DEF_RATTR_BUFF_LEN (1024 * 8)
@@ -66,6 +69,8 @@ struct opts_t {
     bool enumerate;
     bool do_raw;
     bool o_readonly;
+    bool verbose_given;
+    bool version_given;
     int elem_addr;
     int filter;
     int fai;
@@ -240,7 +245,7 @@ static int
 sg_ll_read_attr(int sg_fd, void * resp, int * residp, bool noisy,
                 const struct opts_t * op)
 {
-    int k, ret, res, sense_cat;
+    int ret, res, sense_cat;
     uint8_t ra_cdb[SG_READ_ATTRIBUTE_CMDLEN] =
           {SG_READ_ATTRIBUTE_CMD, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
            0, 0, 0, 0};
@@ -260,10 +265,11 @@ sg_ll_read_attr(int sg_fd, void * resp, int * residp, bool noisy,
     if (op->cache)
         ra_cdb[14] |= 0x1;
     if (op->verbose) {
-        pr2serr("    Read attribute cdb: ");
-        for (k = 0; k < SG_READ_ATTRIBUTE_CMDLEN; ++k)
-            pr2serr("%02x ", ra_cdb[k]);
-        pr2serr("\n");
+        char b[128];
+
+        pr2serr("    Read attribute cdb: %s\n",
+                sg_get_command_str(ra_cdb, SG_READ_ATTRIBUTE_CMDLEN, false,
+                                   sizeof(b), b));
     }
 
     ptvp = construct_scsi_pt_obj();
@@ -275,11 +281,14 @@ sg_ll_read_attr(int sg_fd, void * resp, int * residp, bool noisy,
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     set_scsi_pt_data_in(ptvp, (uint8_t *)resp, op->maxlen);
     res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, op->verbose);
-    ret = sg_cmds_process_resp(ptvp, "read attribute", res, op->maxlen,
-                               sense_b, noisy, op->verbose, &sense_cat);
-    if (-1 == ret)
-        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
-    else if (-2 == ret) {
+    ret = sg_cmds_process_resp(ptvp, "read attribute", res, noisy,
+                               op->verbose, &sense_cat);
+    if (-1 == ret) {
+        if (get_scsi_pt_transport_err(ptvp))
+            ret = SG_LIB_TRANSPORT_ERROR;
+        else
+            ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
+    } else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
         case SG_LIB_CAT_NO_SENSE:
@@ -317,7 +326,7 @@ find_sa_acron(const char * cp)
         for (mp = cp, k = 0; *mp; ++mp, ++k) {
             if (0 == anvp->acron[k])
                 return anvp->val;
-            if (tolower(*mp) != anvp->acron[k])
+            if (tolower((uint8_t)*mp) != (uint8_t)anvp->acron[k])
                 break;
         }
         if ((0 == *mp) && (0 == anvp->acron[k]))
@@ -368,185 +377,6 @@ enum_sa_acrons(void)
         printf("  %d:\t\t%s\t%s\n", anvp->val, anvp->acron, anvp->name);
 }
 
-/* Read ASCII hex bytes or binary from fname (a file named '-' taken as
- * stdin). If reading ASCII hex then there should be either one entry per
- * line or a comma, space or tab separated list of bytes. If no_space is
- * set then a string of ACSII hex digits is expected, 2 per byte. Everything
- * from and including a '#' on a line is ignored. Returns 0 if ok, or 1 if
- * error. */
-static int
-f2hex_arr(const char * fname, bool as_binary, bool no_space,
-          uint8_t * mp_arr, int * mp_arr_len, int max_arr_len)
-{
-    bool split_line, has_stdin;
-    int fn_len, in_len, k, j, m, fd;
-    int off = 0;
-    unsigned int h;
-    const char * lcp;
-    FILE * fp;
-    char line[512];
-    char carry_over[4];
-
-    if ((NULL == fname) || (NULL == mp_arr) || (NULL == mp_arr_len))
-        return 1;
-    fn_len = strlen(fname);
-    if (0 == fn_len)
-        return 1;
-    has_stdin = ((1 == fn_len) && ('-' == fname[0]));  /* read from stdin */
-    if (as_binary) {
-        if (has_stdin) {
-            fd = STDIN_FILENO;
-                if (sg_set_binary_mode(STDIN_FILENO) < 0)
-                    perror("sg_set_binary_mode");
-        } else {
-            fd = open(fname, O_RDONLY);
-            if (fd < 0) {
-                pr2serr("unable to open binary file %s: %s\n", fname,
-                         safe_strerror(errno));
-                return 1;
-            } else if (sg_set_binary_mode(fd) < 0)
-                perror("sg_set_binary_mode");
-        }
-        k = read(fd, mp_arr, max_arr_len);
-        if (k <= 0) {
-            if (0 == k)
-                pr2serr("read 0 bytes from binary file %s\n", fname);
-            else
-                pr2serr("read from binary file %s: %s\n", fname,
-                        safe_strerror(errno));
-            if (! has_stdin)
-                close(fd);
-            return 1;
-        }
-        *mp_arr_len = k;
-        if (! has_stdin)
-            close(fd);
-        return 0;
-    } else {    /* So read the file as ASCII hex */
-        if (has_stdin)
-            fp = stdin;
-        else {
-            fp = fopen(fname, "r");
-            if (NULL == fp) {
-                pr2serr("Unable to open %s for reading\n", fname);
-                return 1;
-            }
-        }
-    }
-
-    carry_over[0] = 0;
-    for (j = 0; j < 512; ++j) {
-        if (NULL == fgets(line, sizeof(line), fp))
-            break;
-        in_len = strlen(line);
-        if (in_len > 0) {
-            if ('\n' == line[in_len - 1]) {
-                --in_len;
-                line[in_len] = '\0';
-                split_line = false;
-            } else
-                split_line = true;
-        }
-        if (in_len < 1) {
-            carry_over[0] = 0;
-            continue;
-        }
-        if (carry_over[0]) {
-            if (isxdigit(line[0])) {
-                carry_over[1] = line[0];
-                carry_over[2] = '\0';
-                if (1 == sscanf(carry_over, "%4x", &h))
-                    mp_arr[off - 1] = h;       /* back up and overwrite */
-                else {
-                    pr2serr("%s: carry_over error ['%s'] around line %d\n",
-                            __func__, carry_over, j + 1);
-                    goto bad;
-                }
-                lcp = line + 1;
-                --in_len;
-            } else
-                lcp = line;
-            carry_over[0] = 0;
-        } else
-            lcp = line;
-
-        m = strspn(lcp, " \t");
-        if (m == in_len)
-            continue;
-        lcp += m;
-        in_len -= m;
-        if ('#' == *lcp)
-            continue;
-        k = strspn(lcp, "0123456789aAbBcCdDeEfF ,\t");
-        if ((k < in_len) && ('#' != lcp[k]) && ('\r' != lcp[k])) {
-            pr2serr("%s: syntax error at line %d, pos %d\n", __func__,
-                    j + 1, m + k + 1);
-            goto bad;
-        }
-        if (no_space) {
-            for (k = 0; isxdigit(*lcp) && isxdigit(*(lcp + 1));
-                 ++k, lcp += 2) {
-                if (1 != sscanf(lcp, "%2x", &h)) {
-                    pr2serr("%s: bad hex number in line %d, pos %d\n",
-                            __func__, j + 1, (int)(lcp - line + 1));
-                    goto bad;
-                }
-                if ((off + k) >= max_arr_len) {
-                    pr2serr("%s: array length exceeded\n", __func__);
-                    goto bad;
-                }
-                mp_arr[off + k] = h;
-            }
-            if (isxdigit(*lcp) && (! isxdigit(*(lcp + 1))))
-                carry_over[0] = *lcp;
-            off += k;
-        } else {
-            for (k = 0; k < 1024; ++k) {
-                if (1 == sscanf(lcp, "%4x", &h)) {
-                    if (h > 0xff) {
-                        pr2serr("%s: hex number larger than 0xff in line %d, "
-                                "pos %d\n", __func__, j + 1,
-                                (int)(lcp - line + 1));
-                        goto bad;
-                    }
-                    if (split_line && (1 == strlen(lcp))) {
-                        /* single trailing hex digit might be a split pair */
-                        carry_over[0] = *lcp;
-                    }
-                    if ((off + k) >= max_arr_len) {
-                        pr2serr("%s: array length exceeded\n", __func__);
-                        goto bad;
-                    }
-                    mp_arr[off + k] = h;
-                    lcp = strpbrk(lcp, " ,\t");
-                    if (NULL == lcp)
-                        break;
-                    lcp += strspn(lcp, " ,\t");
-                    if ('\0' == *lcp)
-                        break;
-                } else {
-                    if (('#' == *lcp) || ('\r' == *lcp)) {
-                        --k;
-                        break;
-                    }
-                    pr2serr("%s: error in line %d, at pos %d\n", __func__,
-                            j + 1, (int)(lcp - line + 1));
-                    goto bad;
-                }
-            }
-            off += (k + 1);
-        }
-    }
-    *mp_arr_len = off;
-    if (stdin != fp)
-        fclose(fp);
-    return 0;
-bad:
-    if (stdin != fp)
-        fclose(fp);
-    return 1;
-}
-
 /* Returns 1 if 'bp' all 0xff bytes, returns 2 is all 0xff bytes apart
  * from last being 0xfe; otherwise returns 0. */
 static int
@@ -906,12 +736,14 @@ decode_all_sa_s(const uint8_t * rabp, int len, const struct opts_t * op)
 int
 main(int argc, char * argv[])
 {
-    int sg_fd, res, c, len, resid, rlen, in_len;
+    int sg_fd, res, c, len, resid, rlen;
     unsigned int ra_len;
+    int in_len = 0;
     int ret = 0;
     const char * device_name = NULL;
     const char * fname = NULL;
     uint8_t * rabp = NULL;
+    uint8_t * free_rabp = NULL;
     struct opts_t opts;
     struct opts_t * op;
     char b[80];
@@ -998,7 +830,7 @@ main(int argc, char * argv[])
             op->o_readonly = true;
             break;
         case 's':
-           if (isdigit(*optarg)) {
+           if (isdigit((uint8_t)*optarg)) {
                op->sa = sg_get_num(optarg);
                if ((op->sa < 0) || (op->sa > 63)) {
                     pr2serr("bad argument to '--sa=SA', expect 0 to 63\n");
@@ -1014,11 +846,12 @@ main(int argc, char * argv[])
             }
             break;
         case 'v':
+            op->verbose_given = true;
             ++op->verbose;
             break;
         case 'V':
-            pr2serr("version: %s\n", version_str);
-            return 0;
+            op->version_given = true;
+            break;
         default:
             pr2serr("unrecognised option code 0x%x ??\n", c);
             usage();
@@ -1037,6 +870,26 @@ main(int argc, char * argv[])
             return SG_LIB_SYNTAX_ERROR;
         }
     }
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (op->verbose_given && op->version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        op->verbose_given = false;
+        op->version_given = false;
+        op->verbose = 0;
+    } else if (! op->verbose_given) {
+        pr2serr("set '-vv'\n");
+        op->verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", op->verbose);
+#else
+    if (op->verbose_given && op->version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (op->version_given) {
+        pr2serr("version: %s\n", version_str);
+        return 0;
+    }
 
     if (op->enumerate) {
         enum_attributes();
@@ -1052,19 +905,17 @@ main(int argc, char * argv[])
 
     if (0 == op->maxlen)
         op->maxlen = DEF_RATTR_BUFF_LEN;
-    rabp = (uint8_t *)calloc(1, op->maxlen);
+    rabp = (uint8_t *)sg_memalign(op->maxlen, 0, &free_rabp, op->verbose > 3);
     if (NULL == rabp) {
-        pr2serr("unable to calloc %d bytes\n", op->maxlen);
-        return SG_LIB_CAT_OTHER;
+        pr2serr("unable to sg_memalign %d bytes\n", op->maxlen);
+        return sg_convert_errno(ENOMEM);
     }
 
     if (NULL == device_name) {
         if (fname) {
-            if (f2hex_arr(fname, op->do_raw, 0 /* no space */, rabp,
-                          &in_len, op->maxlen)) {
-                ret = SG_LIB_FILE_ERROR;
+            if ((ret = sg_f2hex_arr(fname, op->do_raw, false /* no space */,
+                                    rabp, &in_len, op->maxlen)))
                 goto clean_up;
-            }
             if (op->do_raw)
                 op->do_raw = false;    /* can interfere on decode */
             if (in_len < 4) {
@@ -1094,7 +945,7 @@ main(int argc, char * argv[])
     if (sg_fd < 0) {
         pr2serr("open error: %s: %s\n", device_name,
                 safe_strerror(-sg_fd));
-        ret = SG_LIB_FILE_ERROR;
+        ret = sg_convert_errno(-sg_fd);
         goto clean_up;
     }
 
@@ -1139,10 +990,15 @@ close_then_end:
     if (res < 0) {
         pr2serr("close error: %s\n", safe_strerror(-res));
         if (0 == ret)
-            ret = SG_LIB_FILE_ERROR;
+            ret = sg_convert_errno(-res);
     }
 clean_up:
-    if (rabp)
-        free(rabp);
+    if (free_rabp)
+        free(free_rabp);
+    if (0 == op->verbose) {
+        if (! sg_if_can2stderr("sg_read_attr failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' or '-vv' for "
+                    "more information\n");
+    }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
