@@ -3,6 +3,8 @@
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <unistd.h>
@@ -31,7 +33,7 @@
  * to the given SCSI device.
  */
 
-static const char * version_str = "1.25 20180219";
+static const char * version_str = "1.27 20180628";
 
 #define REPORT_TGT_GRP_BUFF_LEN 1024
 
@@ -59,7 +61,9 @@ static struct option long_options[] = {
         {0, 0, 0, 0},
 };
 
-static void usage()
+
+static void
+usage()
 {
     pr2serr("Usage: sg_rtpg   [--decode] [--extended] [--help] [--hex] "
             "[--raw] [--readonly]\n"
@@ -78,7 +82,8 @@ static void usage()
 
 }
 
-static void dStrRaw(const uint8_t * str, int len)
+static void
+dStrRaw(const uint8_t * str, int len)
 {
     int k;
 
@@ -86,7 +91,8 @@ static void dStrRaw(const uint8_t * str, int len)
         printf("%c", str[k]);
 }
 
-static void decode_status(const int st)
+static void
+decode_status(const int st)
 {
     switch (st) {
     case STATUS_CODE_NOSTATUS:
@@ -106,7 +112,8 @@ static void decode_status(const int st)
     }
 }
 
-static void decode_tpgs_state(const int st)
+static void
+decode_tpgs_state(const int st)
 {
     switch (st) {
     case TPGS_STATE_OPTIMIZED:
@@ -136,14 +143,18 @@ static void decode_tpgs_state(const int st)
     }
 }
 
-int main(int argc, char * argv[])
+int
+main(int argc, char * argv[])
 {
     bool decode = false;
-    int hex = false;
+    bool hex = false;
     bool raw = false;
     bool o_readonly = false;
     bool extended = false;
-    int sg_fd, k, j, off, res, c, report_len, tgt_port_count;
+    bool verbose_given = false;
+    bool version_given = false;
+    int k, j, off, res, c, report_len, tgt_port_count;
+    int sg_fd = -1;
     int ret = 0;
     int verbose = 0;
     uint8_t reportTgtGrpBuff[REPORT_TGT_GRP_BUFF_LEN];
@@ -179,11 +190,12 @@ int main(int argc, char * argv[])
             o_readonly = true;
             break;
         case 'v':
+            verbose_given = true;
             ++verbose;
             break;
         case 'V':
-            pr2serr("Version: %s\n", version_str);
-            return 0;
+            version_given = true;
+            break;
         default:
             pr2serr("unrecognised option code 0x%x ??\n", c);
             usage();
@@ -202,9 +214,28 @@ int main(int argc, char * argv[])
             return SG_LIB_SYNTAX_ERROR;
         }
     }
-
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr("Version: %s\n", version_str);
+        return 0;
+    }
     if (NULL == device_name) {
-        pr2serr("missing device name!\n");
+        pr2serr("Missing device name!\n\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
@@ -217,8 +248,11 @@ int main(int argc, char * argv[])
 
     sg_fd = sg_cmds_open_device(device_name, o_readonly, verbose);
     if (sg_fd < 0) {
-        pr2serr("open error: %s: %s\n", device_name, safe_strerror(-sg_fd));
-        return SG_LIB_FILE_ERROR;
+        if (verbose)
+            pr2serr("open error: %s: %s\n", device_name,
+                    safe_strerror(-sg_fd));
+        ret = sg_convert_errno(-sg_fd);
+        goto err_out;
     }
 
     memset(reportTgtGrpBuff, 0x0, sizeof(reportTgtGrpBuff));
@@ -312,11 +346,18 @@ int main(int argc, char * argv[])
     }
 
 err_out:
-    res = sg_cmds_close_device(sg_fd);
-    if (res < 0) {
-        pr2serr("close error: %s\n", safe_strerror(-res));
-        if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+    if (sg_fd >= 0) {
+        res = sg_cmds_close_device(sg_fd);
+        if (res < 0) {
+            pr2serr("close error: %s\n", safe_strerror(-res));
+            if (0 == ret)
+                ret = sg_convert_errno(-res);
+        }
+    }
+    if (0 == verbose) {
+        if (! sg_if_can2stderr("sg_rtpg failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' "
+                    "or '-vv' for more information\n");
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
