@@ -3,6 +3,8 @@
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <unistd.h>
@@ -32,7 +34,7 @@
  *  to the 'SCSI Accessed Fault-Tolerant Enclosures' (SAF-TE) spec.
  */
 
-static const char * version_str = "0.31 20180219";
+static const char * version_str = "0.33 20180628";
 
 
 #define SENSE_BUFF_LEN 64       /* Arbitrary, could be larger */
@@ -469,8 +471,8 @@ do_safte_global_flags(int sg_fd, int do_hex, int do_raw, int verbose)
     return 0;
 }
 
-static
-void usage()
+static void
+usage()
 {
     pr2serr("Usage:  sg_safte [--config] [--devstatus] [--encstatus] "
             "[--flags] [--help]\n"
@@ -514,7 +516,10 @@ main(int argc, char * argv[])
 {
     bool do_insertions = false;
     bool no_hex_raw;
-    int sg_fd, c, ret, peri_type;
+    bool verbose_given = false;
+    bool version_given = false;
+    int c, ret, peri_type;
+    int sg_fd = -1;
     int res = SG_LIB_CAT_OTHER;
     const char * device_name = NULL;
     char ebuff[EBUFF_SZ];
@@ -572,11 +577,12 @@ main(int argc, char * argv[])
                 do_usage = true;
                 break;
             case 'v':
+                verbose_given = true;
                 ++verbose;
                 break;
             case 'V':
-                pr2serr("Version string: %s\n", version_str);
-                exit(0);
+                version_given = true;
+                break;
             default:
                 pr2serr("unrecognised option code 0x%x ??\n", c);
                 usage();
@@ -596,8 +602,29 @@ main(int argc, char * argv[])
         }
     }
 
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr("Version string: %s\n", version_str);
+        return 0;
+    }
+
     if (NULL == device_name) {
-        pr2serr("missing device name!\n");
+        pr2serr("Missing device name!\n\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
@@ -610,15 +637,18 @@ main(int argc, char * argv[])
 
     if ((sg_fd = sg_cmds_open_device(device_name, false /* rw */, verbose))
         < 0) {
-        snprintf(ebuff, EBUFF_SZ, "sg_safte: error opening file: %s (rw)",
-                 device_name);
-        perror(ebuff);
-        return SG_LIB_FILE_ERROR;
+        if (verbose) {
+            snprintf(ebuff, EBUFF_SZ, "sg_safte: error opening file: %s (rw)",
+                     device_name);
+            perror(ebuff);
+        }
+        ret = sg_convert_errno(-sg_fd);
+        goto fini;
     }
     no_hex_raw = ((0 == do_hex) && (0 == do_raw));
 
     if (no_hex_raw) {
-        if (0 == sg_simple_inquiry(sg_fd, &inq_resp, 1, verbose)) {
+        if (0 == sg_simple_inquiry(sg_fd, &inq_resp, true, verbose)) {
             printf("  %.8s  %.16s  %.4s\n", inq_resp.vendor,
                    inq_resp.product, inq_resp.revision);
             peri_type = inq_resp.peripheral_type;
@@ -728,11 +758,19 @@ err_out:
         break;
     }
     ret = res;
-    res = sg_cmds_close_device(sg_fd);
-    if (res < 0) {
-        pr2serr("close error: %s\n", safe_strerror(-res));
-        if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+fini:
+    if (sg_fd >= 0) {
+        res = sg_cmds_close_device(sg_fd);
+        if (res < 0) {
+            pr2serr("close error: %s\n", safe_strerror(-res));
+            if (0 == ret)
+                ret = sg_convert_errno(-res);
+        }
+    }
+    if (0 == verbose) {
+        if (! sg_if_can2stderr("sg_safte failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' "
+                    "or '-vv' for more information\n");
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
