@@ -1,8 +1,10 @@
 /*
- * Copyright (c) 2011-2018 Douglas Gilbert.
+ * Copyright (c) 2011-2021 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <unistd.h>
@@ -31,7 +33,7 @@
 #include "sg_unaligned.h"
 #include "sg_pr2serr.h"
 
-static const char * version_str = "1.07 20180219";
+static const char * version_str = "1.16 20210830";
 
 /* Not all environments support the Unix sleep() */
 #if defined(MSC_VER) || defined(__MINGW32__)
@@ -70,6 +72,8 @@ static struct option long_options[] = {
     {"count", required_argument, 0, 'c'},
     {"crypto", no_argument, 0, 'C'},
     {"desc", no_argument, 0, 'd'},
+    {"dry-run", no_argument, 0, 'D'},
+    {"dry_run", no_argument, 0, 'D'},
     {"early", no_argument, 0, 'e'},
     {"fail", no_argument, 0, 'F'},
     {"help", no_argument, 0, 'h'},
@@ -92,11 +96,14 @@ struct opts_t {
     bool block;
     bool crypto;
     bool desc;
+    bool dry_run;
     bool early;
     bool fail;
     bool invert;
     bool overwrite;
     bool quick;
+    bool verbose_given;
+    bool version_given;
     bool wait;
     bool znr;
     int count;
@@ -113,13 +120,14 @@ static void
 usage()
 {
   pr2serr("Usage: sg_sanitize [--ause] [--block] [--count=OC] [--crypto] "
-          "[--early]\n"
-          "                   [--fail] [--help] [--invert] [--ipl=LEN] "
-          "[--overwrite]\n"
-          "                   [--pattern=PF] [--quick] [--test=TE] "
-          "[--timeout=SEC]\n"
-          "                   [--verbose] [--version] [--wait] [--zero] "
-          "[--znr] DEVICE\n"
+          "[--dry-run]\n"
+          "                   [--early] [--fail] [--help] [--invert] "
+          "[--ipl=LEN]\n"
+          "                   [--overwrite] [--pattern=PF] [--quick] "
+          "[--test=TE]\n"
+          "                   [--timeout=SECS] [--verbose] [--version] "
+          "[--wait]\n"
+          "                   [--zero] [--znr] DEVICE\n"
           "  where:\n"
           "    --ause|-A            set AUSE bit in cdb\n"
           "    --block|-B           do BLOCK ERASE sanitize\n"
@@ -129,6 +137,8 @@ usage()
           "    --desc|-d            polling request sense sets 'desc' "
           "field\n"
           "                         (def: clear 'desc' field)\n"
+          "    --dry-run|-D         to preparation but bypass SANITIZE "
+          "commnd\n"
           "    --early|-e           exit once sanitize started (IMMED set "
           "in cdb)\n"
           "                         user can monitor progress with REQUEST "
@@ -149,7 +159,8 @@ usage()
           "    --test=TE|-T TE      TE is placed in TEST field of "
           "OVERWRITE\n"
           "                         parameter list (def: 0)\n"
-          "    --timeout=SEC|-t SEC    SANITIZE command timeout in seconds\n"
+          "    --timeout=SECS|-t SECS    SANITIZE command timeout in "
+          "seconds\n"
           "    --verbose|-v         increase verbosity\n"
           "    --version|-V         print version string then exit\n"
           "    --wait|-w            wait for command to finish (could "
@@ -172,7 +183,7 @@ do_sanitize(int sg_fd, const struct opts_t * op, const void * param_lstp,
             int param_lst_len)
 {
     bool immed;
-    int k, ret, res, sense_cat, timeout;
+    int ret, res, sense_cat, timeout;
     uint8_t san_cdb[SANITIZE_OP_LEN];
     uint8_t sense_b[SENSE_BUFF_LEN];
     struct sg_pt_base * ptvp;
@@ -206,18 +217,23 @@ do_sanitize(int sg_fd, const struct opts_t * op, const void * param_lstp,
     sg_put_unaligned_be16((uint16_t)param_lst_len, san_cdb + 7);
 
     if (op->verbose > 1) {
-        pr2serr("    Sanitize cdb: ");
-        for (k = 0; k < SANITIZE_OP_LEN; ++k)
-            pr2serr("%02x ", san_cdb[k]);
-        pr2serr("\n");
+        char b[128];
+
+        pr2serr("    Sanitize cdb: %s\n",
+                sg_get_command_str(san_cdb, SANITIZE_OP_LEN, false,
+                                   sizeof(b), b));
         if (op->verbose > 2) {
             if (param_lst_len > 0) {
                 pr2serr("    Parameter list contents:\n");
-                hex2stderr((const uint8_t *)param_lstp, param_lst_len, 1);
+                hex2stderr((const uint8_t *)param_lstp, param_lst_len, -1);
             }
             pr2serr("    Sanitize command timeout: %d seconds\n", timeout);
         }
     }
+    if (op->dry_run) {
+        pr2serr("Due to --dry-run option, bypassing SANITIZE command\n");
+        return 0;
+    }
     ptvp = construct_scsi_pt_obj();
     if (NULL == ptvp) {
         pr2serr("Sanitize: out of memory\n");
@@ -227,11 +243,14 @@ do_sanitize(int sg_fd, const struct opts_t * op, const void * param_lstp,
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     set_scsi_pt_data_out(ptvp, (uint8_t *)param_lstp, param_lst_len);
     res = do_scsi_pt(ptvp, sg_fd, timeout, op->verbose);
-    ret = sg_cmds_process_resp(ptvp, "Sanitize", res, SG_NO_DATA_IN, sense_b,
-                               true /*noisy */, op->verbose, &sense_cat);
-    if (-1 == ret)
-        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
-    else if (-2 == ret) {
+    ret = sg_cmds_process_resp(ptvp, "Sanitize", res, true /*noisy */,
+                               op->verbose, &sense_cat);
+    if (-1 == ret) {
+        if (get_scsi_pt_transport_err(ptvp))
+            ret = SG_LIB_TRANSPORT_ERROR;
+        else
+            ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
+    } else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
         case SG_LIB_CAT_NO_SENSE:
@@ -255,8 +274,12 @@ do_sanitize(int sg_fd, const struct opts_t * op, const void * param_lstp,
             ret = sense_cat;
             break;
         }
-    } else
+    } else {
         ret = 0;
+        if (op->verbose)
+            pr2serr("Sanitize command %s without error\n",
+                    (immed ? "launched" : "completed"));
+    }
 
     destruct_scsi_pt_obj(ptvp);
     return ret;
@@ -374,14 +397,9 @@ print_dev_id(int fd, uint8_t * sinq_resp, int max_rlen, int verbose)
     if (n > (SAFE_STD_INQ_RESP_LEN - 4))
         n = (SAFE_STD_INQ_RESP_LEN - 4);
     for (k = 0, has_sn = 0, has_di = 0; k < n; ++k) {
-        if (VPD_UNIT_SERIAL_NUM == b[4 + k]) {
-            if (has_di) {
-                if (verbose)
-                    pr2serr("VPD_SUPPORTED_VPDS dis-ordered\n");
-                return 0;
-            }
+        if (VPD_UNIT_SERIAL_NUM == b[4 + k])
             ++has_sn;
-        } else if (VPD_DEVICE_ID == b[4 + k]) {
+        else if (VPD_DEVICE_ID == b[4 + k]) {
             ++has_di;
             break;
         }
@@ -432,7 +450,8 @@ int
 main(int argc, char * argv[])
 {
     bool got_stdin = false;
-    int sg_fd, k, res, c, infd, progress, vb, n, resp_len;
+    int k, res, c, infd, progress, vb, n, resp_len, err;
+    int sg_fd = -1;
     int param_lst_len = 0;
     int ret = -1;
     const char * device_name = NULL;
@@ -440,6 +459,7 @@ main(int argc, char * argv[])
     char b[80];
     uint8_t rsBuff[DEF_REQS_RESP_LEN];
     uint8_t * wBuff = NULL;
+    uint8_t * free_wBuff = NULL;
     struct opts_t opts;
     struct opts_t * op;
     struct stat a_stat;
@@ -451,8 +471,8 @@ main(int argc, char * argv[])
     while (1) {
         int option_index = 0;
 
-        c = getopt_long(argc, argv, "ABc:CdeFhi:IOp:Qt:T:vVwzZ", long_options,
-                        &option_index);
+        c = getopt_long(argc, argv, "ABc:CdDeFhi:IOp:Qt:T:vVwzZ",
+                        long_options, &option_index);
         if (c == -1)
             break;
 
@@ -476,6 +496,9 @@ main(int argc, char * argv[])
         case 'd':
             op->desc = true;
             break;
+        case 'D':
+            op->dry_run = true;
+            break;
         case 'e':
             op->early = true;
             break;
@@ -508,7 +531,7 @@ main(int argc, char * argv[])
         case 't':
             op->timeout = sg_get_num(optarg);
             if (op->timeout < 0) {
-                pr2serr("bad argument to '--timeout=SEC', want 0 or more\n");
+                pr2serr("bad argument to '--timeout=SECS', want 0 or more\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             break;
@@ -520,11 +543,12 @@ main(int argc, char * argv[])
             }
             break;
         case 'v':
+            op->verbose_given = true;
             ++op->verbose;
             break;
         case 'V':
-            pr2serr(ME "version: %s\n", version_str);
-            return 0;
+            op->version_given = true;
+            break;
         case 'w':
             op->wait = true;
             break;
@@ -552,8 +576,29 @@ main(int argc, char * argv[])
             return SG_LIB_SYNTAX_ERROR;
         }
     }
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (op->verbose_given && op->version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        op->verbose_given = false;
+        op->version_given = false;
+        op->verbose = 0;
+    } else if (! op->verbose_given) {
+        pr2serr("set '-vv'\n");
+        op->verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", op->verbose);
+#else
+    if (op->verbose_given && op->version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (op->version_given) {
+        pr2serr(ME "version: %s\n", version_str);
+        return 0;
+    }
+
     if (NULL == device_name) {
-        pr2serr("missing device name!\n");
+        pr2serr("Missing device name!\n\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
@@ -562,29 +607,31 @@ main(int argc, char * argv[])
     if (1 != n) {
         pr2serr("one and only one of '--block', '--crypto', '--fail' or "
                 "'--overwrite' please\n");
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
     if (op->overwrite) {
         if (op->zero) {
             if (op->pattern_fn) {
                 pr2serr("confused: both '--pattern=PF' and '--zero' "
                         "options\n");
-                return SG_LIB_SYNTAX_ERROR;
+                return SG_LIB_CONTRADICT;
             }
             op->ipl = 4;
         } else {
             if (NULL == op->pattern_fn) {
                 pr2serr("'--overwrite' requires '--pattern=PF' or '--zero' "
                         "option\n");
-                return SG_LIB_SYNTAX_ERROR;
+                return SG_LIB_CONTRADICT;
             }
             got_stdin = (0 == strcmp(op->pattern_fn, "-"));
             if (! got_stdin) {
                 memset(&a_stat, 0, sizeof(a_stat));
                 if (stat(op->pattern_fn, &a_stat) < 0) {
+                    err = errno;
                     pr2serr("pattern file: unable to stat(%s): %s\n",
-                            op->pattern_fn, safe_strerror(errno));
-                    return SG_LIB_FILE_ERROR;
+                            op->pattern_fn, safe_strerror(err));
+                    ret = sg_convert_errno(err);
+                    goto err_out;
                 }
                 if (op->ipl <= 0) {
                     op->ipl = (int)a_stat.st_size;
@@ -598,16 +645,18 @@ main(int argc, char * argv[])
             if (op->ipl < 1) {
                 pr2serr("'--overwrite' requires '--ipl=LEN' option if can't "
                         "get PF length\n");
-                return SG_LIB_SYNTAX_ERROR;
+                return SG_LIB_CONTRADICT;
             }
         }
     }
 
     sg_fd = sg_cmds_open_device(device_name, false /* rw */, vb);
     if (sg_fd < 0) {
-        pr2serr(ME "open error: %s: %s\n", device_name,
-                safe_strerror(-sg_fd));
-        return SG_LIB_FILE_ERROR;
+        if (op->verbose)
+            pr2serr(ME "open error: %s: %s\n", device_name,
+                    safe_strerror(-sg_fd));
+        ret = sg_convert_errno(-sg_fd);
+        goto err_out;
     }
 
     ret = print_dev_id(sg_fd, inq_resp, sizeof(inq_resp), op->verbose);
@@ -616,11 +665,11 @@ main(int argc, char * argv[])
 
     if (op->overwrite) {
         param_lst_len = op->ipl + 4;
-        wBuff = (uint8_t*)calloc(op->ipl + 4, 1);
+        wBuff = (uint8_t*)sg_memalign(op->ipl + 4, 0, &free_wBuff, false);
         if (NULL == wBuff) {
             pr2serr("unable to allocate %d bytes of memory with calloc()\n",
                     op->ipl + 4);
-            ret = SG_LIB_SYNTAX_ERROR;
+            ret = sg_convert_errno(ENOMEM);
             goto err_out;
         }
         if (op->zero) {
@@ -635,22 +684,24 @@ main(int argc, char * argv[])
                     perror("sg_set_binary_mode");
             } else {
                 if ((infd = open(op->pattern_fn, O_RDONLY)) < 0) {
+                    err = errno;
                     snprintf(ebuff, EBUFF_SZ, ME "could not open %s for "
                              "reading", op->pattern_fn);
                     perror(ebuff);
-                    ret = SG_LIB_FILE_ERROR;
+                    ret = sg_convert_errno(err);
                     goto err_out;
                 } else if (sg_set_binary_mode(infd) < 0)
                     perror("sg_set_binary_mode");
             }
             res = read(infd, wBuff + 4, op->ipl);
             if (res < 0) {
+                err = errno;
                 snprintf(ebuff, EBUFF_SZ, ME "couldn't read from %s",
                          op->pattern_fn);
                 perror(ebuff);
                 if (! got_stdin)
                     close(infd);
-                ret = SG_LIB_FILE_ERROR;
+                ret = sg_convert_errno(err);
                 goto err_out;
             }
             if (res < op->ipl) {
@@ -661,7 +712,7 @@ main(int argc, char * argv[])
             if (! got_stdin)
                 close(infd);
         }
-        wBuff[0] = op->count & 0x1f;;
+        wBuff[0] = op->count & 0x1f;
         if (op->test)
             wBuff[0] |= ((op->test & 0x3) << 5);
         if (op->invert)
@@ -692,6 +743,10 @@ main(int argc, char * argv[])
 
     if ((0 == ret) && (! op->early) && (! op->wait)) {
         for (k = 0; ;++k) {     /* unbounded, exits via break */
+            if (op->dry_run && (k > 0)) {
+                pr2serr("Due to --dry-run option, leave poll loop\n");
+                break;
+            }
             sleep_for(POLL_DURATION_SECS);
             memset(rsBuff, 0x0, sizeof(rsBuff));
             res = sg_ll_request_sense(sg_fd, op->desc, rsBuff, sizeof(rsBuff),
@@ -721,7 +776,7 @@ main(int argc, char * argv[])
             resp_len = rsBuff[7] + 8;
             if (vb > 2) {
                 pr2serr("Parameter data in hex\n");
-                hex2stderr(rsBuff, resp_len, 1);
+                hex2stderr(rsBuff, resp_len, -1);
             }
             progress = -1;
             sg_get_sense_progress_fld(rsBuff, resp_len, &progress);
@@ -730,6 +785,9 @@ main(int argc, char * argv[])
                 if (vb > 1)
                      pr2serr("No progress indication found, iteration %d\n",
                              k + 1);
+                if ((0 == k) && vb)
+                     pr2serr("Sanitize seems to be successful and finished "
+                             "quickly\n");
                 /* N.B. exits first time there isn't a progress indication */
                 break;
             } else
@@ -739,13 +797,20 @@ main(int argc, char * argv[])
     }
 
 err_out:
-    if (wBuff)
-        free(wBuff);
-    res = sg_cmds_close_device(sg_fd);
-    if (res < 0) {
-        pr2serr("close error: %s\n", safe_strerror(-res));
-        if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+    if (free_wBuff)
+        free(free_wBuff);
+    if (sg_fd >= 0) {
+        res = sg_cmds_close_device(sg_fd);
+        if (res < 0) {
+            pr2serr("close error: %s\n", safe_strerror(-res));
+            if (0 == ret)
+                ret = sg_convert_errno(-res);
+        }
+    }
+    if (0 == op->verbose) {
+        if (! sg_if_can2stderr("sg_sanitize failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' "
+                    "or '-vv' for more information\n");
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
