@@ -1,8 +1,10 @@
 /*
- * Copyright (c) 2006-2018 Douglas Gilbert.
+ * Copyright (c) 2006-2021 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <unistd.h>
@@ -28,7 +30,7 @@
 
 /* This program uses a ATA PASS-THROUGH SCSI command to package an
  * ATA IDENTIFY (PACKAGE) DEVICE command. It is based on the SCSI to
- * ATA Translation (SAT) drafts and standards. See http://www.t10.org
+ * ATA Translation (SAT) drafts and standards. See https://www.t10.org
  * for drafts. SAT is a standard: SAT ANSI INCITS 431-2007 (draft prior
  * to that is sat-r09.pdf). SAT-2 is also a standard: SAT-2 ANSI INCITS
  * 465-2010 and the draft prior to that is sat2r09.pdf . The SAT-3 is
@@ -52,9 +54,10 @@
 
 #define EBUFF_SZ 256
 
-static const char * version_str = "1.16 20180219";
+static const char * version_str = "1.18 20210630";
 
 static struct option long_options[] = {
+        {"ck-cond", no_argument, 0, 'c'},
         {"ck_cond", no_argument, 0, 'c'},
         {"extend", no_argument, 0, 'e'},
         {"help", no_argument, 0, 'h'},
@@ -69,7 +72,9 @@ static struct option long_options[] = {
         {0, 0, 0, 0},
 };
 
-static void usage()
+
+static void
+usage()
 {
     pr2serr("Usage: sg_sat_identify [--ck_cond] [--extend] [--help] [--hex] "
             "[--ident]\n"
@@ -100,7 +105,8 @@ static void usage()
             "support the SCSI ATA\nPASS-THROUGH(32) command.\n");
 }
 
-static void dStrRaw(const uint8_t * str, int len)
+static void
+dStrRaw(const uint8_t * str, int len)
 {
     int k;
 
@@ -108,12 +114,13 @@ static void dStrRaw(const uint8_t * str, int len)
         printf("%c", str[k]);
 }
 
-static int do_identify_dev(int sg_fd, bool do_packet, int cdb_len,
-                           bool ck_cond, bool extend, bool do_ident,
-                           int do_hex, bool do_raw, int verbose)
+static int
+do_identify_dev(int sg_fd, bool do_packet, int cdb_len, bool ck_cond,
+                bool extend, bool do_ident, int do_hex, bool do_raw,
+                int verbose)
 {
-    bool t_type = false;/* false -> 512 byte blocks,
-                           true -> device's LB size */
+    const bool t_type = false;/* false -> 512 byte blocks,
+                                 true -> device's LB size */
     bool t_dir = true;  /* false -> to device, true -> from device */
     bool byte_block = true; /* false -> bytes, true -> 512 byte blocks (if
                                t_type=false) */
@@ -384,7 +391,8 @@ static int do_identify_dev(int sg_fd, bool do_packet, int cdb_len,
     return 0;
 }
 
-int main(int argc, char * argv[])
+int
+main(int argc, char * argv[])
 {
     bool do_packet = false;
     bool do_ident = false;
@@ -392,7 +400,10 @@ int main(int argc, char * argv[])
     bool o_readonly = false;
     bool ck_cond = false;    /* set to true to read register(s) back */
     bool extend = false;    /* set to true to send 48 bit LBA with command */
-    int sg_fd, c, res;
+    bool verbose_given = false;
+    bool version_given = false;
+    int c, res;
+    int sg_fd = -1;
     int cdb_len = SAT_ATA_PASS_THROUGH16_LEN;
     int do_hex = 0;
     int verbose = 0;
@@ -446,11 +457,12 @@ int main(int argc, char * argv[])
             o_readonly = true;
             break;
         case 'v':
+            verbose_given = true;
             ++verbose;
             break;
         case 'V':
-            pr2serr("version: %s\n", version_str);
-            return 0;
+            version_given = true;
+            break;
         default:
             pr2serr("unrecognised option code 0x%x ??\n", c);
             usage();
@@ -469,9 +481,29 @@ int main(int argc, char * argv[])
             return SG_LIB_SYNTAX_ERROR;
         }
     }
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr("version: %s\n", version_str);
+        return 0;
+    }
 
     if (NULL == device_name) {
-        pr2serr("missing device name!\n");
+        pr2serr("Missing device name!\n\n");
         usage();
         return 1;
     }
@@ -483,19 +515,29 @@ int main(int argc, char * argv[])
     }
 
     if ((sg_fd = sg_cmds_open_device(device_name, o_readonly, verbose)) < 0) {
-        pr2serr("error opening file: %s: %s\n", device_name,
-                safe_strerror(-sg_fd));
-        return SG_LIB_FILE_ERROR;
+        if (verbose)
+            pr2serr("error opening file: %s: %s\n", device_name,
+                    safe_strerror(-sg_fd));
+        ret = sg_convert_errno(-sg_fd);
+        goto fini;
     }
 
     ret = do_identify_dev(sg_fd, do_packet, cdb_len, ck_cond, extend,
                           do_ident, do_hex, do_raw, verbose);
 
-    res = sg_cmds_close_device(sg_fd);
-    if (res < 0) {
-        pr2serr("close error: %s\n", safe_strerror(-res));
-        if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+fini:
+    if (sg_fd >= 0) {
+        res = sg_cmds_close_device(sg_fd);
+        if (res < 0) {
+            pr2serr("close error: %s\n", safe_strerror(-res));
+            if (0 == ret)
+                ret = sg_convert_errno(-res);
+        }
+    }
+    if (0 == verbose) {
+        if (! sg_if_can2stderr("sg_sat_identify failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' "
+                    "or '-vv' for more information\n");
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
