@@ -3,6 +3,8 @@
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <unistd.h>
@@ -11,6 +13,7 @@
 #include <stdlib.h>
 #include <stdarg.h>
 #include <stdbool.h>
+#include <errno.h>
 #include <string.h>
 #include <getopt.h>
 #define __STDC_FORMAT_MACROS 1
@@ -25,11 +28,11 @@
 #include "sg_cmds_extra.h"
 #include "sg_pr2serr.h"
 
-static const char * version_str = "1.11 20180219";
+static const char * version_str = "1.13 20180628";
 
 /* This program uses a ATA PASS-THROUGH SCSI command. This usage is
  * defined in the SCSI to ATA Translation (SAT) drafts and standards.
- * See http://www.t10.org for drafts. SAT is a standard: SAT ANSI INCITS
+ * See https://www.t10.org for drafts. SAT is a standard: SAT ANSI INCITS
  * 431-2007 (draft prior to that is sat-r09.pdf). SAT-2 is also a
  * standard: SAT-2 ANSI INCITS 465-2010 and the draft prior to that is
  * sat2r09.pdf . The SAT-3 project has started and the most recent draft
@@ -62,6 +65,7 @@ static const char * version_str = "1.11 20180219";
 
 static struct option long_options[] = {
         {"ck_cond", no_argument, 0, 'c'},
+        {"ck-cond", no_argument, 0, 'c'},
         {"extend", no_argument, 0, 'e'},
         {"hex", no_argument, 0, 'H'},
         {"ignore", no_argument, 0, 'i'},
@@ -361,12 +365,14 @@ do_read_log_ext(int sg_fd, int log_addr, bool page_in_log, int feature,
 
 int main(int argc, char * argv[])
 {
+    bool ck_cond = false;   /* set to true to read register(s) back */
     bool extend = false;
     bool ignore = false;
     bool raw = false;
     bool reset = false;
-    bool ck_cond = false;   /* set to true to read register(s) back */
-    int sg_fd, c, k, j, res, id, len, vendor;
+    bool verbose_given = false;
+    bool version_given = false;
+    int sg_fd, c, k, j, res, id, len, vendor, err;
     char * device_name = 0;
     char ebuff[EBUFF_SZ];
     uint8_t inBuff[READ_LOG_EXT_RESPONSE_LEN];
@@ -417,11 +423,12 @@ int main(int argc, char * argv[])
             reset = true;
             break;
         case 'v':
+            verbose_given = true;
             ++verbose;
             break;
         case 'V':
-            pr2serr("version: %s\n", version_str);
-            exit(0);
+            version_given = true;
+            break;
         default:
             pr2serr("unrecognised option code %c [0x%x]\n", c, c);
             usage();
@@ -440,8 +447,28 @@ int main(int argc, char * argv[])
             return SG_LIB_SYNTAX_ERROR;
         }
     }
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr("version: %s\n", version_str);
+        return 0;
+    }
     if (0 == device_name) {
-        pr2serr("no DEVICE name detected\n");
+        pr2serr("no DEVICE name detected\n\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
@@ -453,10 +480,11 @@ int main(int argc, char * argv[])
     }
 
     if ((sg_fd = open(device_name, O_RDWR)) < 0) {
+        err = errno;
         snprintf(ebuff, EBUFF_SZ,
                  "sg_sat_phy_event: error opening file: %s", device_name);
         perror(ebuff);
-        return SG_LIB_FILE_ERROR;
+        return sg_convert_errno(err);
     }
     ret = do_read_log_ext(sg_fd, SATA_PHY_EVENT_LPAGE,
                           false /* page_in_log */,
@@ -493,9 +521,10 @@ int main(int argc, char * argv[])
 
     res = close(sg_fd);
     if (res < 0) {
-        pr2serr("close error: %s\n", safe_strerror(-res));
+        err = errno;
+        pr2serr("close error: %s\n", safe_strerror(err));
         if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+            ret = sg_convert_errno(err);
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
