@@ -3,6 +3,8 @@
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <unistd.h>
@@ -27,7 +29,7 @@
 
 /* This program uses a ATA PASS-THROUGH SCSI command. This usage is
  * defined in the SCSI to ATA Translation (SAT) drafts and standards.
- * See http://www.t10.org for drafts. SAT is a standard: SAT ANSI INCITS
+ * See https://www.t10.org for drafts. SAT is a standard: SAT ANSI INCITS
  * 431-2007 (draft prior to that is sat-r09.pdf). SAT-2 is also a
  * standard: SAT-2 ANSI INCITS 465-2010 and the draft prior to that is
  * sat2r09.pdf . The SAT-3 project has started and the most recent draft
@@ -52,7 +54,7 @@
 
 #define DEF_TIMEOUT 20
 
-static const char * version_str = "1.18 20180219";
+static const char * version_str = "1.20 20180628";
 
 struct opts_t {
     bool ck_cond;
@@ -69,6 +71,7 @@ struct opts_t {
 static struct option long_options[] = {
     {"count", required_argument, 0, 'c'},
     {"ck_cond", no_argument, 0, 'C'},
+    {"ck-cond", no_argument, 0, 'C'},
     {"dma", no_argument, 0, 'd'},
     {"help", no_argument, 0, 'h'},
     {"hex", no_argument, 0, 'H'},
@@ -328,9 +331,13 @@ do_read_gplog(int sg_fd, int ata_cmd, uint8_t *inbuff,
 int
 main(int argc, char * argv[])
 {
-    int sg_fd, c, ret, res, n;
+    bool verbose_given = false;
+    bool version_given = false;
+    int c, ret, res, n;
+    int sg_fd = -1;
     int ata_cmd = ATA_READ_LOG_EXT;
-    uint8_t *inbuff;
+    uint8_t *inbuff = NULL;
+    uint8_t *free_inbuff = NULL;
     struct opts_t opts;
     struct opts_t * op;
 
@@ -392,11 +399,12 @@ main(int argc, char * argv[])
             op->rdonly = true;
             break;
         case 'v':
+            verbose_given = true;
             ++op->verbose;
             break;
         case 'V':
-            pr2serr("version: %s\n", version_str);
-            return 0;
+            version_given = true;
+            break;
         default:
             pr2serr("unrecognised option code 0x%x ??\n", c);
             usage();
@@ -417,8 +425,29 @@ main(int argc, char * argv[])
         }
     }
 
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        op->verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        op->verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", op->verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr("version: %s\n", version_str);
+        return 0;
+    }
+
     if (NULL == op->device_name) {
-        pr2serr("missing device name!\n");
+        pr2serr("Missing device name!\n\n");
         usage();
         return 1;
     }
@@ -431,7 +460,7 @@ main(int argc, char * argv[])
     }
 
     n = op->count * 512;
-    inbuff = (uint8_t *)malloc(n);
+    inbuff = (uint8_t *)sg_memalign(n, 0, &free_inbuff, op->verbose > 3);
     if (!inbuff) {
         pr2serr("Cannot allocate output buffer of size %d\n", n);
         return SG_LIB_CAT_OTHER;
@@ -439,21 +468,30 @@ main(int argc, char * argv[])
 
     if ((sg_fd = sg_cmds_open_device(op->device_name, op->rdonly,
                                      op->verbose)) < 0) {
-        pr2serr("error opening file: %s: %s\n", op->device_name,
-                safe_strerror(-sg_fd));
-        ret = SG_LIB_FILE_ERROR;
+        if (op->verbose)
+            pr2serr("error opening file: %s: %s\n", op->device_name,
+                    safe_strerror(-sg_fd));
+        ret = sg_convert_errno(-sg_fd);
         goto fini;
     }
 
     ret = do_read_gplog(sg_fd, ata_cmd, inbuff, op);
 
-    res = sg_cmds_close_device(sg_fd);
-    if (res < 0) {
-        pr2serr("close error: %s\n", safe_strerror(-res));
-        if (0 == ret)
-            ret = SG_LIB_FILE_ERROR;
-    }
 fini:
-    free(inbuff);
+    if (sg_fd >= 0) {
+        res = sg_cmds_close_device(sg_fd);
+        if (res < 0) {
+            pr2serr("close error: %s\n", safe_strerror(-res));
+            if (0 == ret)
+                ret = sg_convert_errno(-res);
+        }
+    }
+    if (0 == op->verbose) {
+        if (! sg_if_can2stderr("sg_sat_read_gplog failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' "
+                    "or '-vv' for more information\n");
+    }
+    if (free_inbuff)
+        free(free_inbuff);
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
