@@ -3,6 +3,8 @@
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <unistd.h>
@@ -26,7 +28,7 @@
 
 /* This program uses a ATA PASS-THROUGH SCSI command. This usage is
  * defined in the SCSI to ATA Translation (SAT) drafts and standards.
- * See http://www.t10.org for drafts. SAT is a standard: SAT ANSI INCITS
+ * See https://www.t10.org for drafts. SAT is a standard: SAT ANSI INCITS
  * 431-2007 (draft prior to that is sat-r09.pdf). SAT-2 is also a
  * standard: SAT-2 ANSI INCITS 465-2010 and the draft prior to that is
  * sat2r09.pdf . The SAT-3 project has started and the most recent draft
@@ -50,11 +52,12 @@
 
 #define DEF_TIMEOUT 20
 
-static const char * version_str = "1.16 20180219";
+static const char * version_str = "1.18 20180628";
 
 static struct option long_options[] = {
     {"count", required_argument, 0, 'c'},
     {"ck_cond", no_argument, 0, 'C'},
+    {"ck-cond", no_argument, 0, 'C'},
     {"extended", no_argument, 0, 'e'},
     {"feature", required_argument, 0, 'f'},
     {"help", no_argument, 0, 'h'},
@@ -304,7 +307,10 @@ main(int argc, char * argv[])
     bool ck_cond = false;
     bool extend = false;
     bool rdonly = false;
-    int sg_fd, c, ret, res;
+    bool verbose_given = false;
+    bool version_given = false;
+    int c, ret, res;
+    int sg_fd = -1;
     int count = 0;
     int feature = 0;
     int verbose = 0;
@@ -363,11 +369,12 @@ main(int argc, char * argv[])
             rdonly = true;
             break;
         case 'v':
+            verbose_given = true;
             ++verbose;
             break;
         case 'V':
-            pr2serr("version: %s\n", version_str);
-            return 0;
+            version_given = true;
+            break;
         default:
             pr2serr("unrecognised option code 0x%x ??\n", c);
             usage();
@@ -387,8 +394,29 @@ main(int argc, char * argv[])
         }
     }
 
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr("version: %s\n", version_str);
+        return 0;
+    }
+
     if (NULL == device_name) {
-        pr2serr("missing device name!\n");
+        pr2serr("Missing device name!\n\n");
         usage();
         return 1;
     }
@@ -409,19 +437,29 @@ main(int argc, char * argv[])
     }
 
     if ((sg_fd = sg_cmds_open_device(device_name, rdonly, verbose)) < 0) {
-        pr2serr("error opening file: %s: %s\n", device_name,
-                safe_strerror(-sg_fd));
-        return SG_LIB_FILE_ERROR;
+        if (verbose)
+            pr2serr("error opening file: %s: %s\n", device_name,
+                    safe_strerror(-sg_fd));
+        ret = sg_convert_errno(-sg_fd);
+        goto fini;
     }
 
     ret = do_set_features(sg_fd, feature, count, lba, cdb_len, ck_cond,
                           extend, verbose);
 
-    res = sg_cmds_close_device(sg_fd);
-    if (res < 0) {
-        pr2serr("close error: %s\n", safe_strerror(-res));
-        if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+fini:
+    if (sg_fd >= 0) {
+        res = sg_cmds_close_device(sg_fd);
+        if (res < 0) {
+            pr2serr("close error: %s\n", safe_strerror(-res));
+            if (0 == ret)
+                ret = sg_convert_errno(-res);
+        }
+    }
+    if (0 == verbose) {
+        if (! sg_if_can2stderr("sg_sat_set_feature failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' "
+                    "or '-vv' for more information\n");
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
