@@ -1,9 +1,12 @@
-/* A utility program originally written for the Linux OS SCSI subsystem
-*  Copyright (C) 2003-2018 D. Gilbert
-*  This program is free software; you can redistribute it and/or modify
-*  it under the terms of the GNU General Public License as published by
-*  the Free Software Foundation; either version 2, or (at your option)
-*  any later version.
+/*
+ * A utility program originally written for the Linux OS SCSI subsystem
+ *    Copyright (C) 2003-2021 D. Gilbert
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2, or (at your option)
+ * any later version.
+ *
+ * SPDX-License-Identifier: GPL-2.0-or-later
 
    This program issues the SCSI SEND DIAGNOSTIC command and in one case
    the SCSI RECEIVE DIAGNOSTIC command to list supported diagnostic pages.
@@ -31,7 +34,7 @@
 #include "sg_pr2serr.h"
 
 
-static const char * version_str = "0.59 20180302";
+static const char * version_str = "0.64 20210610";
 
 #define ME "sg_senddiag: "
 
@@ -66,15 +69,16 @@ struct opts_t {
     bool do_pf;
     bool do_raw;
     bool do_uoff;
-    bool do_version;
     bool opt_new;
+    bool verbose_given;
+    bool version_given;
     int do_help;
     int do_hex;
     int maxlen;
     int page_code;
     int do_selftest;
     int timeout;
-    int do_verbose;
+    int verbose;
     const char * device_name;
     const char * raw_arg;
 };
@@ -87,7 +91,7 @@ usage()
            "[--list]\n"
            "                   [--maxlen=LEN] [--page=PG] [--pf] "
            "[--raw=H,H...]\n"
-           "                   [--selftest=ST] [--test] [--timeout=SEC] "
+           "                   [--selftest=ST] [--test] [--timeout=SECS] "
            "[--uoff]\n"
            "                   [--verbose] [--version] [DEVICE]\n"
            "  where:\n"
@@ -117,7 +121,7 @@ usage()
            "                           5->foreground short, 6->foreground "
            "extended\n"
            "    --test|-t       default self-test\n"
-           "    --timeout=SEC|-T SEC    timeout for foreground self tests\n"
+           "    --timeout=SECS|-T SECS    timeout for foreground self tests\n"
            "                            unit: second (def: 7200 seconds)\n"
            "    --uoff|-u       unit offline (def: 0, only with '--test')\n"
            "    --verbose|-v    increase verbosity\n"
@@ -133,7 +137,7 @@ usage_old()
 {
     printf("Usage: sg_senddiag [-doff] [-e] [-h] [-H] [-l] [-pf]"
            " [-raw=H,H...]\n"
-           "                   [-s=SF] [-t] [-T=SEC] [-uoff] [-v] [-V] "
+           "                   [-s=SF] [-t] [-T=SECS] [-uoff] [-v] [-V] "
            "[DEVICE]\n"
            "  where:\n"
            "    -doff   device online (def: 0, only with '-t')\n"
@@ -151,7 +155,7 @@ usage_old()
            " 4->abort test\n"
            "            5->foreground short, 6->foreground extended\n"
            "    -t      default self-test\n"
-           "    -T SEC    timeout for foreground self tests\n"
+           "    -T SECS    timeout for foreground self tests\n"
            "    -uoff   unit offline (def: 0, only with '-t')\n"
            "    -v      increase verbosity (print issued SCSI cmds)\n"
            "    -V      output version string\n"
@@ -237,7 +241,7 @@ new_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
         case 'T':
             n = sg_get_num(optarg);
             if (n < 0) {
-                pr2serr("bad argument to '--timeout=SEC'\n");
+                pr2serr("bad argument to '--timeout=SECS'\n");
                 return SG_LIB_SYNTAX_ERROR;
             }
             op->timeout = n;
@@ -246,10 +250,11 @@ new_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
             op->do_uoff = true;
             break;
         case 'v':
-            ++op->do_verbose;
+            op->verbose_given = true;
+            ++op->verbose;
             break;
         case 'V':
-            op->do_version = true;
+            op->version_given = true;
             break;
         default:
             pr2serr("unrecognised option code %c [0x%x]\n", c, c);
@@ -333,10 +338,11 @@ old_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
                         jmp_out = true;
                     break;
                 case 'v':
-                    ++op->do_verbose;
+                    op->verbose_given = true;
+                    ++op->verbose;
                     break;
                 case 'V':
-                    op->do_version = true;
+                    op->version_given = true;
                     break;
                 case '?':
                     ++op->do_help;
@@ -364,7 +370,7 @@ old_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
             } else if (0 == strncmp("T=", cp, 2)) {
                 num = sscanf(cp + 2, "%d", &n);
                 if ((1 != num) || (n < 0)) {
-                    printf("Bad page code after '-T=SEC' option\n");
+                    printf("Bad page code after '-T=SECS' option\n");
                     usage_old();
                     return SG_LIB_SYNTAX_ERROR;
                 }
@@ -505,7 +511,7 @@ build_diag_page(const char * inp, uint8_t * mp_arr, int * mp_arr_len,
                 continue;
             }
             if (carry_over[0]) {
-                if (isxdigit(line[0])) {
+                if (isxdigit((uint8_t)line[0])) {
                     carry_over[1] = line[0];
                     carry_over[2] = '\0';
                     if (1 == sscanf(carry_over, "%x", &h))
@@ -641,7 +647,7 @@ static const char *
 find_page_code_desc(int page_num)
 {
     int k;
-    int num = sizeof(pc_desc_arr) / sizeof(pc_desc_arr[0]);
+    int num = SG_ARRAY_SIZE(pc_desc_arr);
     const struct page_code_desc * pcdp = &pc_desc_arr[0];
 
     for (k = 0; k < num; ++k, ++pcdp) {
@@ -657,7 +663,7 @@ static void
 list_page_codes()
 {
     int k;
-    int num = sizeof(pc_desc_arr) / sizeof(pc_desc_arr[0]);
+    int num = SG_ARRAY_SIZE(pc_desc_arr);
     const struct page_code_desc * pcdp = &pc_desc_arr[0];
 
     printf("Page_Code  Description\n");
@@ -670,7 +676,8 @@ list_page_codes()
 int
 main(int argc, char * argv[])
 {
-    int sg_fd, k, num, rsp_len, res, rsp_buff_size, pg, bd_len, resid, vb;
+    int k, num, rsp_len, res, rsp_buff_size, pg, bd_len, resid, vb;
+    int sg_fd = -1;
     int read_in_len = 0;
     int ret = 0;
     struct opts_t opts;
@@ -695,10 +702,27 @@ main(int argc, char * argv[])
             usage_old();
         return 0;
     }
-    if (op->do_version) {
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (op->verbose_given && op->version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        op->verbose_given = false;
+        op->version_given = false;
+        op->verbose = 0;
+    } else if (! op->verbose_given) {
+        pr2serr("set '-vv'\n");
+        op->verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", op->verbose);
+#else
+    if (op->verbose_given && op->version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (op->version_given) {
         pr2serr("Version string: %s\n", version_str);
         return 0;
     }
+
     rsp_buff_size = op->maxlen;
 
     if (NULL == op->device_name) {
@@ -706,14 +730,14 @@ main(int argc, char * argv[])
             list_page_codes();
             return 0;
         }
-        pr2serr("No DEVICE argument given\n");
+        pr2serr("No DEVICE argument given\n\n");
         if (op->opt_new)
             usage();
         else
             usage_old();
         return SG_LIB_SYNTAX_ERROR;
     }
-    vb = op->do_verbose;
+    vb = op->verbose;
     if (op->do_raw) {
         read_in = sg_memalign(op->maxlen, 0, &free_read_in, vb > 3);
         if (NULL == read_in) {
@@ -741,7 +765,7 @@ main(int argc, char * argv[])
             printf("setting -doff or -uoff only useful when -t is set\n");
             usage_old();
         }
-        ret = SG_LIB_SYNTAX_ERROR;
+        ret = SG_LIB_CONTRADICT;
         goto fini;
     }
     if ((op->do_selftest > 0) && op->do_deftest) {
@@ -752,7 +776,7 @@ main(int argc, char * argv[])
             printf("either set -s=SF or -t (not both)\n");
             usage_old();
         }
-        ret = SG_LIB_SYNTAX_ERROR;
+        ret = SG_LIB_CONTRADICT;
         goto fini;
     }
     if (op->do_raw) {
@@ -767,7 +791,7 @@ main(int argc, char * argv[])
                        "'-l'\n");
                 usage_old();
             }
-            ret = SG_LIB_SYNTAX_ERROR;
+            ret = SG_LIB_CONTRADICT;
             goto fini;
         }
         if (! op->do_pf) {
@@ -791,8 +815,9 @@ main(int argc, char * argv[])
 
     if ((sg_fd = sg_cmds_open_device(op->device_name, false /* rw */, vb)) <
          0) {
-        pr2serr(ME "error opening file: %s: %s\n", op->device_name,
-                safe_strerror(-sg_fd));
+        if (vb)
+            pr2serr(ME "error opening file: %s: %s\n", op->device_name,
+                    safe_strerror(-sg_fd));
         ret = sg_convert_errno(-sg_fd);
         goto fini;
     }
@@ -858,9 +883,12 @@ main(int argc, char * argv[])
                         hex2stdout(rsp_buff, rsp_len, 1);
                     else {
                         for (k = 0; k < (rsp_len - 4); ++k) {
-                            cp = find_page_code_desc(rsp_buff[k + 4]);
-                            printf("  0x%02x  %s\n", rsp_buff[k + 4],
-                                   (cp ? cp : "<unknown>"));
+                            pg = rsp_buff[k + 4];
+                            cp = find_page_code_desc(pg);
+                            if (NULL == cp)
+                                cp = (pg < 0x80) ? "<unknown>" :
+                                                   "<vendor specific>";
+                            printf("  0x%02x  %s\n", pg, cp);
                         }
                     }
                 } else {
@@ -917,13 +945,20 @@ err_out9:
     if (vb < 2)
         pr2serr("  try again with '-vv' for more information\n");
 close_fini:
-    res = sg_cmds_close_device(sg_fd);
-    if ((res < 0) && (0 == ret))
-        ret = SG_LIB_FILE_ERROR;
+    if (sg_fd >= 0) {
+        res = sg_cmds_close_device(sg_fd);
+        if (0 == ret)
+            ret = sg_convert_errno(-res);
+    }
 fini:
     if (free_read_in)
         free(free_read_in);
     if (free_rsp_buff)
         free(free_rsp_buff);
+    if (0 == vb) {
+        if (! sg_if_can2stderr("sg_senddiag failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' "
+                    "or '-vv' for more information\n");
+    }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
