@@ -1,8 +1,10 @@
 /*
- * Copyright (c) 2014-2018 Douglas Gilbert.
+ * Copyright (c) 2014-2021 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <unistd.h>
@@ -14,6 +16,7 @@
 #include <stdint.h>
 #include <ctype.h>
 #include <string.h>
+#include <errno.h>
 #include <getopt.h>
 #include <sys/stat.h>
 #include <sys/types.h>
@@ -25,6 +28,7 @@
 #endif
 
 #include "sg_lib.h"
+#include "sg_lib_data.h"
 #include "sg_cmds_basic.h"
 #include "sg_cmds_extra.h"
 #include "sg_unaligned.h"
@@ -41,7 +45,7 @@
  * RESULTS commands in order to send microcode to the given SES device.
  */
 
-static const char * version_str = "1.13 20180219";    /* ses4r02 */
+static const char * version_str = "1.19 20210610";    /* ses4r02 */
 
 #define ME "sg_ses_microcode: "
 #define MAX_XFER_LEN (128 * 1024 * 1024)
@@ -117,13 +121,10 @@ static struct mode_s mode_arr[] = {
     {NULL, 0, NULL},
 };
 
-struct diag_page_code {
-    int page_code;
-    const char * desc;
-};
-
-/* An array of Download microcode status field values and descriptions */
-static struct diag_page_code mc_status_arr[] = {
+/* An array of Download microcode status field values and descriptions.
+ * This table is a subset of one in sg_read_buffer for the read microcode
+ * status page. */
+static struct sg_lib_simple_value_name_t mc_status_arr[] = {
     {0x0, "No download microcode operation in progress"},
     {0x1, "Download in progress, awaiting more"},
     {0x2, "Download complete, updating storage"},
@@ -238,11 +239,11 @@ print_modes(void)
 static const char *
 get_mc_status_str(uint8_t status_val)
 {
-    const struct diag_page_code * mcsp;
+    const struct sg_lib_simple_value_name_t * mcsp;
 
-    for (mcsp = mc_status_arr; mcsp->desc; ++mcsp) {
-        if (status_val == mcsp->page_code)
-            return mcsp->desc;
+    for (mcsp = mc_status_arr; mcsp->name; ++mcsp) {
+        if (status_val == mcsp->value)
+            return mcsp->name;
     }
     return "";
 }
@@ -480,7 +481,10 @@ main(int argc, char * argv[])
 {
     bool last, got_stdin, is_reg;
     bool want_file = false;
-    int sg_fd, res, c, len, k, n, rsp_len, resid, act_len, din_len, verb;
+    bool verbose_given = false;
+    bool version_given = false;
+    int res, c, len, k, n, rsp_len, resid, act_len, din_len, verb;
+    int sg_fd = -1;
     int infd = -1;
     int do_help = 0;
     int ret = 0;
@@ -553,7 +557,7 @@ main(int argc, char * argv[])
              op->mc_len_given = true;
              break;
         case 'm':
-            if (isdigit(*optarg)) {
+            if (isdigit((uint8_t)*optarg)) {
                 op->mc_mode = sg_get_num_nomult(optarg);
                 if ((op->mc_mode < 0) || (op->mc_mode > 255)) {
                     pr2serr("argument to '--mode' should be in the range 0 "
@@ -610,11 +614,12 @@ main(int argc, char * argv[])
             }
             break;
         case 'v':
+            verbose_given = true;
             ++op->verbose;
             break;
         case 'V':
-            pr2serr(ME "version: %s\n", version_str);
-            return 0;
+            version_given = true;
+            break;
         default:
             pr2serr("unrecognised option code 0x%x ??\n", c);
             usage();
@@ -643,6 +648,27 @@ main(int argc, char * argv[])
         }
     }
 
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        op->verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        op->verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", op->verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr(ME "version: %s\n", version_str);
+        return 0;
+    }
+
     if (NULL == device_name) {
         pr2serr("missing device name!\n\n");
         usage();
@@ -686,9 +712,11 @@ main(int argc, char * argv[])
 
     sg_fd = sg_cmds_open_device(device_name, false /* rw */, op->verbose);
     if (sg_fd < 0) {
-        pr2serr(ME "open error: %s: %s\n", device_name,
-                safe_strerror(-sg_fd));
-        return SG_LIB_FILE_ERROR;
+        if (op->verbose)
+            pr2serr(ME "open error: %s: %s\n", device_name,
+                    safe_strerror(-sg_fd));
+        ret = sg_convert_errno(-sg_fd);
+        goto fini;
     }
 
     if (file_name && (! want_file))
@@ -699,10 +727,10 @@ main(int argc, char * argv[])
             infd = STDIN_FILENO;
         else {
             if ((infd = open(file_name, O_RDONLY)) < 0) {
+                ret = sg_convert_errno(errno);
                 snprintf(ebuff, EBUFF_SZ,
                          ME "could not open %s for reading", file_name);
                 perror(ebuff);
-                ret = SG_LIB_FILE_ERROR;
                 goto fini;
             } else if (sg_set_binary_mode(infd) < 0)
                 perror("sg_set_binary_mode");
@@ -747,19 +775,19 @@ main(int argc, char * argv[])
                 goto fini;
             }
             if (lseek(infd, op->mc_skip, SEEK_SET) < 0) {
+                ret = sg_convert_errno(errno);
                 snprintf(ebuff,  EBUFF_SZ, ME "couldn't skip to "
                          "required position on %s", file_name);
                 perror(ebuff);
-                ret = SG_LIB_FILE_ERROR;
                 goto fini;
             }
         }
         res = read(infd, dmp, op->mc_len);
         if (res < 0) {
+            ret = sg_convert_errno(errno);
             snprintf(ebuff, EBUFF_SZ, ME "couldn't read from %s",
                      file_name);
             perror(ebuff);
-            ret = SG_LIB_FILE_ERROR;
             goto fini;
         }
         if (res < op->mc_len) {
@@ -786,7 +814,7 @@ main(int argc, char * argv[])
         infd = -1;
     } else if (want_file) {
         pr2serr("need --in=FILE option with given mode\n");
-        ret = SG_LIB_SYNTAX_ERROR;
+        ret = SG_LIB_CONTRADICT;
         goto fini;
     }
     if (op->mc_tlen < op->mc_len)
@@ -896,21 +924,18 @@ fini:
         free(dout.free_doutp);
     if (free_dip)
         free(free_dip);
-    res = sg_cmds_close_device(sg_fd);
-    if (res < 0) {
-        pr2serr("close error: %s\n", safe_strerror(-res));
-        if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+    if (sg_fd >= 0) {
+        res = sg_cmds_close_device(sg_fd);
+        if (res < 0) {
+            pr2serr("close error: %s\n", safe_strerror(-res));
+            if (0 == ret)
+                ret = sg_convert_errno(-res);
+        }
     }
-    if (ret && (0 == op->verbose)) {
-        if (SG_LIB_CAT_INVALID_OP == ret)
-            pr2serr("%sRECEIVE DIAGNOSTIC RESULTS command not supported\n",
-                    ((MODE_DNLD_STATUS == op->mc_mode) ?
-                     "" : "SEND DIAGNOSTIC or "));
-        else if (ret > 0)
-            pr2serr("Failed, exit status %d\n", ret);
-        else if (ret < 0)
-            pr2serr("Some error occurred\n");
+    if (0 == op->verbose) {
+        if (! sg_if_can2stderr("sg_ses_microcode failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' "
+                    "or '-vv' for more information\n");
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
