@@ -1,12 +1,14 @@
 /*
- *  Copyright (C) 1999-2018 D. Gilbert
+ *  Copyright (C) 1999-2020 D. Gilbert
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
  *  the Free Software Foundation; either version 2, or (at your option)
  *  any later version.
+ *
+ * SPDX-License-Identifier: GPL-2.0-or-later
 
-    Start/Stop parameter by Kurt Garloff <garloff at suse dot de>, 6/2000
-    Sync cache parameter by Kurt Garloff <garloff at suse dot de>, 1/2001
+    Start/Stop parameter by Kurt Garloff, 6/2000
+    Sync cache parameter by Kurt Garloff, 1/2001
     Guard block device answering sg's ioctls.
                      <dgilbert at interlog dot com> 12/2002
     Convert to SG_IO ioctl so can use sg or block devices in 2.6.* 3/2003
@@ -34,7 +36,7 @@
 #include "sg_pr2serr.h"
 
 
-static const char * version_str = "0.64 20180302";  /* sbc3r14; mmc6r01a */
+static const char * version_str = "0.67 20200930";  /* sbc3r14; mmc6r01a */
 
 static struct option long_options[] = {
         {"eject", no_argument, 0, 'e'},
@@ -65,13 +67,14 @@ struct opts_t {
     bool do_readonly;
     bool do_start;
     bool do_stop;
-    bool do_version;
     bool opt_new;
+    bool verbose_given;
+    bool version_given;
     int do_fl;
     int do_help;
     int do_mod;
     int do_pc;
-    int do_verbose;
+    int verbose;
     const char * device_name;
 };
 
@@ -159,7 +162,7 @@ usage_old()
             "action\n"
             "    --stop     stop unit (same as '0')\n"
             "    -v         verbose (print out SCSI commands)\n"
-            "    -N|--new   use new interface\n"
+            "    --new|-N   use new interface\n"
             "    -V         print version string then exit\n\n"
             "    Example: 'sg_start --stop /dev/sdb'    stops unit\n"
             "             'sg_start --eject /dev/scd0'  stops unit and "
@@ -169,7 +172,7 @@ usage_old()
 }
 
 static int
-news_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
+new_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
 {
     int c, n, err;
 
@@ -247,10 +250,11 @@ news_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
             op->do_stop = true;
             break;
         case 'v':
-            ++op->do_verbose;
+            op->verbose_given = true;
+            ++op->verbose;
             break;
         case 'V':
-            op->do_version = true;
+            op->version_given = true;
             break;
         default:
             pr2serr("unrecognised option code %c [0x%x]\n", c, c);
@@ -315,10 +319,11 @@ old_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
                     op->do_readonly = true;
                     break;
                 case 'v':
-                    ++op->do_verbose;
+                    op->verbose_given = true;
+                    ++op->verbose;
                     break;
                 case 'V':
-                    op->do_version = true;
+                    op->version_given = true;
                     break;
                 case 'h':
                 case '?':
@@ -446,7 +451,7 @@ old_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
             pr2serr("please, only one of 0, 1, --eject, "
                     "--load, --start or --stop\n");
             usage_old();
-            return SG_LIB_SYNTAX_ERROR;
+            return SG_LIB_CONTRADICT;
         } else if (startstop_set) {
             if (startstop)
                 op->do_start = true;
@@ -468,10 +473,10 @@ parse_cmd_line(struct opts_t * op, int argc, char * argv[])
         op->opt_new = false;
         res = old_parse_cmd_line(op, argc, argv);
         if ((0 == res) && op->opt_new)
-            res = news_parse_cmd_line(op, argc, argv);
+            res = new_parse_cmd_line(op, argc, argv);
     } else {
         op->opt_new = true;
-        res = news_parse_cmd_line(op, argc, argv);
+        res = new_parse_cmd_line(op, argc, argv);
         if ((0 == res) && (! op->opt_new))
             res = old_parse_cmd_line(op, argc, argv);
     }
@@ -482,7 +487,8 @@ parse_cmd_line(struct opts_t * op, int argc, char * argv[])
 int
 main(int argc, char * argv[])
 {
-    int fd, res;
+    int res;
+    int sg_fd = -1;
     int ret = 0;
     struct opts_t opts;
     struct opts_t * op;
@@ -492,7 +498,7 @@ main(int argc, char * argv[])
     op->do_fl = -1;    /* only when >= 0 set FL bit */
     res = parse_cmd_line(op, argc, argv);
     if (res)
-        return SG_LIB_SYNTAX_ERROR;
+        return res;
     if (op->do_help) {
         if (op->opt_new)
             usage();
@@ -500,18 +506,35 @@ main(int argc, char * argv[])
             usage_old();
         return 0;
     }
-    if (op->do_version) {
+
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (op->verbose_given && op->version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        op->verbose_given = false;
+        op->version_given = false;
+        op->verbose = 0;
+    } else if (! op->verbose_given) {
+        pr2serr("set '-vv'\n");
+        op->verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", op->verbose);
+#else
+    if (op->verbose_given && op->version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (op->version_given) {
         pr2serr("Version string: %s\n", version_str);
         return 0;
     }
 
     if (op->do_start && op->do_stop) {
         pr2serr("Ambiguous to give both '--start' and '--stop'\n");
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
     if (op->do_load && op->do_eject) {
         pr2serr("Ambiguous to give both '--load' and '--eject'\n");
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
     if (op->do_load)
        op->do_start = true;
@@ -536,48 +559,60 @@ main(int argc, char * argv[])
         if (! op->do_start) {
             pr2serr("Giving '--fl=FL' with '--stop' (or '--eject') is "
                     "invalid\n");
-            return SG_LIB_SYNTAX_ERROR;
+            return SG_LIB_CONTRADICT;
         }
         if (op->do_pc > 0) {
             pr2serr("Giving '--fl=FL' with '--pc=PC' when PC is non-zero "
                     "is invalid\n");
-            return SG_LIB_SYNTAX_ERROR;
+            return SG_LIB_CONTRADICT;
         }
     }
 
-    fd = sg_cmds_open_device(op->device_name, op->do_readonly,
-                             op->do_verbose);
-    if (fd < 0) {
-        pr2serr("Error trying to open %s: %s\n", op->device_name,
-                safe_strerror(-fd));
-        return SG_LIB_FILE_ERROR;
+    sg_fd = sg_cmds_open_device(op->device_name, op->do_readonly,
+                                op->verbose);
+    if (sg_fd < 0) {
+        if (op->verbose)
+            pr2serr("Error trying to open %s: %s\n", op->device_name,
+                    safe_strerror(-sg_fd));
+        ret = sg_convert_errno(-sg_fd);
+        goto fini;
     }
 
     if (op->do_fl >= 0)
-        res = sg_ll_start_stop_unit(fd, op->do_immed, op->do_fl, 0 /* pc */,
+        res = sg_ll_start_stop_unit(sg_fd, op->do_immed, op->do_fl, 0 /* pc */,
                                     true /* fl */, true /* loej */,
                                     true /*start */, true /* noisy */,
-                                    op->do_verbose);
+                                    op->verbose);
     else if (op->do_pc > 0)
-        res = sg_ll_start_stop_unit(fd, op->do_immed, op->do_mod,
+        res = sg_ll_start_stop_unit(sg_fd, op->do_immed, op->do_mod,
                                     op->do_pc, op->do_noflush, false, false,
-                                    true, op->do_verbose);
+                                    true, op->verbose);
     else
-        res = sg_ll_start_stop_unit(fd, op->do_immed, 0, false,
+        res = sg_ll_start_stop_unit(sg_fd, op->do_immed, 0, false,
                                     op->do_noflush, op->do_loej,
-                                    op->do_start, true, op->do_verbose);
+                                    op->do_start, true, op->verbose);
     ret = res;
     if (res) {
-        if (op->do_verbose < 2) {
+        if (op->verbose < 2) {
             char b[80];
 
-            sg_get_category_sense_str(res, sizeof(b), b, op->do_verbose);
+            sg_get_category_sense_str(res, sizeof(b), b, op->verbose);
             pr2serr("%s\n", b);
         }
         pr2serr("START STOP UNIT command failed\n");
     }
-    res = sg_cmds_close_device(fd);
-    if ((res < 0) && (0 == ret))
-        return SG_LIB_FILE_ERROR;
+fini:
+    if (sg_fd >= 0) {
+        res = sg_cmds_close_device(sg_fd);
+        if (res < 0) {
+            if (0 == ret)
+                ret = sg_convert_errno(-res);
+        }
+    }
+    if (0 == op->verbose) {
+        if (! sg_if_can2stderr("sg_start failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' "
+                    "or '-vv' for more information\n");
+    }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
