@@ -1,8 +1,10 @@
 /*
- * Copyright (c) 2004-2018 Douglas Gilbert.
+ * Copyright (c) 2004-2021 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <unistd.h>
@@ -35,7 +37,7 @@
  * (e.g. disks).
  */
 
-static const char * version_str = "1.20 20180219";
+static const char * version_str = "1.26 20210830";
 
 #define SYNCHRONIZE_CACHE16_CMD     0x91
 #define SYNCHRONIZE_CACHE16_CMDLEN  16
@@ -61,7 +63,7 @@ static void usage()
 {
     pr2serr("Usage: sg_sync    [--16] [--count=COUNT] [--group=GN] [--help] "
             "[--immed]\n"
-            "                  [--lba=LBA] [--sync-nv] [--timeout=SEC] "
+            "                  [--lba=LBA] [--sync-nv] [--timeout=SECS] "
             "[--verbose]\n"
             "                  [--version] DEVICE\n"
             "  where:\n"
@@ -82,7 +84,7 @@ static void usage()
             "    --sync-nv|-s        synchronize to non-volatile storage "
             "(if distinct\n"
             "                        from medium). Obsolete in sbc3r35d.\n"
-            "    --timeout=SEC|-t SEC    command timeout in seconds, only "
+            "    --timeout=SECS|-t SECS    command timeout in seconds, only "
             "active\n"
             "                              if '--16' given (def: 60 seconds)\n"
             "    --verbose|-v        increase verbosity\n"
@@ -95,7 +97,7 @@ sg_ll_sync_cache_16(int sg_fd, bool sync_nv, bool immed, int group,
                     uint64_t lba, unsigned int num_lb, int to_secs,
                     bool noisy, int verbose)
 {
-    int res, ret, k, sense_cat;
+    int res, ret, sense_cat;
     uint8_t sc_cdb[SYNCHRONIZE_CACHE16_CMDLEN] =
                 {SYNCHRONIZE_CACHE16_CMD, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                  0, 0, 0, 0, 0, 0};
@@ -111,10 +113,11 @@ sg_ll_sync_cache_16(int sg_fd, bool sync_nv, bool immed, int group,
     sg_put_unaligned_be32((uint32_t)num_lb, sc_cdb + 10);
 
     if (verbose) {
-        pr2serr("    synchronize cache(16) cdb: ");
-        for (k = 0; k < SYNCHRONIZE_CACHE16_CMDLEN; ++k)
-            pr2serr("%02x ", sc_cdb[k]);
-        pr2serr("\n");
+        char b[128];
+
+        pr2serr("    Synchronize cache(16) cdb: %s\n",
+                sg_get_command_str(sc_cdb, SYNCHRONIZE_CACHE16_CMDLEN, false,
+                                   sizeof(b), b));
     }
     ptvp = construct_scsi_pt_obj();
     if (NULL == ptvp) {
@@ -125,11 +128,13 @@ sg_ll_sync_cache_16(int sg_fd, bool sync_nv, bool immed, int group,
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     res = do_scsi_pt(ptvp, sg_fd, to_secs, verbose);
     ret = sg_cmds_process_resp(ptvp, "synchronize cache(16)", res,
-                               SG_NO_DATA_IN, sense_b, noisy, verbose,
-                               &sense_cat);
-    if (-1 == ret)
-        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
-    else if (-2 == ret) {
+                               noisy, verbose, &sense_cat);
+    if (-1 == ret) {
+        if (get_scsi_pt_transport_err(ptvp))
+            ret = SG_LIB_TRANSPORT_ERROR;
+        else
+            ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
+    } else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
         case SG_LIB_CAT_NO_SENSE:
@@ -152,7 +157,10 @@ int main(int argc, char * argv[])
     bool do_16 = false;
     bool immed = false;
     bool sync_nv = false;
-    int sg_fd, res, c;
+    bool verbose_given = false;
+    bool version_given = false;
+    int res, c;
+    int sg_fd = -1;
     int group = 0;
     int ret = 0;
     int to_secs = DEF_PT_TIMEOUT;
@@ -214,11 +222,12 @@ int main(int argc, char * argv[])
             }
             break;
         case 'v':
+            verbose_given = true;
             ++verbose;
             break;
         case 'V':
-            pr2serr("version: %s\n", version_str);
-            return 0;
+            version_given = true;
+            break;
         default:
             pr2serr("unrecognised option code 0x%x ??\n", c);
             usage();
@@ -238,16 +247,39 @@ int main(int argc, char * argv[])
         }
     }
 
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr("version: %s\n", version_str);
+        return 0;
+    }
+
     if (NULL == device_name) {
-        pr2serr("missing device name!\n");
+        pr2serr("Missing device name!\n\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
     sg_fd = sg_cmds_open_device(device_name, false /* rw */, verbose);
     if (sg_fd < 0) {
-        pr2serr("open error: %s: %s\n", device_name,
-                safe_strerror(-sg_fd));
-        return SG_LIB_FILE_ERROR;
+        if (verbose)
+            pr2serr("open error: %s: %s\n", device_name,
+                    safe_strerror(-sg_fd));
+        ret = sg_convert_errno(-sg_fd);
+        goto fini;
     }
 
     if (do_16)
@@ -264,11 +296,19 @@ int main(int argc, char * argv[])
         pr2serr("Synchronize cache failed: %s\n", b);
     }
 
-    res = sg_cmds_close_device(sg_fd);
-    if (res < 0) {
-        pr2serr("close error: %s\n", safe_strerror(-res));
-        if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+fini:
+    if (sg_fd >= 0) {
+        res = sg_cmds_close_device(sg_fd);
+        if (res < 0) {
+            pr2serr("close error: %s\n", safe_strerror(-res));
+            if (0 == ret)
+                ret = sg_convert_errno(-res);
+        }
+    }
+    if (0 == verbose) {
+        if (! sg_if_can2stderr("sg_sync failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' "
+                    "or '-vv' for more information\n");
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
