@@ -1,7 +1,7 @@
 /*
- * (c) 2000 Kurt Garloff <garloff at suse dot de>
+ * (c) 2000 Kurt Garloff
  * heavily based on Douglas Gilbert's sg_rbuf program.
- * (c) 1999-2018 Douglas Gilbert
+ * (c) 1999-2019 Douglas Gilbert
  *
  * Program to test the SCSI host adapter by issuing
  * write and read operations on a device's buffer
@@ -16,6 +16,8 @@
  *  the Free Software Foundation; either version 2, or (at your option)
  *  any later version.
  *
+ * SPDX-License-Identifier: GPL-2.0-or-later
+ *
  * $Id: sg_test_rwbuf.c,v 1.1 2000/03/02 13:50:03 garloff Exp $
  *
  *   2003/11/11  switch sg3_utils version to use SG_IO ioctl [dpg]
@@ -29,6 +31,7 @@
 #include <stdarg.h>
 #include <stdbool.h>
 #include <string.h>
+#include <errno.h>
 #include <getopt.h>
 #include <sys/ioctl.h>
 #include <sys/types.h>
@@ -44,7 +47,7 @@
 #include "sg_pr2serr.h"
 
 
-static const char * version_str = "1.15 20180219";
+static const char * version_str = "1.20 20191220";
 
 #define BPI (signed)(sizeof(int))
 
@@ -64,6 +67,7 @@ static int base = 0x12345678;
 static int buf_capacity = 0;
 static int buf_granul = 255;
 static uint8_t *cmpbuf = NULL;
+static uint8_t *free_cmpbuf = NULL;
 
 
 /* Options */
@@ -92,7 +96,7 @@ find_out_about_buffer(int sg_fd)
         uint8_t rbBuff[RB_DESC_LEN];
         uint8_t sense_buffer[32];
         struct sg_io_hdr io_hdr;
-        int k, res;
+        int res;
 
         rb_cdb[1] = RB_MODE_DESC;
         rb_cdb[8] = RB_DESC_LEN;
@@ -108,10 +112,11 @@ find_out_about_buffer(int sg_fd)
         io_hdr.timeout = 60000;     /* 60000 millisecs == 60 seconds */
 
         if (verbose) {
-                pr2serr("    read buffer [mode desc] cdb: ");
-                for (k = 0; k < (int)sizeof(rb_cdb); ++k)
-                        pr2serr("%02x ", rb_cdb[k]);
-                pr2serr("\n");
+                char b[128];
+
+                pr2serr("    read buffer [mode desc] cdb: %s\n",
+                        sg_get_command_str(rb_cdb, (int)sizeof(rb_cdb), false,
+                                           sizeof(b), b));
         }
         if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
                 perror(ME "SG_IO READ BUFFER descriptor error");
@@ -226,10 +231,12 @@ void do_fill_buffer (int *buf, int len)
 
 int read_buffer (int sg_fd, unsigned ssize)
 {
-        int res, k;
+        int res;
         uint8_t rb_cdb[] = {READ_BUFFER, 0, 0, 0, 0, 0, 0, 0, 0, 0};
         int bufSize = ssize + addread;
-        uint8_t * rbBuff = (uint8_t *)malloc(bufSize);
+        uint8_t * free_rbBuff = NULL;
+        uint8_t * rbBuff = (uint8_t *)sg_memalign(bufSize, 0, &free_rbBuff,
+                                                  false);
         uint8_t sense_buffer[32];
         struct sg_io_hdr io_hdr;
 
@@ -249,10 +256,11 @@ int read_buffer (int sg_fd, unsigned ssize)
         io_hdr.pack_id = 2;
         io_hdr.timeout = 60000;     /* 60000 millisecs == 60 seconds */
         if (verbose) {
-                pr2serr("    read buffer [mode data] cdb: ");
-                for (k = 0; k < (int)sizeof(rb_cdb); ++k)
-                        pr2serr("%02x ", rb_cdb[k]);
-                pr2serr("\n");
+                char b[128];
+
+                pr2serr("    read buffer [mode data] cdb: %s\n",
+                        sg_get_command_str(rb_cdb, (int)sizeof(rb_cdb), false,
+                                           sizeof(b), b));
         }
 
         if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
@@ -280,7 +288,8 @@ int read_buffer (int sg_fd, unsigned ssize)
         }
 
         res = do_checksum((int*)rbBuff, ssize, false);
-        free(rbBuff);
+        if (free_rbBuff)
+                free(free_rbBuff);
         return res;
 }
 
@@ -288,10 +297,12 @@ int write_buffer (int sg_fd, unsigned ssize)
 {
         uint8_t wb_cdb[] = {WRITE_BUFFER, 0, 0, 0, 0, 0, 0, 0, 0, 0};
         int bufSize = ssize + addwrite;
-        uint8_t * wbBuff = (uint8_t *)malloc(bufSize);
+        uint8_t * free_wbBuff = NULL;
+        uint8_t * wbBuff = (uint8_t *)sg_memalign(bufSize, 0, &free_wbBuff,
+                                                  false);
         uint8_t sense_buffer[32];
         struct sg_io_hdr io_hdr;
-        int k, res;
+        int res;
 
         if (NULL == wbBuff)
                 return -1;
@@ -311,10 +322,11 @@ int write_buffer (int sg_fd, unsigned ssize)
         io_hdr.pack_id = 1;
         io_hdr.timeout = 60000;     /* 60000 millisecs == 60 seconds */
         if (verbose) {
-                pr2serr("    write buffer [mode data] cdb: ");
-                for (k = 0; k < (int)sizeof(wb_cdb); ++k)
-                        pr2serr("%02x ", wb_cdb[k]);
-                pr2serr("\n");
+                char b[128];
+
+                pr2serr("    write buffer [mode data] cdb: %s\n",
+                        sg_get_command_str(wb_cdb, (int)sizeof(wb_cdb), false,
+                                           sizeof(b), b));
         }
 
         if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
@@ -340,7 +352,8 @@ int write_buffer (int sg_fd, unsigned ssize)
                 free(wbBuff);
                 return res;
         }
-        free(wbBuff);
+        if (free_wbBuff)
+                free(free_wbBuff);
         return res;
 }
 
@@ -378,11 +391,14 @@ void usage ()
 
 int main (int argc, char * argv[])
 {
+        bool verbose_given = false;
+        bool version_given = false;
         int sg_fd, res;
         const char * device_name = NULL;
         int times = 1;
         int ret = 0;
         int k = 0;
+        int err;
 
         while (1) {
                 int option_index = 0;
@@ -422,11 +438,12 @@ int main (int argc, char * argv[])
                         }
                         break;
                 case 'v':
+                        verbose_given = true;
                         verbose++;
                         break;
                 case 'V':
-                        pr2serr(ME "version: %s\n", version_str);
-                        return 0;
+                        version_given = true;
+                        break;
                 case 'w':
                         addwrite = sg_get_num(optarg);
                         if (-1 == addwrite) {
@@ -479,6 +496,29 @@ int main (int argc, char * argv[])
                         return SG_LIB_SYNTAX_ERROR;
                 }
         }
+
+#ifdef DEBUG
+        pr2serr("In DEBUG mode, ");
+        if (verbose_given && version_given) {
+                pr2serr("but override: '-vV' given, zero verbose and "
+                        "continue\n");
+                verbose_given = false;
+                version_given = false;
+                verbose = 0;
+        } else if (! verbose_given) {
+                pr2serr("set '-vv'\n");
+                verbose = 2;
+        } else
+                pr2serr("keep verbose=%d\n", verbose);
+#else
+        if (verbose_given && version_given)
+                pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+        if (version_given) {
+                pr2serr(ME "version: %s\n", version_str);
+                return 0;
+        }
+
         if (NULL == device_name) {
                 pr2serr("no device name given\n");
                 usage();
@@ -493,8 +533,9 @@ int main (int argc, char * argv[])
 
         sg_fd = open(device_name, O_RDWR | O_NONBLOCK);
         if (sg_fd < 0) {
+                err = errno;
                 perror("sg_test_rwbuf: open error");
-                return SG_LIB_FILE_ERROR;
+                return sg_convert_errno(err);
         }
         ret = find_out_about_buffer(sg_fd);
         if (ret)
@@ -511,7 +552,7 @@ int main (int argc, char * argv[])
                 goto err_out;
         }
 
-        cmpbuf = (uint8_t *)malloc(size);
+        cmpbuf = (uint8_t *)sg_memalign(size, 0, &free_cmpbuf, false);
         for (k = 0; k < times; ++k) {
                 ret = write_buffer (sg_fd, size);
                 if (ret) {
@@ -526,13 +567,13 @@ int main (int argc, char * argv[])
         }
 
 err_out:
-        if (cmpbuf)
-                free(cmpbuf);
+        if (free_cmpbuf)
+                free(free_cmpbuf);
         res = close(sg_fd);
         if (res < 0) {
                 perror(ME "close error");
                 if (0 == ret)
-                        ret = SG_LIB_FILE_ERROR;
+                        ret = sg_convert_errno(errno);
         }
         if ((0 == ret) && (! do_quick))
                 printf ("Success\n");
