@@ -1,10 +1,11 @@
 /*
- * Copyright (C) 2000-2018 D. Gilbert
+ * Copyright (C) 2000-2021 D. Gilbert
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2, or (at your option)
  * any later version.
  *
+ * SPDX-License-Identifier: GPL-2.0-or-later
  */
 
 /*
@@ -21,8 +22,11 @@
 #include <stdarg.h>
 #include <stdbool.h>
 #include <string.h>
+#include <time.h>
 #include <errno.h>
 #include <getopt.h>
+#define __STDC_FORMAT_MACROS 1
+#include <inttypes.h>
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -41,23 +45,13 @@
 #include "sg_pr2serr.h"
 
 
-static const char * version_str = "3.40 20180321";
-
-#if defined(MSC_VER) || defined(__MINGW32__)
-#define HAVE_MS_SLEEP
-#endif
-
-#ifdef HAVE_MS_SLEEP
-#include <windows.h>
-#define sleep_for(seconds)    Sleep( (seconds) * 1000)
-#else
-#define sleep_for(seconds)    sleep(seconds)
-#endif
+static const char * version_str = "3.49 20210830";
 
 #define DEF_PT_TIMEOUT  60       /* 60 seconds */
 
 
 static struct option long_options[] = {
+        {"delay", required_argument, 0, 'd'},
         {"help", no_argument, 0, 'h'},
         {"low", no_argument, 0, 'l'},
         {"new", no_argument, 0, 'N'},
@@ -73,14 +67,17 @@ static struct option long_options[] = {
 };
 
 struct opts_t {
+    bool delay_given;
     bool do_low;
     bool do_progress;
     bool do_time;
-    bool do_version;
     bool opts_new;
+    bool verbose_given;
+    bool version_given;
+    int delay;
     int do_help;
     int do_number;
-    int do_verbose;
+    int verbose;
     const char * device_name;
 };
 
@@ -94,10 +91,13 @@ struct loop_res_t {
 static void
 usage()
 {
-    printf("Usage: sg_turs [--help] [--low] [--number=NUM] [--num=NUM] "
-           "[--progress]\n"
-           "               [--time] [--verbose] [--version] DEVICE\n"
+    printf("Usage: sg_turs [--delay=MS] [--help] [--low] [--number=NUM] "
+           "[--num=NUM]\n"
+           "               [--progress] [--time] [--verbose] [--version] "
+           "DEVICE\n"
            "  where:\n"
+           "    --delay=MS|-d MS    delay MS miiliseconds before sending "
+           "each tur\n"
            "    --help|-h        print usage message then exit\n"
            "    --low|-l         use low level (sg_pt) interface for "
            "speed\n"
@@ -107,19 +107,23 @@ usage()
            "    --old|-O         use old interface (use as first option)\n"
            "    --progress|-p    outputs progress indication (percentage) "
            "if available\n"
+           "                     waits 30 seconds before TUR unless "
+           "--delay=MS given\n"
            "    --time|-t        outputs total duration and commands per "
            "second\n"
            "    --verbose|-v     increase verbosity\n"
            "    --version|-V     print version string then exit\n\n"
-           "Performs a SCSI TEST UNIT READY command (or many of them).\n");
+           "Performs a SCSI TEST UNIT READY command (or many of them).\n"
+           "This SCSI command is often known by its abbreviation: TUR .\n");
 }
 
 static void
 usage_old()
 {
-    printf("Usage: sg_turs [-l] [-n=NUM] [-p] [-t] [-v] [-V] "
+    printf("Usage: sg_turs [-d=MS] [-l] [-n=NUM] [-p] [-t] [-v] [-V] "
            "DEVICE\n"
            "  where:\n"
+           "    -d=MS     same as --delay=MS in new interface\n"
            "    -l        use low level interface (sg_pt) for speed\n"
            "    -n=NUM    number of test_unit_ready commands "
            "(def: 1)\n"
@@ -150,12 +154,22 @@ new_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
     while (1) {
         int option_index = 0;
 
-        c = getopt_long(argc, argv, "hln:NOptvV", long_options,
+        c = getopt_long(argc, argv, "d:hln:NOptvV", long_options,
                         &option_index);
         if (c == -1)
             break;
 
         switch (c) {
+        case 'd':
+            n = sg_get_num(optarg);
+            if (n < 0) {
+                pr2serr("bad argument to '--delay='\n");
+                usage();
+                return SG_LIB_SYNTAX_ERROR;
+            }
+            op->delay = n;
+            op->delay_given = true;
+            break;
         case 'h':
         case '?':
             ++op->do_help;
@@ -184,10 +198,11 @@ new_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
             op->do_time = true;
             break;
         case 'v':
-            ++op->do_verbose;
+            op->verbose_given = true;
+            ++op->verbose;
             break;
         case 'V':
-            op->do_version = true;
+            op->version_given = true;
             break;
         default:
             pr2serr("unrecognised option code %c [0x%x]\n", c, c);
@@ -242,13 +257,14 @@ old_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
                     op->do_time = true;
                     break;
                 case 'v':
-                    ++op->do_verbose;
+                    op->verbose_given = true;
+                    ++op->verbose;
                     break;
                 case 'V':
-                    op->do_version = true;
+                    op->version_given = true;
                     break;
                 case '?':
-                    usage_old();
+                    ++op->do_help;
                     return 0;
                 default:
                     jmp_out = true;
@@ -259,7 +275,15 @@ old_parse_cmd_line(struct opts_t * op, int argc, char * argv[])
             }
             if (plen <= 0)
                 continue;
-            if (0 == strncmp("n=", cp, 2)) {
+            if (0 == strncmp("d=", cp, 2)) {
+                op->delay = sg_get_num(cp + 2);
+                if (op->delay < 0) {
+                    printf("Couldn't decode number after 'd=' option\n");
+                    usage_old();
+                    return SG_LIB_SYNTAX_ERROR;
+                }
+                op->delay_given = true;
+            } else if (0 == strncmp("n=", cp, 2)) {
                 op->do_number = sg_get_num(cp + 2);
                 if (op->do_number <= 0) {
                     printf("Couldn't decode number after 'n=' option\n");
@@ -306,37 +330,69 @@ parse_cmd_line(struct opts_t * op, int argc, char * argv[])
     return res;
 }
 
+static void
+wait_millisecs(int millisecs)
+{
+    struct timespec wait_period, rem;
+
+    wait_period.tv_sec = millisecs / 1000;
+    wait_period.tv_nsec = (millisecs % 1000) * 1000000;
+    while ((nanosleep(&wait_period, &rem) < 0) && (EINTR == errno))
+                wait_period = rem;
+}
+
+/* Returns true if prints estimate of duration to ready */
+bool
+check_for_lu_becoming(struct sg_pt_base * ptvp)
+{
+    int s_len = get_scsi_pt_sense_len(ptvp);
+    uint64_t info;
+    uint8_t * sense_b = get_scsi_pt_sense_buf(ptvp);
+    struct sg_scsi_sense_hdr ssh;
+
+    /* Check for "LU is in process of becoming ready" with a non-zero INFO
+     * field that isn't too big. As per 20-061r2 it means the following: */
+    if (sg_scsi_normalize_sense(sense_b, s_len, &ssh) && (ssh.asc == 0x4) &&
+        (ssh.ascq == 0x1) && sg_get_sense_info_fld(sense_b, s_len, &info) &&
+        (info > 0x0) && (info < 0x1000000)) {
+        printf("device not ready, estimated to be ready in %" PRIu64
+               " milliseconds\n", info);
+        return true;
+    }
+    return false;
+}
+
 /* Returns number of TURs performed */
 static int
-loop_turs(int sg_fd, struct loop_res_t * resp, struct opts_t * op)
+loop_turs(struct sg_pt_base * ptvp, struct loop_res_t * resp,
+          struct opts_t * op)
 {
     int k, res;
-    int vb = op->do_verbose;
+    int packet_id = 0;
+    int vb = op->verbose;
     char b[80];
+    uint8_t sense_b[32];
 
     if (op->do_low) {
-        int err, rs, n, sense_cat;
-        struct sg_pt_base * pbp;
+        int rs, n, sense_cat;
         uint8_t cdb[6];
-        uint8_t sense_b[32];
 
-        pbp = construct_scsi_pt_obj_with_fd(sg_fd, vb);
-        err = 0;
-        if ((NULL == pbp) || ((err = get_scsi_pt_os_err(pbp)))) {
-            resp->ret = sg_convert_errno(err ? err : ENOMEM);
-            return 0;
-        }
         for (k = 0; k < op->do_number; ++k) {
+            if (op->delay > 0)
+                wait_millisecs(op->delay);
             /* Might get Unit Attention on first invocation */
             memset(cdb, 0, sizeof(cdb));    /* TUR's cdb is 6 zeros */
-            set_scsi_pt_cdb(pbp, cdb, sizeof(cdb));
-            set_scsi_pt_sense(pbp, sense_b, sizeof(sense_b));
-            rs = do_scsi_pt(pbp, -1, DEF_PT_TIMEOUT, vb);
-            n = sg_cmds_process_resp(pbp, "Test unit ready", rs,
-                                     SG_NO_DATA_IN, sense_b,
-                                     (0 == k), vb, &sense_cat);
+            set_scsi_pt_cdb(ptvp, cdb, sizeof(cdb));
+            set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
+            set_scsi_pt_packet_id(ptvp, ++packet_id);
+            rs = do_scsi_pt(ptvp, -1, DEF_PT_TIMEOUT, vb);
+            n = sg_cmds_process_resp(ptvp, "Test unit ready", rs, (0 == k),
+                                     vb, &sense_cat);
             if (-1 == n) {
-                resp->ret = sg_convert_errno(get_scsi_pt_os_err(pbp));
+                if (get_scsi_pt_transport_err(ptvp))
+                    resp->ret = SG_LIB_TRANSPORT_ERROR;
+                else
+                    resp->ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
                 return k;
             } else if (-2 == n) {
                 switch (sense_cat) {
@@ -345,9 +401,10 @@ loop_turs(int sg_fd, struct loop_res_t * resp, struct opts_t * op)
                     break;
                 case SG_LIB_CAT_NOT_READY:
                     ++resp->num_errs;
-                    if (1 ==  op->do_number) {
+                    if ((1 == op->do_number) || (op->delay > 0)) {
+                        if (! check_for_lu_becoming(ptvp))
+                            printf("device not ready\n");
                         resp->ret = sense_cat;
-                        printf("device not ready\n");
                         resp->reported = true;
                     }
                     break;
@@ -370,21 +427,25 @@ loop_turs(int sg_fd, struct loop_res_t * resp, struct opts_t * op)
                     break;
                 }
             }
-            clear_scsi_pt_obj(pbp);
+            partial_clear_scsi_pt_obj(ptvp);
         }
-        destruct_scsi_pt_obj(pbp);
         return k;
     } else {
         for (k = 0; k < op->do_number; ++k) {
+            if (op->delay > 0)
+                wait_millisecs(op->delay);
+            set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
             /* Might get Unit Attention on first invocation */
-            res = sg_ll_test_unit_ready(sg_fd, k, (0 == k), vb);
+            res = sg_ll_test_unit_ready_pt(ptvp, k, (0 == k), vb);
             if (res) {
                 ++resp->num_errs;
                 resp->ret = res;
-                if (1 == op->do_number) {
-                    if (SG_LIB_CAT_NOT_READY == res)
-                        printf("device not ready\n");
-                    else {
+                if ((1 == op->do_number) || (op->delay > 0)) {
+                    if (SG_LIB_CAT_NOT_READY == res) {
+                        if (! check_for_lu_becoming(ptvp))
+                            printf("device not ready\n");
+                        continue;
+                    } else {
                         sg_get_category_sense_str(res, sizeof(b), b, vb);
                         printf("%s\n", b);
                     }
@@ -402,8 +463,10 @@ int
 main(int argc, char * argv[])
 {
     bool start_tm_valid = false;
-    int sg_fd, k, res, progress, pr, rem, num_done;
+    int k, res, progress, pr, rem, num_done;
+    int err = 0;
     int ret = 0;
+    int sg_fd = -1;
     int64_t elapsed_usecs = 0;
 #if defined(HAVE_CLOCK_GETTIME) && defined(CLOCK_MONOTONIC)
     struct timespec start_tm, end_tm;
@@ -412,6 +475,7 @@ main(int argc, char * argv[])
 #endif
     struct loop_res_t loop_res;
     struct loop_res_t * resp = &loop_res;
+    struct sg_pt_base * ptvp = NULL;
     struct opts_t opts;
     struct opts_t * op = &opts;
 
@@ -421,15 +485,33 @@ main(int argc, char * argv[])
     op->do_number = 1;
     res = parse_cmd_line(op, argc, argv);
     if (res)
-        return SG_LIB_SYNTAX_ERROR;
+        return res;
     if (op->do_help) {
         usage_for(op);
         return 0;
     }
-    if (op->do_version) {
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (op->verbose_given && op->version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        op->verbose_given = false;
+        op->version_given = false;
+        op->verbose = 0;
+    } else if (! op->verbose_given) {
+        pr2serr("set '-vv'\n");
+        op->verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", op->verbose);
+#else
+    if (op->verbose_given && op->version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (op->version_given) {
         pr2serr("Version string: %s\n", version_str);
         return 0;
     }
+    if (op->do_progress && (! op->delay_given))
+        op->delay = 30 * 1000;  /* progress has 30 second default delay */
 
     if (NULL == op->device_name) {
         pr2serr("No DEVICE argument given\n");
@@ -438,18 +520,29 @@ main(int argc, char * argv[])
     }
 
     if ((sg_fd = sg_cmds_open_device(op->device_name, true /* ro */,
-                                     op->do_verbose)) < 0) {
-        pr2serr("sg_turs: error opening file: %s: %s\n", op->device_name,
-                safe_strerror(-sg_fd));
-        return sg_convert_errno(-sg_fd);
+                                     op->verbose)) < 0) {
+        pr2serr("%s: error opening file: %s: %s\n", __func__,
+                op->device_name, safe_strerror(-sg_fd));
+        ret = sg_convert_errno(-sg_fd);
+        goto fini;
+    }
+    ptvp = construct_scsi_pt_obj_with_fd(sg_fd, op->verbose);
+    if ((NULL == ptvp) || ((err = get_scsi_pt_os_err(ptvp)))) {
+        pr2serr("%s: unable to construct pt object\n", __func__);
+        ret = sg_convert_errno(err ? err : ENOMEM);
+        goto fini;
     }
     if (op->do_progress) {
         for (k = 0; k < op->do_number; ++k) {
-            if (k > 0)
-                sleep_for(30);
+            if (op->delay > 0) {
+                if (op->delay_given)
+                    wait_millisecs(op->delay);
+                else if (k > 0)
+                    wait_millisecs(op->delay);
+            }
             progress = -1;
-            res = sg_ll_test_unit_ready_progress(sg_fd, k, &progress,
-                     (1 == op->do_number), op->do_verbose);
+            res = sg_ll_test_unit_ready_progress_pt(ptvp, k, &progress,
+                             (1 == op->do_number), op->verbose);
             if (progress < 0) {
                 ret = res;
                 break;
@@ -483,12 +576,11 @@ main(int argc, char * argv[])
         start_tm_valid = false;
 #endif
 
-        num_done = loop_turs(sg_fd, resp, op);
+        num_done = loop_turs(ptvp, resp, op);
 
         if (op->do_time && start_tm_valid) {
 #if defined(HAVE_CLOCK_GETTIME) && defined(CLOCK_MONOTONIC)
             if (start_tm.tv_sec || start_tm.tv_nsec) {
-                int err;
 
                 res = clock_gettime(CLOCK_MONOTONIC, &end_tm);
                 if (res < 0) {
@@ -526,8 +618,11 @@ main(int argc, char * argv[])
                    op->do_number, resp->num_errs);
         if (1 == op->do_number)
             ret = resp->ret;
-
     }
-    sg_cmds_close_device(sg_fd);
+fini:
+    if (ptvp)
+        destruct_scsi_pt_obj(ptvp);
+    if (sg_fd >= 0)
+        sg_cmds_close_device(sg_fd);
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
