@@ -1,8 +1,10 @@
 /*
- * Copyright (c) 2004-2018 Douglas Gilbert.
+ * Copyright (c) 2004-2021 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <unistd.h>
@@ -30,7 +32,7 @@
  * mode page on the given device.
  */
 
-static const char * version_str = "1.22 20180219";
+static const char * version_str = "1.27 20210610";
 
 #define ME "sg_wr_mode: "
 
@@ -48,19 +50,25 @@ static struct option long_options[] = {
         {"len", required_argument, 0, 'l'},
         {"mask", required_argument, 0, 'm'},
         {"page", required_argument, 0, 'p'},
+        {"rtd", no_argument, 0, 'R'},
         {"save", no_argument, 0, 's'},
+        {"six", no_argument, 0, '6'},
         {"verbose", no_argument, 0, 'v'},
         {"version", no_argument, 0, 'V'},
         {0, 0, 0, 0},
 };
 
-static void usage()
+
+static void
+usage()
 {
     pr2serr("Usage: sg_wr_mode [--contents=H,H...] [--dbd] [--force] "
             "[--help]\n"
             "                  [--len=10|6] [--mask=M,M...] "
             "[--page=PG_H[,SPG_H]]\n"
-            "                  [--save] [--verbose] [--version] DEVICE\n"
+            "                  [--rtd] [--save] [--six] [--verbose] "
+            "[--version]\n"
+            "                  DEVICE\n"
             "  where:\n"
             "    --contents=H,H... | -c H,H...    comma separated string "
             "of hex numbers\n"
@@ -83,9 +91,13 @@ static void usage()
             "    --page=PG_H,SPG_H | -p PG_H,SPG_H    page and subpage code "
             "to be\n"
             "                                         written (in hex)\n"
+            "    --rtd | -R            set RTD bit (revert to defaults) in "
+            "cdb\n"
             "    --save | -s           set 'save page' (SP) bit; default "
             "don't so\n"
             "                          only 'current' values changed\n"
+            "    --six | -6            do SCSI MODE SENSE/SELECT(6) "
+            "commands\n"
             "    --verbose | -v        increase verbosity\n"
             "    --version | -V        print version string and exit\n\n"
             "writes given mode page with SCSI MODE SELECT (10 or 6) "
@@ -97,9 +109,10 @@ static void usage()
  * either be comma or space separated list. Space separated list need to be
  * quoted. For stdin (indicated by *inp=='-') there should be either
  * one entry per line, a comma separated list or space separated list.
- * Returns 0 if ok, or 1 if error. */
-static int build_mode_page(const char * inp, uint8_t * mp_arr,
-                           int * mp_arr_len, int max_arr_len)
+ * Returns 0 if ok, or sg3_utils error code if error. */
+static int
+build_mode_page(const char * inp, uint8_t * mp_arr, int * mp_arr_len,
+                int max_arr_len)
 {
     int in_len, k, j, m;
     unsigned int h;
@@ -109,7 +122,7 @@ static int build_mode_page(const char * inp, uint8_t * mp_arr,
 
     if ((NULL == inp) || (NULL == mp_arr) ||
         (NULL == mp_arr_len))
-        return 1;
+        return SG_LIB_LOGIC_ERROR;
     lcp = inp;
     in_len = strlen(inp);
     if (0 == in_len)
@@ -138,15 +151,15 @@ static int build_mode_page(const char * inp, uint8_t * mp_arr,
                 continue;
             }
             if (carry_over[0]) {
-                if (isxdigit(line[0])) {
+                if (isxdigit((uint8_t)line[0])) {
                     carry_over[1] = line[0];
                     carry_over[2] = '\0';
                     if (1 == sscanf(carry_over, "%x", &h))
                         mp_arr[off - 1] = h;       /* back up and overwrite */
                     else {
-                        pr2serr("build_mode_page: carry_over error ['%s'] "
-                                "around line %d\n", carry_over, j + 1);
-                        return 1;
+                        pr2serr("%s: carry_over error ['%s'] around line "
+                                "%d\n", __func__, carry_over, j + 1);
+                        return SG_LIB_SYNTAX_ERROR;
                     }
                     lcp = line + 1;
                     --in_len;
@@ -164,9 +177,9 @@ static int build_mode_page(const char * inp, uint8_t * mp_arr,
                 continue;
             k = strspn(lcp, "0123456789aAbBcCdDeEfF ,\t");
             if ((k < in_len) && ('#' != lcp[k])) {
-                pr2serr("build_mode_page: syntax error at "
-                        "line %d, pos %d\n", j + 1, m + k + 1);
-                return 1;
+                pr2serr("%s: syntax error at line %d, pos %d\n", __func__,
+                        j + 1, m + k + 1);
+                return SG_LIB_SYNTAX_ERROR;
             }
             for (k = 0; k < 1024; ++k) {
                 if (1 == sscanf(lcp, "%x", &h)) {
@@ -174,7 +187,7 @@ static int build_mode_page(const char * inp, uint8_t * mp_arr,
                         pr2serr("%s: hex number larger than 0xff in line %d, "
                                 "pos %d\n", __func__, j + 1,
                                 (int)(lcp - line + 1));
-                        return 1;
+                        return SG_LIB_SYNTAX_ERROR;
                     }
                     if (split_line && (1 == strlen(lcp))) {
                         /* single trailing hex digit might be a split pair */
@@ -182,7 +195,7 @@ static int build_mode_page(const char * inp, uint8_t * mp_arr,
                     }
                     if ((off + k) >= max_arr_len) {
                         pr2serr("%s: array length exceeded\n", __func__);
-                        return 1;
+                        return SG_LIB_SYNTAX_ERROR;
                     }
                     mp_arr[off + k] = h;
                     lcp = strpbrk(lcp, " ,\t");
@@ -198,7 +211,7 @@ static int build_mode_page(const char * inp, uint8_t * mp_arr,
                     }
                     pr2serr("%s: error in line %d, at pos %d\n", __func__,
                             j + 1, (int)(lcp - line + 1));
-                    return 1;
+                    return SG_LIB_SYNTAX_ERROR;
                 }
             }
             off += (k + 1);
@@ -208,14 +221,14 @@ static int build_mode_page(const char * inp, uint8_t * mp_arr,
         k = strspn(inp, "0123456789aAbBcCdDeEfF, ");
         if (in_len != k) {
             pr2serr("%s: error at pos %d\n", __func__, k + 1);
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
         for (k = 0; k < max_arr_len; ++k) {
             if (1 == sscanf(lcp, "%x", &h)) {
                 if (h > 0xff) {
                     pr2serr("%s: hex number larger than 0xff at pos %d\n",
                             __func__, (int)(lcp - inp + 1));
-                    return 1;
+                    return SG_LIB_SYNTAX_ERROR;
                 }
                 mp_arr[k] = h;
                 cp = (char *)strchr(lcp, ',');
@@ -230,13 +243,13 @@ static int build_mode_page(const char * inp, uint8_t * mp_arr,
             } else {
                 pr2serr("%s: error at pos %d\n", __func__,
                         (int)(lcp - inp + 1));
-                return 1;
+                return SG_LIB_SYNTAX_ERROR;
             }
         }
         *mp_arr_len = k + 1;
         if (k == max_arr_len) {
             pr2serr("%s: array length exceeded\n", __func__);
-            return 1;
+            return SG_LIB_SYNTAX_ERROR;
         }
     }
     return 0;
@@ -245,8 +258,9 @@ static int build_mode_page(const char * inp, uint8_t * mp_arr,
 /* Read hex numbers from command line (comma separated list).
  * Can also be (single) space separated list but needs to be quoted on the
  * command line. Returns 0 if ok, or 1 if error. */
-static int build_mask(const char * inp, uint8_t * mask_arr,
-                      int * mask_arr_len, int max_arr_len)
+static int
+build_mask(const char * inp, uint8_t * mask_arr, int * mask_arr_len,
+           int max_arr_len)
 {
     int in_len, k;
     unsigned int h;
@@ -303,20 +317,26 @@ static int build_mask(const char * inp, uint8_t * mask_arr,
 }
 
 
-int main(int argc, char * argv[])
+int
+main(int argc, char * argv[])
 {
     bool dbd = false;
     bool force = false;
     bool got_contents = false;
     bool got_mask = false;
-    bool mode_6 = false;
+    bool mode_6 = false;        /* so default is mode_10 */
+    bool rtd = false;   /* added in spc5r11 */
     bool save = false;
-    int sg_fd, res, c, num, alloc_len, off, pdt;
-    int k, md_len, hdr_len, bd_len, mask_in_len;
+    bool verbose_given = false;
+    bool version_given = false;
+    int res, c, num, alloc_len, off, pdt, k, md_len, hdr_len, bd_len;
+    int mask_in_len;
+    int sg_fd = -1;
     int pg_code = -1;
     int sub_pg_code = 0;
     int verbose = 0;
     int read_in_len = 0;
+    int ret = 0;
     unsigned u, uu;
     const char * device_name = NULL;
     uint8_t read_in[MX_ALLOC_LEN];
@@ -326,23 +346,25 @@ int main(int argc, char * argv[])
     char errStr[128];
     char b[80];
     struct sg_simple_inquiry_resp inq_data;
-    int ret = 0;
 
     while (1) {
         int option_index = 0;
 
-        c = getopt_long(argc, argv, "c:dfhl:m:p:svV", long_options,
+        c = getopt_long(argc, argv, "6c:dfhl:m:p:RsvV", long_options,
                         &option_index);
         if (c == -1)
             break;
 
         switch (c) {
+        case '6':
+            mode_6 = true;
+            break;
         case 'c':
             memset(read_in, 0, sizeof(read_in));
-            if (0 != build_mode_page(optarg, read_in, &read_in_len,
-                                     sizeof(read_in))) {
-                pr2serr("bad argument to '--contents'\n");
-                return SG_LIB_SYNTAX_ERROR;
+            if ((ret = build_mode_page(optarg, read_in, &read_in_len,
+                                       sizeof(read_in)))) {
+                pr2serr("bad argument to '--contents='\n");
+                return ret;
             }
             got_contents = true;
             break;
@@ -397,15 +419,19 @@ int main(int argc, char * argv[])
                 return SG_LIB_SYNTAX_ERROR;
             }
             break;
+        case 'R':
+            rtd = true;
+            break;
         case 's':
             save = true;
             break;
         case 'v':
+            verbose_given = true;
             ++verbose;
             break;
         case 'V':
-            pr2serr(ME "version: %s\n", version_str);
-            return 0;
+            version_given = true;
+            break;
         default:
             pr2serr("unrecognised option code 0x%x ??\n", c);
             usage();
@@ -424,28 +450,56 @@ int main(int argc, char * argv[])
             return SG_LIB_SYNTAX_ERROR;
         }
     }
+
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr(ME "version: %s\n", version_str);
+        return 0;
+    }
+
     if (NULL == device_name) {
-        pr2serr("missing device name!\n");
+        pr2serr("missing device name!\n\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
-    if (pg_code < 0) {
-        pr2serr("need page code (see '--page=')\n");
+    if ((pg_code < 0) && (! rtd)) {
+        pr2serr("need page code (see '--page=')\n\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
     if (got_mask && force) {
-        pr2serr("cannot use both '--force' and '--mask'\n");
+        pr2serr("cannot use both '--force' and '--mask'\n\n");
         usage();
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
 
     sg_fd = sg_cmds_open_device(device_name, false /* rw */, verbose);
     if (sg_fd < 0) {
-        pr2serr(ME "open error: %s: %s\n", device_name, safe_strerror(-sg_fd));
-        return SG_LIB_FILE_ERROR;
+        if (verbose)
+            pr2serr(ME "open error: %s: %s\n", device_name,
+                    safe_strerror(-sg_fd));
+        ret = sg_convert_errno(-sg_fd);
+        goto fini;
     }
-    if (0 == sg_simple_inquiry(sg_fd, &inq_data, 0, verbose))
+    if (rtd)
+        goto revert_to_defaults;
+
+    if (0 == sg_simple_inquiry(sg_fd, &inq_data, false, verbose))
         pdt = inq_data.peripheral_type;
     else
         pdt = 0x1f;
@@ -471,23 +525,23 @@ int main(int argc, char * argv[])
             sg_get_category_sense_str(res, sizeof(b), b, verbose);
             pr2serr("%s%s\n", errStr, b);
         }
-        goto err_out;
+        goto fini;
     }
     off = sg_mode_page_offset(ref_md, alloc_len, mode_6, ebuff, EBUFF_SZ);
     if (off < 0) {
         pr2serr("%s%s\n", errStr, ebuff);
-        goto err_out;
+        goto fini;
     }
     md_len = sg_msense_calc_length(ref_md, alloc_len, mode_6, &bd_len);
     if (md_len < 0) {
         pr2serr("%ssg_msense_calc_length() failed\n", errStr);
-        goto err_out;
+        goto fini;
     }
     hdr_len = mode_6 ? 4 : 8;
     if (got_contents) {
         if (read_in_len < 2) {
             pr2serr("contents length=%d too short\n", read_in_len);
-            goto err_out;
+            goto fini;
         }
         ref_md[0] = 0;  /* mode data length reserved for mode select */
         if (! mode_6)
@@ -497,7 +551,7 @@ int main(int argc, char * argv[])
         if (md_len > alloc_len) {
             pr2serr("mode data length=%d exceeds allocation length=%d\n",
                     md_len, alloc_len);
-            goto err_out;
+            goto fini;
         }
         if (got_mask) {
             for (k = 0; k < (md_len - off); ++k) {
@@ -514,45 +568,42 @@ int main(int argc, char * argv[])
             if ((! (ref_md[off] & 0x80)) && save) {
                 pr2serr("PS bit in existing mode page indicates that it is "
                         "not saveable\n    but '--save' option given\n");
-                goto err_out;
+                goto fini;
             }
             read_in[0] &= 0x7f; /* mask out PS bit, reserved in mode select */
             if ((md_len - off) != read_in_len) {
                 pr2serr("contents length=%d but reference mode page "
                         "length=%d\n", read_in_len, md_len - off);
-                goto err_out;
+                goto fini;
             }
             if (pg_code != (read_in[0] & 0x3f)) {
                 pr2serr("contents page_code=0x%x but reference "
                         "page_code=0x%x\n", (read_in[0] & 0x3f), pg_code);
-                goto err_out;
+                goto fini;
             }
             if ((read_in[0] & 0x40) != (ref_md[off] & 0x40)) {
                 pr2serr("contents flags subpage but reference page does not "
                         "(or vice versa)\n");
-                goto err_out;
+                goto fini;
             }
             if ((read_in[0] & 0x40) && (read_in[1] != sub_pg_code)) {
                 pr2serr("contents subpage_code=0x%x but reference "
                         "sub_page_code=0x%x\n", read_in[1], sub_pg_code);
-                goto err_out;
+                goto fini;
             }
         } else
             md_len = off + read_in_len; /* force length */
 
         memcpy(ref_md + off, read_in, read_in_len);
         if (mode_6)
-            res = sg_ll_mode_select6(sg_fd, true /* PF */, save, ref_md,
-                                     md_len, true, verbose);
+            res = sg_ll_mode_select6_v2(sg_fd, true /* PF */, rtd, save,
+                                        ref_md, md_len, true, verbose);
         else
-            res = sg_ll_mode_select10(sg_fd, true /* PF */, save, ref_md,
-                                      md_len, true, verbose);
+            res = sg_ll_mode_select10_v2(sg_fd, true /* PF */, rtd, save,
+                                         ref_md, md_len, true, verbose);
         ret = res;
-        if (res) {
-            sg_get_category_sense_str(res, sizeof(b), b, verbose);
-            pr2serr("MODE SELECT (%d): %s\n", (mode_6 ? 6 : 10), b);
-            goto err_out;
-        }
+        if (res)
+            goto fini;
     } else {
         printf(">>> No contents given, so show current mode page data:\n");
         printf("  header:\n");
@@ -565,12 +616,33 @@ int main(int argc, char * argv[])
         printf("  mode page:\n");
         hex2stdout(ref_md + off, md_len - off, -1);
     }
-err_out:
-    res = sg_cmds_close_device(sg_fd);
-    if (res < 0) {
-        pr2serr("close error: %s\n", safe_strerror(-res));
-        if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+    ret = 0;
+    goto fini;
+
+revert_to_defaults:
+    if (verbose)
+        pr2serr("Doing MODE SELECT(%d) with revert to defaults (RTD) set "
+                "and SP=%d\n", mode_6 ? 6 : 10, !! save);
+    if (mode_6)
+        res = sg_ll_mode_select6_v2(sg_fd, false /* PF */, true /* rtd */,
+                                    save, NULL, 0, true, verbose);
+    else
+        res = sg_ll_mode_select10_v2(sg_fd, false /* PF */, true /* rtd */,
+                                     save, NULL, 0, true, verbose);
+    ret = res;
+fini:
+    if (sg_fd >= 0) {
+        res = sg_cmds_close_device(sg_fd);
+        if (res < 0) {
+            pr2serr("close error: %s\n", safe_strerror(-res));
+            if (0 == ret)
+                ret = sg_convert_errno(-res);
+        }
+    }
+    if (0 == verbose) {
+        if (! sg_if_can2stderr("sg_wr_mode failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' or '-vv' for "
+                    "more information\n");
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
