@@ -1,8 +1,10 @@
 /*
- * Copyright (c) 2006-2018 Luben Tuikov and Douglas Gilbert.
+ * Copyright (c) 2006-2021 Luben Tuikov and Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <unistd.h>
@@ -13,6 +15,7 @@
 #include <stdint.h>
 #include <stdbool.h>
 #include <ctype.h>
+#include <errno.h>
 #include <string.h>
 #include <getopt.h>
 #define __STDC_FORMAT_MACROS 1
@@ -37,7 +40,7 @@
  * This utility issues the SCSI WRITE BUFFER command to the given device.
  */
 
-static const char * version_str = "1.25 20180217";    /* spc5r19 */
+static const char * version_str = "1.30 20210610";    /* spc6r05 */
 
 #define ME "sg_write_buffer: "
 #define DEF_XFER_LEN (8 * 1024 * 1024)
@@ -192,8 +195,11 @@ main(int argc, char * argv[])
     bool bpw_then_activate = false;
     bool dry_run = false;
     bool got_stdin = false;
+    bool verbose_given = false;
+    bool version_given = false;
     bool wb_len_given = false;
-    int sg_fd, infd, res, c, len, k, n;
+    int infd, res, c, len, k, n;
+    int sg_fd = -1;
     int bpw = 0;
     int do_help = 0;
     int ret = 0;
@@ -208,6 +214,7 @@ main(int argc, char * argv[])
     const char * device_name = NULL;
     const char * file_name = NULL;
     uint8_t * dop = NULL;
+    uint8_t * read_buf = NULL;
     uint8_t * free_dop = NULL;
     char * cp;
     const struct mode_s * mp;
@@ -261,7 +268,7 @@ main(int argc, char * argv[])
              wb_len_given = true;
              break;
         case 'm':
-            if (isdigit(*optarg)) {
+            if (isdigit((uint8_t)*optarg)) {
                 wb_mode = sg_get_num(optarg);
                 if ((wb_mode < 0) || (wb_mode > 31)) {
                     pr2serr("argument to '--mode' should be in the range 0 "
@@ -314,11 +321,12 @@ main(int argc, char * argv[])
             }
             break;
         case 'v':
+            verbose_given = true;
             ++verbose;
             break;
         case 'V':
-            pr2serr(ME "version: %s\n", version_str);
-            return 0;
+            version_given = true;
+            break;
         default:
             pr2serr("unrecognised option code 0x%x ??\n", c);
             usage();
@@ -347,8 +355,29 @@ main(int argc, char * argv[])
         }
     }
 
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr("version: %s\n", version_str);
+        return 0;
+    }
+
     if (NULL == device_name) {
-        pr2serr("missing device name!\n");
+        pr2serr("Missing device name!\n\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
@@ -369,9 +398,11 @@ main(int argc, char * argv[])
 
     sg_fd = sg_cmds_open_device(device_name, false /* rw */, verbose);
     if (sg_fd < 0) {
-        pr2serr(ME "open error: %s: %s\n", device_name,
-                safe_strerror(-sg_fd));
-        return SG_LIB_FILE_ERROR;
+        if (verbose)
+            pr2serr(ME "open error: %s: %s\n", device_name,
+                    safe_strerror(-sg_fd));
+        ret = sg_convert_errno(-sg_fd);
+        goto err_out;
     }
     if (file_name || (wb_len > 0)) {
         if (0 == wb_len)
@@ -379,7 +410,7 @@ main(int argc, char * argv[])
         dop = sg_memalign(wb_len, 0, &free_dop, false);
         if (NULL == dop) {
             pr2serr(ME "out of memory\n");
-            ret = SG_LIB_SYNTAX_ERROR;
+            ret = sg_convert_errno(ENOMEM);
             goto err_out;
         }
         memset(dop, 0xff, wb_len);
@@ -394,33 +425,71 @@ main(int argc, char * argv[])
                 infd = STDIN_FILENO;
             } else {
                 if ((infd = open(file_name, O_RDONLY)) < 0) {
+                    ret = sg_convert_errno(errno);
                     snprintf(ebuff, EBUFF_SZ,
                              ME "could not open %s for reading", file_name);
                     perror(ebuff);
-                    ret = SG_LIB_FILE_ERROR;
                     goto err_out;
                 } else if (sg_set_binary_mode(infd) < 0)
                     perror("sg_set_binary_mode");
                 if (wb_skip > 0) {
                     if (lseek(infd, wb_skip, SEEK_SET) < 0) {
+                        ret = sg_convert_errno(errno);
                         snprintf(ebuff,  EBUFF_SZ, ME "couldn't skip to "
                                  "required position on %s", file_name);
                         perror(ebuff);
                         close(infd);
-                        ret = SG_LIB_FILE_ERROR;
                         goto err_out;
                     }
                 }
             }
-            res = read(infd, dop, wb_len);
-            if (res < 0) {
-                snprintf(ebuff, EBUFF_SZ, ME "couldn't read from %s",
-                         file_name);
-                perror(ebuff);
-                if (! got_stdin)
-                    close(infd);
-                ret = SG_LIB_FILE_ERROR;
-                goto err_out;
+            if (infd == STDIN_FILENO) {
+                if (NULL == (read_buf = (uint8_t *)malloc(DEF_XFER_LEN))) {
+                    pr2serr(ME "out of memory\n");
+                    ret = SG_LIB_SYNTAX_ERROR;
+                    goto err_out;
+                }
+                res = read(infd, read_buf, DEF_XFER_LEN);
+                if (res < 0) {
+                    snprintf(ebuff, EBUFF_SZ, ME "couldn't read from STDIN");
+                    perror(ebuff);
+                    ret = SG_LIB_FILE_ERROR;
+                    goto err_out;
+                }
+                char * pch;
+                int val = 0;
+                res = 0;
+                pch = strtok((char*)read_buf, ",. \n\t");
+                while (pch != NULL) {
+                    val = sg_get_num_nomult(pch);
+                    if (val >= 0 && val < 255) {
+                        dop[res] = val;
+                        res++;
+                    } else {
+                        pr2serr("Data read from STDIO is wrong.\nPlease "
+                                "input the data a byte at a time, the bytes "
+                                "should be separated\nby either space, or "
+                                "',' ( or by '.'), and the value per byte "
+                                "should\nbe between 0~255. Hexadecimal "
+                                "numbers should be preceded by either '0x' "
+                                "or\n'OX' (or have a trailing 'h' or "
+                                "'H').\n");
+                        ret = SG_LIB_SYNTAX_ERROR;
+                        goto err_out;
+                    }
+                    pch = strtok(NULL, ",. \n\t");
+                }
+            } else {
+                res = read(infd, dop, wb_len);
+                if (res < 0) {
+                    ret = sg_convert_errno(errno);
+                    snprintf(ebuff, EBUFF_SZ, ME "couldn't read from %s",
+                             file_name);
+                    perror(ebuff);
+                    if (! got_stdin)
+                        close(infd);
+                    goto err_out;
+                }
             }
             if (res < wb_len) {
                 if (wb_len_given) {
@@ -509,11 +578,20 @@ main(int argc, char * argv[])
 err_out:
     if (free_dop)
         free(free_dop);
-    res = sg_cmds_close_device(sg_fd);
-    if (res < 0) {
-        pr2serr("close error: %s\n", safe_strerror(-res));
-        if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+    if (read_buf)
+        free(read_buf);
+    if (sg_fd >= 0) {
+        res = sg_cmds_close_device(sg_fd);
+        if (res < 0) {
+            pr2serr("close error: %s\n", safe_strerror(-res));
+            if (0 == ret)
+                ret = sg_convert_errno(-res);
+        }
+    }
+    if (0 == verbose) {
+        if (! sg_if_can2stderr("sg_write_buffer failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' "
+                    "or '-vv' for more information\n");
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
