@@ -1,8 +1,10 @@
 /*
- * Copyright (c) 2009-2018 Douglas Gilbert.
+ * Copyright (c) 2009-2021 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <unistd.h>
@@ -31,7 +33,7 @@
 #include "sg_unaligned.h"
 #include "sg_pr2serr.h"
 
-static const char * version_str = "1.21 20180219";
+static const char * version_str = "1.33 20210830";
 
 
 #define ME "sg_write_same: "
@@ -62,6 +64,7 @@ static struct option long_options[] = {
     {"16", no_argument, 0, 'S'},
     {"32", no_argument, 0, 'T'},
     {"anchor", no_argument, 0, 'a'},
+    {"ff", no_argument, 0, 'f'},
     {"grpnum", required_argument, 0, 'g'},
     {"help", no_argument, 0, 'h'},
     {"in", required_argument, 0, 'i'},
@@ -81,10 +84,13 @@ static struct option long_options[] = {
 
 struct opts_t {
     bool anchor;
+    bool ff;
     bool ndob;
     bool lbdata;
     bool pbdata;
     bool unmap;
+    bool verbose_given;
+    bool version_given;
     bool want_ws10;
     int grpnum;
     int numblocks;
@@ -102,12 +108,12 @@ static void
 usage()
 {
     pr2serr("Usage: sg_write_same [--10] [--16] [--32] [--anchor] "
-            "[--grpnum=GN] [--help]\n"
-            "                     [--in=IF] [--lba=LBA] [--lbdata] "
-            "[--ndob] [--num=NUM]\n"
-            "                     [--pbdata] [--timeout=TO] [--unmap] "
-            "[--verbose]\n"
-            "                     [--version] [--wrprotect=WRP] "
+            "[-ff] [--grpnum=GN]\n"
+            "                     [--help] [--in=IF] [--lba=LBA] [--lbdata] "
+            "[--ndob]\n"
+            "                     [--num=NUM] [--pbdata] [--timeout=TO] "
+            "[--unmap]\n"
+            "                     [--verbose] [--version] [--wrprotect=WRP] "
             "[xferlen=LEN]\n"
             "                     DEVICE\n"
             "  where:\n"
@@ -119,6 +125,8 @@ usage()
             "then def 16)\n"
             "    --32|-T              send WRITE SAME(32) (def: 10 or 16)\n"
             "    --anchor|-a          set ANCHOR field in cdb\n"
+            "    --ff|-f              use buffer of 0xff bytes for fill "
+            "(def: 0x0 bytes)\n"
             "    --grpnum=GN|-g GN    GN is group number field (def: 0)\n"
             "    --help|-h            print out usage message\n"
             "    --in=IF|-i IF        IF is file to fetch one block of data "
@@ -150,7 +158,8 @@ usage()
             "specified blocks\nwill be filled with zeros or the "
             "'provisioning initialization pattern'\nas indicated by the "
             "LBPRZ field. As a precaution one of the '--in=',\n'--lba=' or "
-            "'--num=' options is required.\n"
+            "'--num=' options is required.\nAnother implementation of WRITE "
+            "SAME is found in the sg_write_x utility.\n"
             );
 }
 
@@ -158,7 +167,7 @@ static int
 do_write_same(int sg_fd, const struct opts_t * op, const void * dataoutp,
               int * act_cdb_lenp)
 {
-    int k, ret, res, sense_cat, cdb_len;
+    int ret, res, sense_cat, cdb_len;
     uint64_t llba;
     uint8_t ws_cdb[WRITE_SAME32_LEN];
     uint8_t sense_b[SENSE_BUFF_LEN];
@@ -246,11 +255,11 @@ do_write_same(int sg_fd, const struct opts_t * op, const void * dataoutp,
     }
 
     if (op->verbose > 1) {
-        pr2serr("    Write same(%d) cdb: ", cdb_len);
-        for (k = 0; k < cdb_len; ++k)
-            pr2serr("%02x ", ws_cdb[k]);
-        pr2serr("\n    Data-out buffer length=%d\n",
-                op->xfer_len);
+        char b[128];
+
+        pr2serr("    Write same(%d) cdb: %s\n", cdb_len,
+                sg_get_command_str(ws_cdb, cdb_len, false, sizeof(b), b));
+        pr2serr("    Data-out buffer length=%d\n", op->xfer_len);
     }
     if ((op->verbose > 3) && (op->xfer_len > 0)) {
         pr2serr("    Data-out buffer contents:\n");
@@ -265,12 +274,14 @@ do_write_same(int sg_fd, const struct opts_t * op, const void * dataoutp,
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     set_scsi_pt_data_out(ptvp, (uint8_t *)dataoutp, op->xfer_len);
     res = do_scsi_pt(ptvp, sg_fd, op->timeout, op->verbose);
-    ret = sg_cmds_process_resp(ptvp, "Write same", res, SG_NO_DATA_IN,
-                               sense_b, true /*noisy */, op->verbose,
-                               &sense_cat);
-    if (-1 == ret)
-        get_scsi_pt_os_err(ptvp);
-    else if (-2 == ret) {
+    ret = sg_cmds_process_resp(ptvp, "Write same", res, true /*noisy */,
+                               op->verbose, &sense_cat);
+    if (-1 == ret) {
+        if (get_scsi_pt_transport_err(ptvp))
+            ret = SG_LIB_TRANSPORT_ERROR;
+        else
+            ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
+    } else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
         case SG_LIB_CAT_NO_SENSE:
@@ -290,6 +301,10 @@ do_write_same(int sg_fd, const struct opts_t * op, const void * dataoutp,
             }
             ret = sense_cat;
             break;
+        case SG_LIB_CAT_ILLEGAL_REQ:
+            if (op->verbose)
+                sg_print_command_len(ws_cdb, cdb_len);
+            /* FALL THROUGH */
         default:
             ret = sense_cat;
             break;
@@ -310,13 +325,15 @@ main(int argc, char * argv[])
     bool lba_given = false;
     bool num_given = false;
     bool prot_en;
-    int sg_fd, res, c, infd, act_cdb_len, vb;
+    int res, c, infd, act_cdb_len, vb, err;
+    int sg_fd = -1;
     int ret = -1;
     uint32_t block_size;
     int64_t ll;
     const char * device_name = NULL;
     struct opts_t * op;
     uint8_t * wBuff = NULL;
+    uint8_t * free_wBuff = NULL;
     char ebuff[EBUFF_SZ];
     char b[80];
     uint8_t resp_buff[RCAP16_RESP_LEN];
@@ -331,7 +348,7 @@ main(int argc, char * argv[])
     while (1) {
         int option_index = 0;
 
-        c = getopt_long(argc, argv, "ag:hi:l:Ln:NPRSt:TUvVw:x:",
+        c = getopt_long(argc, argv, "afg:hi:l:Ln:NPRSt:TUvVw:x:",
                         long_options, &option_index);
         if (c == -1)
             break;
@@ -340,6 +357,9 @@ main(int argc, char * argv[])
         case 'a':
             op->anchor = true;
             break;
+        case 'f':
+            op->ff = true;
+            break;
         case 'g':
             op->grpnum = sg_get_num(optarg);
             if ((op->grpnum < 0) || (op->grpnum > 63))  {
@@ -352,7 +372,8 @@ main(int argc, char * argv[])
             usage();
             return 0;
         case 'i':
-            strncpy(op->ifilename, optarg, sizeof(op->ifilename));
+            strncpy(op->ifilename, optarg, sizeof(op->ifilename) - 1);
+            op->ifilename[sizeof(op->ifilename) - 1] = '\0';
             if_given = true;
             break;
         case 'l':
@@ -387,7 +408,7 @@ main(int argc, char * argv[])
         case 'S':
             if (DEF_WS_CDB_SIZE != op->pref_cdb_size) {
                 pr2serr("only one '--10', '--16' or '--32' please\n");
-                return SG_LIB_SYNTAX_ERROR;
+                return SG_LIB_CONTRADICT;
             }
             op->pref_cdb_size = 16;
             break;
@@ -401,7 +422,7 @@ main(int argc, char * argv[])
         case 'T':
             if (DEF_WS_CDB_SIZE != op->pref_cdb_size) {
                 pr2serr("only one '--10', '--16' or '--32' please\n");
-                return SG_LIB_SYNTAX_ERROR;
+                return SG_LIB_CONTRADICT;
             }
             op->pref_cdb_size = 32;
             break;
@@ -409,11 +430,12 @@ main(int argc, char * argv[])
             op->unmap = true;
             break;
         case 'v':
+            op->verbose_given = true;
             ++op->verbose;
             break;
         case 'V':
-            pr2serr(ME "version: %s\n", version_str);
-            return 0;
+            op->version_given = true;
+            break;
         case 'w':
             op->wrprotect = sg_get_num(optarg);
             if ((op->wrprotect < 0) || (op->wrprotect > 7))  {
@@ -448,10 +470,32 @@ main(int argc, char * argv[])
     }
     if (op->want_ws10 && (DEF_WS_CDB_SIZE != op->pref_cdb_size)) {
         pr2serr("only one '--10', '--16' or '--32' please\n");
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
+
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (op->verbose_given && op->version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        op->verbose_given = false;
+        op->version_given = false;
+        op->verbose = 0;
+    } else if (! op->verbose_given) {
+        pr2serr("set '-vv'\n");
+        op->verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", op->verbose);
+#else
+    if (op->verbose_given && op->version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (op->version_given) {
+        pr2serr(ME "version: %s\n", version_str);
+        return 0;
+    }
+
     if (NULL == device_name) {
-        pr2serr("missing device name!\n");
+        pr2serr("Missing device name!\n\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
@@ -460,28 +504,29 @@ main(int argc, char * argv[])
     if ((! if_given) && (! lba_given) && (! num_given)) {
         pr2serr("As a precaution, one of '--in=', '--lba=' or '--num=' is "
                 "required\n");
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
 
     if (op->ndob) {
         if (if_given) {
             pr2serr("Can't have both --ndob and '--in='\n");
-            return SG_LIB_SYNTAX_ERROR;
+            return SG_LIB_CONTRADICT;
         }
         if (0 != op->xfer_len) {
             pr2serr("With --ndob only '--xferlen=0' (or not given) is "
                     "acceptable\n");
-            return SG_LIB_SYNTAX_ERROR;
+            return SG_LIB_CONTRADICT;
         }
     } else if (op->ifilename[0]) {
         got_stdin = (0 == strcmp(op->ifilename, "-"));
         if (! got_stdin) {
             memset(&a_stat, 0, sizeof(a_stat));
             if (stat(op->ifilename, &a_stat) < 0) {
+                err = errno;
                 if (vb)
                     pr2serr("unable to stat(%s): %s\n", op->ifilename,
-                            safe_strerror(errno));
-                return SG_LIB_FILE_ERROR;
+                            safe_strerror(err));
+                return sg_convert_errno(err);
             }
             if (op->xfer_len <= 0)
                 op->xfer_len = (int)a_stat.st_size;
@@ -490,8 +535,11 @@ main(int argc, char * argv[])
 
     sg_fd = sg_cmds_open_device(device_name, false /* rw */, vb);
     if (sg_fd < 0) {
-        pr2serr(ME "open error: %s: %s\n", device_name, safe_strerror(-sg_fd));
-        return SG_LIB_FILE_ERROR;
+        if (op->verbose)
+            pr2serr(ME "open error: %s: %s\n", device_name,
+                    safe_strerror(-sg_fd));
+        ret = sg_convert_errno(-sg_fd);
+        goto err_out;
     }
 
     if (! op->ndob) {
@@ -550,13 +598,15 @@ main(int argc, char * argv[])
             ret = SG_LIB_SYNTAX_ERROR;
             goto err_out;
         }
-        wBuff = (uint8_t*)calloc(op->xfer_len, 1);
+        wBuff = (uint8_t *)sg_memalign(op->xfer_len, 0, &free_wBuff, false);
         if (NULL == wBuff) {
-            pr2serr("unable to allocate %d bytes of memory with calloc()\n",
-                    op->xfer_len);
-            ret = SG_LIB_SYNTAX_ERROR;
+            pr2serr("unable to allocate %d bytes of memory with "
+                    "sg_memalign()\n", op->xfer_len);
+            ret = sg_convert_errno(ENOMEM);
             goto err_out;
         }
+        if (op->ff)
+            memset(wBuff, 0xff, op->xfer_len);
         if (op->ifilename[0]) {
             if (got_stdin) {
                 infd = STDIN_FILENO;
@@ -564,22 +614,22 @@ main(int argc, char * argv[])
                     perror("sg_set_binary_mode");
             } else {
                 if ((infd = open(op->ifilename, O_RDONLY)) < 0) {
+                    ret = sg_convert_errno(errno);
                     snprintf(ebuff, EBUFF_SZ, ME "could not open %.400s for "
                              "reading", op->ifilename);
                     perror(ebuff);
-                    ret = SG_LIB_FILE_ERROR;
                     goto err_out;
                 } else if (sg_set_binary_mode(infd) < 0)
                     perror("sg_set_binary_mode");
             }
             res = read(infd, wBuff, op->xfer_len);
             if (res < 0) {
+                ret = sg_convert_errno(errno);
                 snprintf(ebuff, EBUFF_SZ, ME "couldn't read from %.400s",
                          op->ifilename);
                 perror(ebuff);
                 if (! got_stdin)
                     close(infd);
-                ret = SG_LIB_FILE_ERROR;
                 goto err_out;
             }
             if (res < op->xfer_len) {
@@ -610,13 +660,20 @@ main(int argc, char * argv[])
     }
 
 err_out:
-    if (wBuff)
-        free(wBuff);
-    res = sg_cmds_close_device(sg_fd);
-    if (res < 0) {
-        pr2serr("close error: %s\n", safe_strerror(-res));
-        if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+    if (free_wBuff)
+        free(free_wBuff);
+    if (sg_fd >= 0) {
+        res = sg_cmds_close_device(sg_fd);
+        if (res < 0) {
+            pr2serr("close error: %s\n", safe_strerror(-res));
+            if (0 == ret)
+                ret = sg_convert_errno(-res);
+        }
+    }
+    if (0 == op->verbose) {
+        if (! sg_if_can2stderr("sg_write_same failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' "
+                    "or '-vv' for more information\n");
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
