@@ -1,9 +1,11 @@
 /*
- * Copyright (c) 2014-2018 Douglas Gilbert
+ * Copyright (c) 2014-2021 Douglas Gilbert
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * This program issues the SCSI command WRITE AND VERIFY to a given SCSI
  * device. It sends the command with the logical block address passed as the
  * LBA argument, for the given number of blocks. The number of bytes sent is
@@ -38,7 +40,7 @@
 #include "sg_unaligned.h"
 #include "sg_pr2serr.h"
 
-static const char * version_str = "1.13 20180219";
+static const char * version_str = "1.19 20210830";
 
 
 #define ME "sg_write_verify: "
@@ -125,18 +127,18 @@ run_scsi_transaction(int sg_fd, const uint8_t *cdbp, int cdb_len,
                      uint8_t *dop, int do_len, int timeout,
                      bool noisy, int verbose)
 {
-    int res, k, sense_cat, ret;
+    int res, sense_cat, ret;
     struct sg_pt_base * ptvp;
     uint8_t sense_b[SENSE_BUFF_LEN];
     char b[32];
 
     snprintf(b, sizeof(b), "Write and verify(%d)", cdb_len);
     if (verbose) {
-       pr2serr("    %s cdb: ", b);
-       for (k = 0; k < cdb_len; ++k)
-           pr2serr("%02x ", cdbp[k]);
-       pr2serr("\n");
-       if ((verbose > 2) && dop && do_len) {
+        char d[128];
+
+        pr2serr("    %s cdb: %s\n", b,
+                sg_get_command_str(cdbp, cdb_len, false, sizeof(d), d));
+        if ((verbose > 2) && dop && do_len) {
             pr2serr("    Data out buffer [%d bytes]:\n", do_len);
             hex2stderr(dop, do_len, -1);
         }
@@ -150,11 +152,13 @@ run_scsi_transaction(int sg_fd, const uint8_t *cdbp, int cdb_len,
     set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
     set_scsi_pt_data_out(ptvp, dop, do_len);
     res = do_scsi_pt(ptvp, sg_fd, timeout, verbose);
-    ret = sg_cmds_process_resp(ptvp, b, res, SG_NO_DATA_IN, sense_b, noisy,
-                               verbose, &sense_cat);
-    if (-1 == ret)
-        ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
-    else if (-2 == ret) {
+    ret = sg_cmds_process_resp(ptvp, b, res, noisy, verbose, &sense_cat);
+    if (-1 == ret) {
+        if (get_scsi_pt_transport_err(ptvp))
+            ret = SG_LIB_TRANSPORT_ERROR;
+        else
+            ret = sg_convert_errno(get_scsi_pt_os_err(ptvp));
+    } else if (-2 == ret) {
         switch (sense_cat) {
         case SG_LIB_CAT_RECOVERED:
         case SG_LIB_CAT_NO_SENSE:
@@ -174,6 +178,10 @@ run_scsi_transaction(int sg_fd, const uint8_t *cdbp, int cdb_len,
             }
             ret = sense_cat;
             break;
+        case SG_LIB_CAT_ILLEGAL_REQ:
+            if (verbose)
+                sg_print_command_len(cdbp, cdb_len);
+            /* FALL THROUGH */
         case SG_LIB_CAT_PROTECTION:     /* PI failure */
         case SG_LIB_CAT_MISCOMPARE:     /* only in bytchk=1 case */
         default:
@@ -241,18 +249,21 @@ sg_ll_write_verify16(int sg_fd, int wrprotect, bool dpo, int bytchk,
     return ret;
 }
 
+/* Returns file descriptor ( >= 0) if successful. Else a negated sg3_utils
+ * error code is returned. */
 static int
 open_if(const char * fn, int got_stdin)
 {
-    int fd;
+    int fd, err;
 
     if (got_stdin)
         fd = STDIN_FILENO;
     else {
         fd = open(fn, O_RDONLY);
         if (fd < 0) {
-            pr2serr(ME "open error: %s: %s\n", fn, safe_strerror(errno));
-            return -SG_LIB_FILE_ERROR;
+            err = errno;
+            pr2serr(ME "open error: %s: %s\n", fn, safe_strerror(err));
+            return -sg_convert_errno(err);
         }
     }
     if (sg_set_binary_mode(fd) < 0) {
@@ -272,6 +283,8 @@ main(int argc, char * argv[])
     bool has_filename = false;
     bool lba_given = false;
     bool repeat = false;
+    bool verbose_given = false;
+    bool version_given = false;
     int sg_fd, res, c, n;
     int bytchk = 0;
     int group = 0;
@@ -288,7 +301,8 @@ main(int argc, char * argv[])
     uint64_t llba = 0;
     int64_t ll;
     uint8_t * wvb = NULL;
-    void * wrkBuff = NULL;
+    uint8_t * wrkBuff = NULL;
+    uint8_t * free_wrkBuff = NULL;
     const char * device_name = NULL;
     const char * ifnp;
     char cmd_name[32];
@@ -373,11 +387,12 @@ main(int argc, char * argv[])
             }
             break;
         case 'v':
+            verbose_given = true;
             ++verbose;
             break;
         case 'V':
-            pr2serr(ME "version: %s\n", version_str);
-            return 0;
+            version_given = true;
+            break;
         case 'w':
             wrprotect = sg_get_num(optarg);
             if ((wrprotect < 0) || (wrprotect > 7))  {
@@ -406,8 +421,29 @@ main(int argc, char * argv[])
        }
     }
 
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr(ME "version: %s\n", version_str);
+        return 0;
+    }
+
     if (NULL == device_name) {
-        pr2serr("missing device name!\n");
+        pr2serr("Missing device name!\n\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
@@ -420,12 +456,12 @@ main(int argc, char * argv[])
         if (! has_filename) {
             pr2serr("with '--repeat' need '--in=IF' option\n");
             usage();
-            return SG_LIB_SYNTAX_ERROR;
+            return SG_LIB_CONTRADICT;
         }
         if (ilen < 1) {
             pr2serr("with '--repeat' need '--ilen=ILEN' option\n");
             usage();
-            return SG_LIB_SYNTAX_ERROR;
+            return SG_LIB_CONTRADICT;
         } else {
             b_p_lb = ilen / num_lb;
             if (b_p_lb < 64) {
@@ -439,8 +475,9 @@ main(int argc, char * argv[])
 
     sg_fd = sg_cmds_open_device(device_name, false /* rw */, verbose);
     if (sg_fd < 0) {
+        ret = sg_convert_errno(-sg_fd);
         pr2serr(ME "open error: %s: %s\n", device_name, safe_strerror(-sg_fd));
-        return SG_LIB_FILE_ERROR;
+        goto err_out;
     }
 
     if ((! do_16) && (llba > UINT_MAX))
@@ -497,9 +534,10 @@ main(int argc, char * argv[])
                     } else if (verbose)
                         pr2serr("Using file size of %d bytes\n", ilen);
                 }
-                if (NULL == (wrkBuff = malloc(ilen))) {
+                if (NULL == (wrkBuff = (uint8_t *)sg_memalign(ilen, 0,
+                                           &free_wrkBuff, verbose > 3))) {
                     pr2serr(ME "out of memory\n");
-                    ret = SG_LIB_CAT_OTHER;
+                    ret = sg_convert_errno(ENOMEM);
                     goto err_out;
                 }
                 wvb = (uint8_t *)wrkBuff;
@@ -523,9 +561,10 @@ main(int argc, char * argv[])
                                 num_lb, 512, 512 * num_lb);
                     ilen = 512 * num_lb;
                 }
-                if (NULL == (wrkBuff = malloc(ilen))) {
+                if (NULL == (wrkBuff = (uint8_t *)sg_memalign(ilen, 0,
+                                           &free_wrkBuff, verbose > 3))) {
                     pr2serr(ME "out of memory\n");
-                    ret = SG_LIB_CAT_OTHER;
+                    ret = sg_convert_errno(ENOMEM);
                     goto err_out;
                 }
                 wvb = (uint8_t *)wrkBuff;
@@ -576,23 +615,20 @@ err_out:
     if (repeat)
         pr2serr("%d [0x%x] logical blocks written, in total\n", tnum_lb_wr,
                 tnum_lb_wr);
-    if (wrkBuff)
-        free(wrkBuff);
+    if (free_wrkBuff)
+        free(free_wrkBuff);
     if ((ifd >= 0) && (STDIN_FILENO != ifd))
         close(ifd);
     res = sg_cmds_close_device(sg_fd);
     if (res < 0) {
         pr2serr("close error: %s\n", safe_strerror(-res));
         if (0 == ret)
-            return SG_LIB_FILE_ERROR;
+            ret = sg_convert_errno(-res);
     }
     if (ret && (0 == verbose)) {
-        if (SG_LIB_CAT_INVALID_OP == ret)
-            pr2serr("%s command not supported\n", cmd_name);
-        else if (ret > 0)
-            pr2serr("%s, exit status %d\n", cmd_name, ret);
-        else if (ret < 0)
-            pr2serr("Some error occurred\n");
+        if (! sg_if_can2stderr("sg_write_verify failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' "
+                    "or '-vv' for more information\n");
     }
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
