@@ -1,7 +1,7 @@
 /* A utility program for copying files. Similar to 'dd' but using
  * the 'Extended Copy' command.
  *
- *  Copyright (c) 2011-2018 Hannes Reinecke, SUSE Labs
+ *  Copyright (c) 2011-2021 Hannes Reinecke, SUSE Labs
  *
  *  Largely taken from 'sg_dd', which has the
  *
@@ -11,6 +11,8 @@
  *  the Free Software Foundation; either version 2, or (at your option)
  *  any later version.
  *
+ * SPDX-License-Identifier: GPL-2.0-or-later
+ *
  * This program is a specialisation of the Unix "dd" command in which
  * either the input or the output file is a scsi generic device, raw
  * device, a block device or a normal file. The block size ('bs') is
@@ -25,7 +27,7 @@
  * in this case) is transferred to or from the sg device in a single SCSI
  * command.
  *
- * This version is designed for the linux kernel 2.4, 2.6, 3 and 4 series.
+ * This version is designed for the linux kernel 2.4, 2.6, 3, 4 and 5 series.
  */
 
 #define _XOPEN_SOURCE 600
@@ -55,7 +57,6 @@
 #include <sys/types.h>
 #endif
 #include <linux/major.h>
-#include <linux/fs.h>   /* <sys/mount.h> */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -68,7 +69,7 @@
 #include "sg_unaligned.h"
 #include "sg_pr2serr.h"
 
-static const char * version_str = "0.63 20180219";
+static const char * version_str = "0.72 20210902";
 
 #define ME "sg_xcopy: "
 
@@ -96,7 +97,7 @@ static const char * version_str = "0.63 20180219";
 #endif
 
 #ifndef RAW_MAJOR
-#define RAW_MAJOR 255   /*unlikey value */
+#define RAW_MAJOR 255   /*unlikely value */
 #endif
 
 #define SG_LIB_FLOCK_ERR 90
@@ -171,6 +172,7 @@ static bool do_time = false;
 static bool start_tm_valid = false;
 static bool xcopy_flag_cat = false;
 static bool xcopy_flag_dc = false;
+static bool xcopy_flag_fco = false;     /* fast copy only, spc5r20 */
 static int blk_sz = 0;
 static int list_id_usage = -1;
 static int priority = 1;
@@ -338,13 +340,13 @@ open_sg(struct xcopy_fp_t * fp, int vb)
                  fp->sg_type & FT_BLOCK ? "block" : "char",
                  devmajor, devminor);
         perror(ebuff);
-        return -1;
+        return -sg_convert_errno(-fp->sg_fd);
     }
-    if (sg_simple_inquiry(fp->sg_fd, &sir, 0, vb)) {
+    if (sg_simple_inquiry(fp->sg_fd, &sir, false, vb)) {
         pr2serr("INQUIRY failed on %s\n", ebuff);
         sg_cmds_close_device(fp->sg_fd);
         fp->sg_fd = -1;
-        return -2;
+        return -1;
     }
 
     fp->pdt = sir.peripheral_type;
@@ -402,21 +404,21 @@ dd_filetype_str(int ft, char * buff)
     int off = 0;
 
     if (FT_DEV_NULL & ft)
-        off += snprintf(buff + off, 32, "null device ");
+        off += sg_scnpr(buff + off, 32, "null device ");
     if (FT_SG & ft)
-        off += snprintf(buff + off, 32, "SCSI generic (sg) device ");
+        off += sg_scnpr(buff + off, 32, "SCSI generic (sg) device ");
     if (FT_BLOCK & ft)
-        off += snprintf(buff + off, 32, "block device ");
+        off += sg_scnpr(buff + off, 32, "block device ");
     if (FT_FIFO & ft)
-        off += snprintf(buff + off, 32, "fifo (named pipe) ");
+        off += sg_scnpr(buff + off, 32, "fifo (named pipe) ");
     if (FT_ST & ft)
-        off += snprintf(buff + off, 32, "SCSI tape device ");
+        off += sg_scnpr(buff + off, 32, "SCSI tape device ");
     if (FT_RAW & ft)
-        off += snprintf(buff + off, 32, "raw device ");
+        off += sg_scnpr(buff + off, 32, "raw device ");
     if (FT_OTHER & ft)
-        off += snprintf(buff + off, 32, "other (perhaps ordinary file) ");
+        off += sg_scnpr(buff + off, 32, "other (perhaps ordinary file) ");
     if (FT_ERROR & ft)
-        snprintf(buff + off, 32, "unable to 'stat' file ");
+        sg_scnpr(buff + off, 32, "unable to 'stat' file ");
     return buff;
 }
 
@@ -532,6 +534,7 @@ primary_help:
             "    conv        ignored\n"
             "    count       number of blocks to copy (def: device size)\n"
             "    dc          xcopy segment descriptor DC bit (default: 0)\n"
+            "    fco         xcopy segment descriptor FCO bit (default: 0)\n"
             "    ibs         input block size (if given must be same as "
             "'bs=')\n"
             "    id_usage    sets list_id_usage field to hold (0), "
@@ -595,6 +598,8 @@ scsi_encode_seg_desc(uint8_t *seg_desc, int seg_desc_type,
         seg_desc[1] |= 0x1;
     if (xcopy_flag_dc)
         seg_desc[1] |= 0x2;
+    if (xcopy_flag_fco)
+        seg_desc[1] |= 0x4;
     if (seg_desc_type == 0x02) {
         seg_desc_len = 0x18;
         seg_desc[4] = 0;
@@ -1070,27 +1075,27 @@ desc_from_vpd_id(int sg_fd, uint8_t *desc, int desc_len,
         if (verbose > 2)
             pr2serr("    Desc %d: assoc %u desig %u len %d\n", off, assoc,
                     desig, i_len);
-        /* Descriptor must be less than 16 bytes */
-        if (i_len > 16)
+        /* Identification descriptor's Designator length must be <= 20. */
+        if (i_len > 20)
             continue;
-        if (desig == 3) {
+        if (desig == /*NAA=*/3) {
             best = bp;
             best_len = i_len;
             break;
         }
-        if (desig == 2) {
+        if (desig == /*EUI64=*/2) {
             if (!best || f_desig < 2) {
                 best = bp;
                 best_len = i_len;
                 f_desig = 2;
             }
-        } else if (desig == 1) {
+        } else if (desig == /*T10*/1) {
             if (!best || f_desig == 0) {
                 best = bp;
                 best_len = i_len;
                 f_desig = desig;
             }
-        } else if (desig == 0) {
+        } else if (desig == /*vend.spec.=*/0) {
             if (!best) {
                 best = bp;
                 best_len = i_len;
@@ -1103,9 +1108,10 @@ desc_from_vpd_id(int sg_fd, uint8_t *desc, int desc_len,
             decode_designation_descriptor(best, best_len);
         if (best_len + 4 < desc_len) {
             memset(desc, 0, 32);
-            desc[0] = 0xe4;
+            desc[0] = 0xe4; /* Identification Descriptor */
             memcpy(desc + 4, best, best_len + 4);
-            desc[4] &= 0x1f;
+            desc[4] &= 0x0f; /* code set */
+            desc[5] &= 0x3f; /* association and designator type */
             if (pad)
                 desc[28] = 0x4;
             sg_put_unaligned_be24((uint32_t)block_size, desc + 29);
@@ -1196,7 +1202,7 @@ process_flags(const char * arg, struct xcopy_fp_t * fp)
 static int
 open_if(struct xcopy_fp_t * ifp, int vb)
 {
-    int infd = -1, flags, fl, res;
+    int infd = -1, flags, fl, res, err;
     char ebuff[EBUFF_SZ];
 
     ifp->sg_type = dd_filetype(ifp);
@@ -1216,10 +1222,11 @@ open_if(struct xcopy_fp_t * ifp, int vb)
     if ((infd = open(ifp->fname, fl | flags)) < 0) {
         fl = O_RDONLY;
         if ((infd = open(ifp->fname, fl | flags)) < 0) {
+            err = errno;
             snprintf(ebuff, EBUFF_SZ,
                      ME "could not open %.500s for sg reading", ifp->fname);
             perror(ebuff);
-            return -SG_LIB_FILE_ERROR;
+            return -sg_convert_errno(err);
         }
     }
     if (vb)
@@ -1245,7 +1252,7 @@ open_if(struct xcopy_fp_t * ifp, int vb)
 static int
 open_of(struct xcopy_fp_t * ofp, int vb)
 {
-    int outfd, flags, res;
+    int outfd, flags, res, err;
     char ebuff[EBUFF_SZ];
 
     ofp->sg_type = dd_filetype(ofp);
@@ -1258,11 +1265,14 @@ open_of(struct xcopy_fp_t * ofp, int vb)
         flags = O_RDWR | O_NONBLOCK;
         if (ofp->excl)
             flags |= O_EXCL;
+        if (ofp->append)
+            flags |= O_APPEND;
         if ((outfd = open(ofp->fname, flags)) < 0) {
+            err = errno;
             snprintf(ebuff, EBUFF_SZ,
                      ME "could not open %.500s for sg writing", ofp->fname);
             perror(ebuff);
-            return -SG_LIB_FILE_ERROR;
+            return -sg_convert_errno(err);
         }
         if (vb)
             pr2serr("        open output(sg_io), flags=0x%x\n", flags);
@@ -1301,6 +1311,8 @@ main(int argc, char * argv[])
     bool list_id_given = false;
     bool on_src = false;
     bool on_src_dst_given = false;
+    bool verbose_given = false;
+    bool version_given = false;
     int res, k, n, keylen, infd, outfd, xcopy_fd;
     int blocks = 0;
     int bpt = DEF_BLOCKS_PER_TRANSFER;
@@ -1329,7 +1341,7 @@ main(int argc, char * argv[])
     if (argc < 2) {
         pr2serr("Won't default both IFILE to stdin _and_ OFILE to stdout\n");
         pr2serr("For more information use '--help'\n");
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
 
     for (k = 1; k < argc; k++) {
@@ -1404,14 +1416,23 @@ main(int argc, char * argv[])
                 return SG_LIB_SYNTAX_ERROR;
             }
             xcopy_flag_dc = !! n;
+        } else if (0 == strcmp(key, "fco")) {
+            n = sg_get_num(buf);
+            if (n < 0 || n > 1) {
+                pr2serr(ME "bad argument to 'fco='\n");
+                return SG_LIB_SYNTAX_ERROR;
+            }
+            xcopy_flag_fco = !! n;
         } else if (0 == strcmp(key, "ibs")) {
             ibs = sg_get_num(buf);
         } else if (strcmp(key, "if") == 0) {
             if ('\0' != ixcf.fname[0]) {
                 pr2serr("Second IFILE argument??\n");
-                return SG_LIB_SYNTAX_ERROR;
-            } else
-                strncpy(ixcf.fname, buf, INOUTF_SZ - 1);
+                return SG_LIB_CONTRADICT;
+            } else {
+                memcpy(ixcf.fname, buf, INOUTF_SZ - 1);
+                ixcf.fname[INOUTF_SZ - 1] = '\0';
+            }
         } else if (0 == strcmp(key, "iflag")) {
             if (process_flags(buf, &ixcf)) {
                 pr2serr(ME "bad argument to 'iflag='\n");
@@ -1422,9 +1443,11 @@ main(int argc, char * argv[])
         } else if (strcmp(key, "of") == 0) {
             if ('\0' != oxcf.fname[0]) {
                 pr2serr("Second OFILE argument??\n");
-                return SG_LIB_SYNTAX_ERROR;
-            } else
-                strncpy(oxcf.fname, buf, INOUTF_SZ - 1);
+                return SG_LIB_CONTRADICT;
+            } else {
+                memcpy(oxcf.fname, buf, INOUTF_SZ - 1);
+                oxcf.fname[INOUTF_SZ - 1] = '\0';
+            }
         } else if (0 == strcmp(key, "oflag")) {
             if (process_flags(buf, &oxcf)) {
                 pr2serr(ME "bad argument to 'oflag='\n");
@@ -1455,7 +1478,7 @@ main(int argc, char * argv[])
                 pr2serr("Syntax error - either specify --on_src OR "
                         "--on_dst\n");
                 pr2serr("For more information use '--help'\n");
-                return SG_LIB_SYNTAX_ERROR;
+                return SG_LIB_CONTRADICT;
             }
             on_src_dst_given = true;
         } else if (0 == strncmp(key, "--on_src", 8)) {
@@ -1464,15 +1487,14 @@ main(int argc, char * argv[])
                 pr2serr("Syntax error - either specify --on_src OR "
                         "--on_dst\n");
                 pr2serr("For more information use '--help'\n");
-                return SG_LIB_SYNTAX_ERROR;
+                return SG_LIB_CONTRADICT;
             }
             on_src_dst_given = true;
-        } else if (0 == strncmp(key, "--verb", 6))
+        } else if (0 == strncmp(key, "--verb", 6)) {
+            verbose_given = true;
             verbose += 1;
-        else if (0 == strncmp(key, "--vers", 6)) {
-            pr2serr(ME "%s\n", version_str);
-            return 0;
-        }
+        } else if (0 == strncmp(key, "--vers", 6))
+            version_given = true;
         else if (0 == strncmp(key, "--xcopy", 7))
             ;   /* ignore; for compatibility with ddpt */
         /* look for short options that start with a single '-', they can be
@@ -1484,11 +1506,13 @@ main(int argc, char * argv[])
             res += n;
             n = num_chs_in_str(key + 1, keylen - 1, 'v');
             verbose += n;
+            if (n > 0)
+                verbose_given = true;
+            res += n;
+            n = num_chs_in_str(key + 1, keylen - 1, 'V');
+            if (n > 0)
+                version_given = true;
             res += n;
-            if (num_chs_in_str(key + 1, keylen - 1, 'V')) {
-                pr2serr("%s\n", version_str);
-                return -1;
-            }
             n = num_chs_in_str(key + 1, keylen - 1, 'x');
             /* accept and ignore; for compatibility with ddpt */
             res += n;
@@ -1511,6 +1535,27 @@ main(int argc, char * argv[])
         usage(num_help);
         return 0;
     }
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr(ME "%s\n", version_str);
+        return 0;
+    }
+
     if (! on_src_dst_given) {
         if (ixcf.xcopy_given == oxcf.xcopy_given) {
             char * csp;
@@ -1542,7 +1587,7 @@ main(int argc, char * argv[])
         (obs && blk_sz && (obs != blk_sz))) {
         pr2serr("If 'ibs' or 'obs' given must be same as 'bs'\n");
         pr2serr("For more information use '--help'\n");
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
     if (blk_sz && !ibs)
         ibs = blk_sz;
@@ -1551,11 +1596,11 @@ main(int argc, char * argv[])
 
     if ((skip < 0) || (seek < 0)) {
         pr2serr("skip and seek cannot be negative\n");
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
     if (oxcf.append && (seek > 0)) {
         pr2serr("Can't use both append and seek switches\n");
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
     if (bpt < 1) {
         pr2serr("bpt must be greater than 0\n");
@@ -1621,7 +1666,7 @@ main(int argc, char * argv[])
     if ((STDIN_FILENO == infd) && (STDOUT_FILENO == outfd)) {
         pr2serr("Can't have both 'if' as stdin _and_ 'of' as stdout\n");
         pr2serr("For more information use '--help'\n");
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
 
     res = scsi_read_capacity(&ixcf);
@@ -1716,22 +1761,25 @@ main(int argc, char * argv[])
             pr2serr("Unit attention (%s), continuing\n",
                     rec_copy_op_params_str);
             res = scsi_operating_parameter(&ixcf, 0);
-        } else {
-            if (-res == SG_LIB_CAT_INVALID_OP) {
-                pr2serr("%s command not supported on %s\n",
-                        rec_copy_op_params_str, ixcf.fname);
-                return EINVAL;
-            } else if (-res == SG_LIB_CAT_NOT_READY)
-                pr2serr("%s failed on %s - not ready\n",
-                        rec_copy_op_params_str, ixcf.fname);
-            else {
-                pr2serr("Unable to %s on %s\n", rec_copy_op_params_str,
-                        ixcf.fname);
-                return -res;
-            }
         }
-    } else if (res == 0)
-        return SG_LIB_CAT_INVALID_OP;
+        if (-res == SG_LIB_CAT_INVALID_OP) {
+            pr2serr("%s command not supported on %s\n",
+                    rec_copy_op_params_str, ixcf.fname);
+            ret = sg_convert_errno(EINVAL);
+            goto fini;
+        } else if (-res == SG_LIB_CAT_NOT_READY)
+            pr2serr("%s failed on %s - not ready\n",
+                    rec_copy_op_params_str, ixcf.fname);
+        else {
+            pr2serr("Unable to %s on %s\n", rec_copy_op_params_str,
+                    ixcf.fname);
+            ret = -res;
+            goto fini;
+        }
+    } else if (res == 0) {
+        ret = SG_LIB_CAT_INVALID_OP;
+        goto fini;
+    }
 
     if (res & TD_VPD) {
         if (verbose)
@@ -1741,10 +1789,12 @@ main(int argc, char * argv[])
                                  sizeof(src_desc), ixcf.sect_sz, ixcf.pad);
         if (src_desc_len > (int)sizeof(src_desc)) {
             pr2serr("source descriptor too large (%d bytes)\n", res);
-            return SG_LIB_CAT_MALFORMED;
+            ret = SG_LIB_CAT_MALFORMED;
+            goto fini;
         }
     } else {
-        return SG_LIB_CAT_INVALID_OP;
+        ret = SG_LIB_CAT_INVALID_OP;
+        goto fini;
     }
 
     res = scsi_operating_parameter(&oxcf, 1);
@@ -1753,22 +1803,25 @@ main(int argc, char * argv[])
             pr2serr("Unit attention (%s), continuing\n",
                     rec_copy_op_params_str);
             res = scsi_operating_parameter(&oxcf, 1);
-        } else {
-            if (-res == SG_LIB_CAT_INVALID_OP) {
-                pr2serr("%s command not supported on %s\n",
-                        rec_copy_op_params_str, oxcf.fname);
-                return EINVAL;
-            } else if (-res == SG_LIB_CAT_NOT_READY)
-                pr2serr("%s failed on %s - not ready\n",
-                        rec_copy_op_params_str, oxcf.fname);
-            else {
-                pr2serr("Unable to %s on %s\n", rec_copy_op_params_str,
-                        oxcf.fname);
-                return -res;
-            }
         }
-    } else if (res == 0)
-        return SG_LIB_CAT_INVALID_OP;
+        if (-res == SG_LIB_CAT_INVALID_OP) {
+            pr2serr("%s command not supported on %s\n",
+                    rec_copy_op_params_str, oxcf.fname);
+            ret = sg_convert_errno(EINVAL);
+            goto fini;
+        } else if (-res == SG_LIB_CAT_NOT_READY)
+            pr2serr("%s failed on %s - not ready\n",
+                    rec_copy_op_params_str, oxcf.fname);
+        else {
+            pr2serr("Unable to %s on %s\n", rec_copy_op_params_str,
+                    oxcf.fname);
+            ret = -res;
+            goto fini;
+        }
+    } else if (res == 0) {
+        ret = SG_LIB_CAT_INVALID_OP;
+        goto fini;
+    }
 
     if (res & TD_VPD) {
         if (verbose)
@@ -1778,10 +1831,12 @@ main(int argc, char * argv[])
                                  sizeof(dst_desc), oxcf.sect_sz, oxcf.pad);
         if (dst_desc_len > (int)sizeof(dst_desc)) {
             pr2serr("destination descriptor too large (%d bytes)\n", res);
-            return SG_LIB_CAT_MALFORMED;
+            ret = SG_LIB_CAT_MALFORMED;
+            goto fini;
         }
     } else {
-        return SG_LIB_CAT_INVALID_OP;
+        ret = SG_LIB_CAT_INVALID_OP;
+        goto fini;
     }
 
     if (dd_count < 0) {
@@ -1865,6 +1920,14 @@ main(int argc, char * argv[])
     else
         pr2serr("sg_xcopy: %" PRId64 " blocks, %d command%s\n", in_full,
                 num_xcopy, ((num_xcopy > 1) ? "s" : ""));
+    ret = res;
 
-    return res;
+fini:
+    /* file handles not explicitly closed; let process cleanup do that */
+    if (0 == verbose) {
+        if (! sg_if_can2stderr("sg_xcopy failed: ", ret))
+            pr2serr("Some error occurred, try again with '-v' or '-vv' for "
+                    "more information\n");
+    }
+    return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
