@@ -1,19 +1,20 @@
 /* A utility program for copying files. Specialised for "files" that
  * represent devices that understand the SCSI command set.
  *
- * Copyright (C) 1999 - 2018 D. Gilbert and P. Allworth
+ * Copyright (C) 1999 - 2021 D. Gilbert and P. Allworth
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2, or (at your option)
  * any later version.
+ *
+ * SPDX-License-Identifier: GPL-2.0-or-later
 
    This program is a specialisation of the Unix "dd" command in which
-   either the input or the output file is a scsi generic device or a
-   raw device. The block size ('bs') is assumed to be 512 if not given.
-   This program complains if 'ibs' or 'obs' are given with a value
-   that differs from 'bs' (or the default 512).
-   If 'if' is not given or 'if=-' then stdin is assumed. If 'of' is
-   not given or 'of=-' then stdout assumed.
+   either the input or the output file is a scsi generic device. The block
+   size ('bs') is assumed to be 512 if not given. This program complains if
+   'ibs' or 'obs' are given with a value that differs from 'bs' (or the
+   default, 512). If 'if' is not given or 'if=-' then stdin is assumed.
+   If 'of' is not given or 'of=-' then stdout assumed.
 
    A non-standard argument "bpt" (blocks per transfer) is added to control
    the maximum number of blocks in each transfer. The default value is 128.
@@ -26,7 +27,7 @@
    then only the read side will be mmap-ed, while the write side will
    use normal IO.
 
-   This version is designed for the linux kernel 2.4, 2.6, 3 and 4 series.
+   This version is designed for the linux kernel 2.4, 2.6, 3, 4 and 5 series.
 */
 
 #define _XOPEN_SOURCE 600
@@ -52,9 +53,11 @@
 #include <sys/time.h>
 #include <sys/mman.h>
 #include <sys/sysmacros.h>
-#include <sys/types.h>  /* needed for lseek64() */
-#include <linux/major.h>
-#include <linux/fs.h>   /* <sys/mount.h> */
+#ifndef major
+#include <sys/types.h>
+#endif
+#include <linux/major.h>        /* for MEM_MAJOR, SCSI_GENERIC_MAJOR, etc */
+#include <linux/fs.h>           /* for BLKSSZGET and friends */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -66,7 +69,7 @@
 #include "sg_pr2serr.h"
 
 
-static const char * version_str = "1.51 20171209";
+static const char * version_str = "1.17 20211024";
 
 #define DEF_BLOCK_SIZE 512
 #define DEF_BLOCKS_PER_TRANSFER 128
@@ -76,7 +79,6 @@ static const char * version_str = "1.51 20171209";
 
 #define ME "sgm_dd: "
 
-/* #define SG_DEBUG */
 
 #ifndef SG_FLAG_MMAP_IO
 #define SG_FLAG_MMAP_IO 4
@@ -89,7 +91,7 @@ static const char * version_str = "1.51 20171209";
 #define DEF_TIMEOUT 60000       /* 60,000 millisecs == 60 seconds */
 
 #ifndef RAW_MAJOR
-#define RAW_MAJOR 255   /*unlikey value */
+#define RAW_MAJOR 255   /*unlikely value */
 #endif
 
 #define FT_OTHER 1              /* filetype other than one of following */
@@ -113,11 +115,14 @@ static int in_partial = 0;
 static int64_t out_full = 0;
 static int out_partial = 0;
 static int verbose = 0;
+static int dry_run = 0;
+static int progress = 0;        /* accept --progress or -p, does nothing */
 
 static bool do_time = false;
 static bool start_tm_valid = false;
 static struct timeval start_tm;
 static int blk_sz = 0;
+static uint32_t glob_pack_id = 0;       /* pre-increment */
 
 static const char * proc_allow_dio = "/proc/scsi/sg/allow_dio";
 
@@ -240,19 +245,19 @@ dd_filetype_str(int ft, char * buff)
     int off = 0;
 
     if (FT_DEV_NULL & ft)
-        off += snprintf(buff + off, 32, "null device ");
+        off += sg_scnpr(buff + off, 32, "null device ");
     if (FT_SG & ft)
-        off += snprintf(buff + off, 32, "SCSI generic (sg) device ");
+        off += sg_scnpr(buff + off, 32, "SCSI generic (sg) device ");
     if (FT_BLOCK & ft)
-        off += snprintf(buff + off, 32, "block device ");
+        off += sg_scnpr(buff + off, 32, "block device ");
     if (FT_ST & ft)
-        off += snprintf(buff + off, 32, "SCSI tape device ");
+        off += sg_scnpr(buff + off, 32, "SCSI tape device ");
     if (FT_RAW & ft)
-        off += snprintf(buff + off, 32, "raw device ");
+        off += sg_scnpr(buff + off, 32, "raw device ");
     if (FT_OTHER & ft)
-        off += snprintf(buff + off, 32, "other (perhaps ordinary file) ");
+        off += sg_scnpr(buff + off, 32, "other (perhaps ordinary file) ");
     if (FT_ERROR & ft)
-        snprintf(buff + off, 32, "unable to 'stat' file ");
+        sg_scnpr(buff + off, 32, "unable to 'stat' file ");
     return buff;
 }
 
@@ -266,10 +271,12 @@ usage()
             "               [--help] [--version]\n\n");
     pr2serr("               [bpt=BPT] [cdbsz=6|10|12|16] [dio=0|1] "
             "[fua=0|1|2|3]\n"
-            "               [sync=0|1] [time=0|1] [verbose=VERB]\n\n"
+            "               [sync=0|1] [time=0|1] [verbose=VERB] "
+            "[--dry-run] [--verbose]\n\n"
             "  where:\n"
             "    bpt         is blocks_per_transfer (default is 128)\n"
-            "    bs          must be device block size (default 512)\n"
+            "    bs          must be device logical block size (default "
+            "512)\n"
             "    cdbsz       size of SCSI READ or WRITE cdb (default is 10)\n"
             "    count       number of blocks to copy (def: device size)\n"
             "    dio         0->indirect IO on write, 1->direct IO on write\n"
@@ -295,8 +302,10 @@ usage()
             "throughput\n"
             "    verbose     0->quiet(def), 1->some noise, 2->more noise, "
             "etc\n"
-            "    --help      print usage message then exit\n"
-            "    --version   print version information then exit\n\n"
+            "    --dry-run|-d    prepare but bypass copy/read\n"
+            "    --help|-h       print usage message then exit\n"
+            "    --verbose|-v    increase verbosity\n"
+            "    --version|-V    print version information then exit\n\n"
             "Copy from IFILE to OFILE, similar to dd command\n"
             "specialized for SCSI devices for which mmap-ed IO attempted\n");
 }
@@ -330,7 +339,7 @@ scsi_read_capacity(int sg_fd, int64_t * num_sect, int * sect_sz)
         *num_sect = (int64_t)ui + 1;
         *sect_sz = sg_get_unaligned_be32(rcBuff + 4);
     }
-    if (verbose)
+    if (verb)
         pr2serr("      number of blocks=%" PRId64 " [0x%" PRIx64 "], block "
                 "size=%d\n", *num_sect, *num_sect, *sect_sz);
     return 0;
@@ -355,9 +364,10 @@ read_blkdev_capacity(int sg_fd, int64_t * num_sect, int * sect_sz)
             return -1;
         }
         *num_sect = ((int64_t)ull / (int64_t)*sect_sz);
-        if (verbose)
+        if (verbose > 1)
             pr2serr("      [bgs64] number of blocks=%" PRId64 " [0x%" PRIx64
-                    "], block size=%d\n", *num_sect, *num_sect, *sect_sz);
+                    "], logical block size=%d\n", *num_sect, *num_sect,
+                    *sect_sz);
  #else
         unsigned long ul;
 
@@ -366,9 +376,10 @@ read_blkdev_capacity(int sg_fd, int64_t * num_sect, int * sect_sz)
             return -1;
         }
         *num_sect = (int64_t)ul;
-        if (verbose)
+        if (verbose > 1)
             pr2serr("      [bgs] number of blocks=%" PRId64 " [0x%" PRIx64
-                    "], block size=%d\n", *num_sect, *num_sect, *sect_sz);
+                    "], logical block size=%d\n", *num_sect, *num_sect,
+                    *sect_sz);
  #endif
     }
     return 0;
@@ -457,7 +468,8 @@ static int
 sg_read(int sg_fd, uint8_t * buff, int blocks, int64_t from_block,
         int bs, int cdbsz, bool fua, bool dpo, bool do_mmap)
 {
-    int k, res;
+    bool print_cdb_after = false;
+    int res;
     uint8_t rdCmd[MAX_SCSI_CDBSZ];
     uint8_t senseBuff[SENSE_BUFF_LEN];
     struct sg_io_hdr io_hdr;
@@ -479,19 +491,19 @@ sg_read(int sg_fd, uint8_t * buff, int blocks, int64_t from_block,
     io_hdr.mx_sb_len = SENSE_BUFF_LEN;
     io_hdr.sbp = senseBuff;
     io_hdr.timeout = DEF_TIMEOUT;
-    io_hdr.pack_id = (int)from_block;
+    io_hdr.pack_id = (int)++glob_pack_id;
     if (do_mmap)
         io_hdr.flags |= SG_FLAG_MMAP_IO;
     if (verbose > 2) {
-        pr2serr("    read cdb: ");
-        for (k = 0; k < cdbsz; ++k)
-            pr2serr("%02x ", rdCmd[k]);
-        pr2serr("\n");
+        char b[128];
+
+        pr2serr("    Read cdb: %s\n",
+                sg_get_command_str(rdCmd, cdbsz, false, sizeof(b), b));
     }
 
 #if 1
     while (((res = ioctl(sg_fd, SG_IO, &io_hdr)) < 0) &&
-           ((EINTR == errno) || (EAGAIN == errno)))
+           ((EINTR == errno) || (EAGAIN == errno) || (EBUSY == errno)))
         sleep(1);
     if (res < 0) {
         perror(ME "SG_IO error (sg_read)");
@@ -499,7 +511,7 @@ sg_read(int sg_fd, uint8_t * buff, int blocks, int64_t from_block,
     }
 #else
     while (((res = write(sg_fd, &io_hdr, sizeof(io_hdr))) < 0) &&
-           ((EINTR == errno) || (EAGAIN == errno)))
+           ((EINTR == errno) || (EAGAIN == errno) || (EBUSY == errno)))
         ;
     if (res < 0) {
         if (ENOMEM == errno)
@@ -509,7 +521,7 @@ sg_read(int sg_fd, uint8_t * buff, int blocks, int64_t from_block,
     }
 
     while (((res = read(sg_fd, &io_hdr, sizeof(io_hdr))) < 0) &&
-           ((EINTR == errno) || (EAGAIN == errno)))
+           ((EINTR == errno) || (EAGAIN == errno) || (EBUSY == errno)))
         ;
     if (res < 0) {
         perror("reading (rd) on sg device, error");
@@ -528,15 +540,20 @@ sg_read(int sg_fd, uint8_t * buff, int blocks, int64_t from_block,
     case SG_LIB_CAT_NOT_READY:
     case SG_LIB_CAT_MEDIUM_HARD:
         return res;
+    case SG_LIB_CAT_ILLEGAL_REQ:
+        if (verbose)
+            print_cdb_after = true;
+        /* FALL THROUGH */
     case SG_LIB_CAT_ABORTED_COMMAND:
     case SG_LIB_CAT_UNIT_ATTENTION:
-    case SG_LIB_CAT_ILLEGAL_REQ:
     default:
         sg_chk_n_print3("reading", &io_hdr, verbose > 1);
+        if (print_cdb_after)
+            sg_print_command_len(rdCmd, cdbsz);
         return res;
     }
     sum_of_resids += io_hdr.resid;
-#ifdef SG_DEBUG
+#ifdef DEBUG
     pr2serr("duration=%u ms\n", io_hdr.duration);
 #endif
     return 0;
@@ -548,7 +565,8 @@ static int
 sg_write(int sg_fd, uint8_t * buff, int blocks, int64_t to_block,
          int bs, int cdbsz, bool fua, bool dpo, bool do_mmap, bool * diop)
 {
-    int k, res;
+    bool print_cdb_after = false;
+    int res;
     uint8_t wrCmd[MAX_SCSI_CDBSZ];
     uint8_t senseBuff[SENSE_BUFF_LEN];
     struct sg_io_hdr io_hdr;
@@ -570,21 +588,21 @@ sg_write(int sg_fd, uint8_t * buff, int blocks, int64_t to_block,
     io_hdr.mx_sb_len = SENSE_BUFF_LEN;
     io_hdr.sbp = senseBuff;
     io_hdr.timeout = DEF_TIMEOUT;
-    io_hdr.pack_id = (int)to_block;
+    io_hdr.pack_id = (int)++glob_pack_id;
     if (do_mmap)
         io_hdr.flags |= SG_FLAG_MMAP_IO;
     else if (diop && *diop)
         io_hdr.flags |= SG_FLAG_DIRECT_IO;
     if (verbose > 2) {
-        pr2serr("    write cdb: ");
-        for (k = 0; k < cdbsz; ++k)
-            pr2serr("%02x ", wrCmd[k]);
-        pr2serr("\n");
+        char b[128];
+
+        pr2serr("    Write cdb: %s\n",
+                sg_get_command_str(wrCmd, cdbsz, false, sizeof(b), b));
     }
 
 #if 1
     while (((res = ioctl(sg_fd, SG_IO, &io_hdr)) < 0) &&
-           ((EINTR == errno) || (EAGAIN == errno)))
+           ((EINTR == errno) || (EAGAIN == errno) || (EBUSY == errno)))
         sleep(1);
     if (res < 0) {
         perror(ME "SG_IO error (sg_write)");
@@ -592,7 +610,7 @@ sg_write(int sg_fd, uint8_t * buff, int blocks, int64_t to_block,
     }
 #else
     while (((res = write(sg_fd, &io_hdr, sizeof(io_hdr))) < 0) &&
-           ((EINTR == errno) || (EAGAIN == errno)))
+           ((EINTR == errno) || (EAGAIN == errno) || (EBUSY == errno)))
         ;
     if (res < 0) {
         if (ENOMEM == errno)
@@ -602,7 +620,7 @@ sg_write(int sg_fd, uint8_t * buff, int blocks, int64_t to_block,
     }
 
     while (((res = read(sg_fd, &io_hdr, sizeof(io_hdr))) < 0) &&
-           ((EINTR == errno) || (EAGAIN == errno)))
+           ((EINTR == errno) || (EAGAIN == errno) || (EBUSY == errno)))
         ;
     if (res < 0) {
         perror("writing (rd) on sg device, error");
@@ -621,11 +639,16 @@ sg_write(int sg_fd, uint8_t * buff, int blocks, int64_t to_block,
     case SG_LIB_CAT_NOT_READY:
     case SG_LIB_CAT_MEDIUM_HARD:
         return res;
+    case SG_LIB_CAT_ILLEGAL_REQ:
+        if (verbose)
+            print_cdb_after = true;
+        /* FALL THROUGH */
     case SG_LIB_CAT_ABORTED_COMMAND:
     case SG_LIB_CAT_UNIT_ATTENTION:
-    case SG_LIB_CAT_ILLEGAL_REQ:
     default:
         sg_chk_n_print3("writing", &io_hdr, verbose > 1);
+        if (print_cdb_after)
+            sg_print_command_len(wrCmd, cdbsz);
         return res;
     }
     if (diop && *diop &&
@@ -677,11 +700,24 @@ process_flags(const char * arg, struct flags_t * fp)
     return 0;
 }
 
+/* Returns the number of times 'ch' is found in string 's' given the
+ * string's length. */
+static int
+num_chs_in_str(const char * s, int slen, int ch)
+{
+    int res = 0;
+
+    while (--slen >= 0) {
+        if (ch == s[slen])
+            ++res;
+    }
+    return res;
+}
+
 
 #define STR_SZ 1024
 #define INOUTF_SZ 512
-#define EBUFF_SZ 512
-
+#define EBUFF_SZ 768
 
 int
 main(int argc, char * argv[])
@@ -690,8 +726,9 @@ main(int argc, char * argv[])
     bool cdbsz_given = false;
     bool do_coe = false;     /* dummy, just accept + ignore */
     bool do_sync = false;
-    int res, k, t, infd, outfd, blocks, n, flags;
-    int blocks_per;
+    bool verbose_given = false;
+    bool version_given = false;
+    int res, k, t, infd, outfd, blocks, n, flags, blocks_per, err, keylen;
     int bpt = DEF_BLOCKS_PER_TRANSFER;
     int ibs = 0;
     int in_res_sz = 0;
@@ -735,13 +772,14 @@ main(int argc, char * argv[])
 
     for (k = 1; k < argc; k++) {
         if (argv[k])
-            strncpy(str, argv[k], STR_SZ);
+            snprintf(str, STR_SZ, "%s", argv[k]);
         else
             continue;
         for (key = str, buf = key; *buf && *buf != '=';)
             buf++;
         if (*buf)
             *buf++ = '\0';
+        keylen = strlen(key);
         if (0 == strcmp(key,"bpt")) {
             bpt = sg_get_num(buf);
             if (-1 == bpt) {
@@ -787,9 +825,11 @@ main(int argc, char * argv[])
         } else if (strcmp(key,"if") == 0) {
             if ('\0' != inf[0]) {
                 pr2serr("Second 'if=' argument??\n");
-                return SG_LIB_SYNTAX_ERROR;
-            } else
-                strncpy(inf, buf, INOUTF_SZ);
+                return SG_LIB_CONTRADICT;
+            } else {
+                memcpy(inf, buf, INOUTF_SZ);
+                inf[INOUTF_SZ - 1] = '\0';
+            }
         } else if (0 == strcmp(key, "iflag")) {
             if (process_flags(buf, &in_flags)) {
                 pr2serr(ME "bad argument to 'iflag'\n");
@@ -798,9 +838,11 @@ main(int argc, char * argv[])
         } else if (strcmp(key,"of") == 0) {
             if ('\0' != outf[0]) {
                 pr2serr("Second 'of=' argument??\n");
-                return SG_LIB_SYNTAX_ERROR;
-            } else
-                strncpy(outf, buf, INOUTF_SZ);
+                return SG_LIB_CONTRADICT;
+            } else {
+                memcpy(outf, buf, INOUTF_SZ);
+                outf[INOUTF_SZ - 1] = '\0';
+            }
         } else if (0 == strcmp(key, "oflag")) {
             if (process_flags(buf, &out_flags)) {
                 pr2serr(ME "bad argument to 'oflag'\n");
@@ -830,37 +872,88 @@ main(int argc, char * argv[])
             do_time = sg_get_num(buf);
         else if (0 == strncmp(key, "verb", 4))
             verbose = sg_get_num(buf);
-        else if ((0 == strncmp(key, "--help", 7)) ||
-                 (0 == strcmp(key, "-h")) || (0 == strcmp(key, "-?"))) {
+        else if ((keylen > 1) && ('-' == key[0]) && ('-' != key[1])) {
+            res = 0;
+            n = num_chs_in_str(key + 1, keylen - 1, 'd');
+            dry_run += n;
+            res += n;
+            n = num_chs_in_str(key + 1, keylen - 1, 'h');
+            if (n > 0) {
+                usage();
+                return 0;
+            }
+            n = num_chs_in_str(key + 1, keylen - 1, 'p');
+            progress += n;
+            res += n;
+            n = num_chs_in_str(key + 1, keylen - 1, 'v');
+            verbose += n;
+            res += n;
+            n = num_chs_in_str(key + 1, keylen - 1, 'V');
+            if (n > 0)
+                version_given = true;
+            res += n;
+            if (res < (keylen - 1)) {
+                pr2serr("Unrecognised short option in '%s', try '--help'\n",
+                        key);
+                return SG_LIB_SYNTAX_ERROR;
+            }
+        } else if ((0 == strncmp(key, "--dry-run", 9)) ||
+                 (0 == strncmp(key, "--dry_run", 9)))
+            ++dry_run;
+        else if ((0 == strncmp(key, "--help", 6)) ||
+                 (0 == strcmp(key, "-?"))) {
             usage();
             return 0;
-        } else if ((0 == strncmp(key, "--vers", 6)) ||
-                   (0 == strcmp(key, "-V"))) {
-            pr2serr(ME ": %s\n", version_str);
-            return 0;
-        }
+        } else if (0 == strncmp(key, "--prog", 6))
+            ++progress;
+        else if (0 == strncmp(key, "--verb", 6))
+            ++verbose;
+        else if (0 == strncmp(key, "--vers", 6))
+            version_given = true;
         else {
             pr2serr("Unrecognized option '%s'\n", key);
             pr2serr("For more information use '--help'\n");
             return SG_LIB_SYNTAX_ERROR;
         }
     }
+#ifdef DEBUG
+    pr2serr("In DEBUG mode, ");
+    if (verbose_given && version_given) {
+        pr2serr("but override: '-vV' given, zero verbose and continue\n");
+        verbose_given = false;
+        version_given = false;
+        verbose = 0;
+    } else if (! verbose_given) {
+        pr2serr("set '-vv'\n");
+        verbose = 2;
+    } else
+        pr2serr("keep verbose=%d\n", verbose);
+#else
+    if (verbose_given && version_given)
+        pr2serr("Not in DEBUG mode, so '-vV' has no special action\n");
+#endif
+    if (version_given) {
+        pr2serr(ME ": %s\n", version_str);
+        return 0;
+    }
+
     if (blk_sz <= 0) {
         blk_sz = DEF_BLOCK_SIZE;
-        pr2serr("Assume default 'bs' (block size) of %d bytes\n", blk_sz);
+        pr2serr("Assume default 'bs' ((logical) block size) of %d bytes\n",
+                blk_sz);
     }
     if ((ibs && (ibs != blk_sz)) || (obs && (obs != blk_sz))) {
         pr2serr("If 'ibs' or 'obs' given must be same as 'bs'\n");
         usage();
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
     if ((skip < 0) || (seek < 0)) {
         pr2serr("skip and seek cannot be negative\n");
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
     if (out_flags.append && (seek > 0)) {
         pr2serr("Can't use both append and seek switches\n");
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
     if (bpt < 1) {
         pr2serr("bpt must be greater than 0\n");
@@ -872,7 +965,7 @@ main(int argc, char * argv[])
     if ((blk_sz >= 2048) && (! bpt_given))
         bpt = DEF_BLOCKS_PER_2048TRANSFER;
 
-#ifdef SG_DEBUG
+#ifdef DEBUG
     pr2serr(ME "if=%s skip=%" PRId64 " of=%s seek=%" PRId64 " count=%" PRId64
             "\n", inf, skip, outf, seek, dd_count);
 #endif
@@ -885,7 +978,7 @@ main(int argc, char * argv[])
     outfd = STDOUT_FILENO;
     if (inf[0] && ('-' != inf[0])) {
         in_type = dd_filetype(inf);
-        if (verbose)
+        if (verbose > 1)
             pr2serr(" >> Input file type: %s\n",
                     dd_filetype_str(in_type, ebuff));
 
@@ -904,10 +997,11 @@ main(int argc, char * argv[])
             if (in_flags.dsync)
                 flags |= O_SYNC;
             if ((infd = open(inf, flags)) < 0) {
+                err = errno;
                 snprintf(ebuff, EBUFF_SZ,
                          ME "could not open %s for sg reading", inf);
                 perror(ebuff);
-                return SG_LIB_FILE_ERROR;
+                return sg_convert_errno(err);
             }
             res = ioctl(infd, SG_GET_VERSION_NUM, &t);
             if ((res < 0) || (t < 30122)) {
@@ -918,24 +1012,27 @@ main(int argc, char * argv[])
             if (0 != (in_res_sz % psz)) /* round up to next page */
                 in_res_sz = ((in_res_sz / psz) + 1) * psz;
             if (ioctl(infd, SG_GET_RESERVED_SIZE, &t) < 0) {
+                err = errno;
                 perror(ME "SG_GET_RESERVED_SIZE error");
-                return SG_LIB_FILE_ERROR;
+                return sg_convert_errno(err);
             }
             if (t < MIN_RESERVED_SIZE)
                 t = MIN_RESERVED_SIZE;
             if (in_res_sz > t) {
                 if (ioctl(infd, SG_SET_RESERVED_SIZE, &in_res_sz) < 0) {
+                    err = errno;
                     perror(ME "SG_SET_RESERVED_SIZE error");
-                    return SG_LIB_FILE_ERROR;
+                    return sg_convert_errno(err);
                 }
             }
             wrkMmap = (uint8_t *)mmap(NULL, in_res_sz,
                                  PROT_READ | PROT_WRITE, MAP_SHARED, infd, 0);
             if (MAP_FAILED == wrkMmap) {
+                err = errno;
                 snprintf(ebuff, EBUFF_SZ,
                          ME "error using mmap() on file: %s", inf);
                 perror(ebuff);
-                return SG_LIB_FILE_ERROR;
+                return sg_convert_errno(err);
             }
         } else {
             flags = O_RDONLY;
@@ -946,22 +1043,24 @@ main(int argc, char * argv[])
             if (in_flags.dsync)
                 flags |= O_SYNC;
             if ((infd = open(inf, flags)) < 0) {
+                err = errno;
                 snprintf(ebuff, EBUFF_SZ,
                          ME "could not open %s for reading", inf);
                 perror(ebuff);
-                return SG_LIB_FILE_ERROR;
+                return sg_convert_errno(err);
             }
             else if (skip > 0) {
                 off64_t offset = skip;
 
                 offset *= blk_sz;       /* could exceed 32 bits here! */
                 if (lseek64(infd, offset, SEEK_SET) < 0) {
+                    err = errno;
                     snprintf(ebuff, EBUFF_SZ, ME "couldn't skip to "
                              "required position on %s", inf);
                     perror(ebuff);
-                    return SG_LIB_FILE_ERROR;
+                    return sg_convert_errno(err);
                 }
-                if (verbose)
+                if (verbose > 1)
                     pr2serr("  >> skip: lseek64 SEEK_SET, byte offset=0x%"
                             PRIx64 "\n", (uint64_t)offset);
             }
@@ -970,7 +1069,7 @@ main(int argc, char * argv[])
 
     if (outf[0] && ('-' != outf[0])) {
         out_type = dd_filetype(outf);
-        if (verbose)
+        if (verbose > 1)
             pr2serr(" >> Output file type: %s\n",
                     dd_filetype_str(out_type, ebuff));
 
@@ -987,10 +1086,11 @@ main(int argc, char * argv[])
             if (out_flags.dsync)
                 flags |= O_SYNC;
             if ((outfd = open(outf, flags)) < 0) {
+                err = errno;
                 snprintf(ebuff, EBUFF_SZ, ME "could not open %s for "
                          "sg writing", outf);
                 perror(ebuff);
-                return SG_LIB_FILE_ERROR;
+                return sg_convert_errno(err);
             }
             res = ioctl(outfd, SG_GET_VERSION_NUM, &t);
             if ((res < 0) || (t < 30122)) {
@@ -998,26 +1098,29 @@ main(int argc, char * argv[])
                 return SG_LIB_FILE_ERROR;
             }
             if (ioctl(outfd, SG_GET_RESERVED_SIZE, &t) < 0) {
+                err = errno;
                 perror(ME "SG_GET_RESERVED_SIZE error");
-                return SG_LIB_FILE_ERROR;
+                return sg_convert_errno(err);
             }
            if (t < MIN_RESERVED_SIZE)
                 t = MIN_RESERVED_SIZE;
             out_res_sz = blk_sz * bpt;
             if (out_res_sz > t) {
                 if (ioctl(outfd, SG_SET_RESERVED_SIZE, &out_res_sz) < 0) {
+                    err = errno;
                     perror(ME "SG_SET_RESERVED_SIZE error");
-                    return SG_LIB_FILE_ERROR;
+                    return sg_convert_errno(err);
                 }
             }
             if (NULL == wrkMmap) {
                 wrkMmap = (uint8_t *)mmap(NULL, out_res_sz,
                                 PROT_READ | PROT_WRITE, MAP_SHARED, outfd, 0);
                 if (MAP_FAILED == wrkMmap) {
+                    err = errno;
                     snprintf(ebuff, EBUFF_SZ,
                              ME "error using mmap() on file: %s", outf);
                     perror(ebuff);
-                    return SG_LIB_FILE_ERROR;
+                    return sg_convert_errno(err);
                 }
             }
         }
@@ -1035,18 +1138,20 @@ main(int argc, char * argv[])
                 if (out_flags.append)
                     flags |= O_APPEND;
                 if ((outfd = open(outf, flags, 0666)) < 0) {
+                    err = errno;
                     snprintf(ebuff, EBUFF_SZ,
                              ME "could not open %s for writing", outf);
                     perror(ebuff);
-                    return SG_LIB_FILE_ERROR;
+                    return sg_convert_errno(err);
                 }
             }
             else {
                 if ((outfd = open(outf, O_WRONLY)) < 0) {
+                    err = errno;
                     snprintf(ebuff, EBUFF_SZ, ME "could not open %s "
                              "for raw writing", outf);
                     perror(ebuff);
-                    return SG_LIB_FILE_ERROR;
+                    return sg_convert_errno(err);
                 }
             }
             if (seek > 0) {
@@ -1054,12 +1159,13 @@ main(int argc, char * argv[])
 
                 offset *= blk_sz;       /* could exceed 32 bits here! */
                 if (lseek64(outfd, offset, SEEK_SET) < 0) {
+                    err = errno;
                     snprintf(ebuff, EBUFF_SZ, ME "couldn't seek to "
                              "required position on %s", outf);
                     perror(ebuff);
-                    return SG_LIB_FILE_ERROR;
+                    return sg_convert_errno(err);
                 }
-                if (verbose)
+                if (verbose > 1)
                     pr2serr("   >> seek: lseek64 SEEK_SET, byte offset=0x%"
                             PRIx64 "\n", (uint64_t)offset);
             }
@@ -1069,7 +1175,7 @@ main(int argc, char * argv[])
         pr2serr("Won't default both IFILE to stdin _and_ OFILE to as "
                 "stdout\n");
         pr2serr("For more information use '--help'\n");
-        return SG_LIB_SYNTAX_ERROR;
+        return SG_LIB_CONTRADICT;
     }
     if (dd_count < 0) {
         in_num_sect = -1;
@@ -1093,8 +1199,8 @@ main(int argc, char * argv[])
                 in_num_sect = -1;
             }
             if (blk_sz != in_sect_sz) {
-                pr2serr("block size on %s confusion; bs=%d, from device=%d\n",
-                        inf, blk_sz, in_sect_sz);
+                pr2serr("logical block size on %s confusion; bs=%d, from "
+                        "device=%d\n", inf, blk_sz, in_sect_sz);
                 in_num_sect = -1;
             }
         }
@@ -1123,14 +1229,14 @@ main(int argc, char * argv[])
                 out_num_sect = -1;
             }
             if (blk_sz != out_sect_sz) {
-                pr2serr("block size on %s confusion: bs=%d, from device=%d\n",
-                        outf, blk_sz, out_sect_sz);
+                pr2serr("logical block size on %s confusion: bs=%d, from "
+                        "device=%d\n", outf, blk_sz, out_sect_sz);
                 out_num_sect = -1;
             }
         }
         if (out_num_sect > seek)
             out_num_sect -= seek;
-#ifdef SG_DEBUG
+#ifdef DEBUG
         pr2serr("Start of loop, count=%" PRId64 ", in_num_sect=%" PRId64 ", "
                 "out_num_sect=%" PRId64 "\n", dd_count, in_num_sect,
                 out_num_sect);
@@ -1187,31 +1293,22 @@ main(int argc, char * argv[])
     if (wrkMmap) {
         wrkPos = wrkMmap;
     } else {
-        if ((FT_RAW == in_type) || (FT_RAW == out_type)) {
-            wrkBuff = (uint8_t *)malloc(blk_sz * bpt + psz);
-            if (NULL == wrkBuff) {
-                pr2serr("Not enough user memory for raw\n");
-                return SG_LIB_FILE_ERROR;
-            }
-            /* perhaps use posix_memalign() instead */
-            wrkPos = (uint8_t *)(((sg_uintptr_t)wrkBuff + psz - 1) &
-                                       (~(psz - 1)));
-        }
-        else {
-            wrkBuff = (uint8_t *)malloc(blk_sz * bpt);
-            if (NULL == wrkBuff) {
-                pr2serr("Not enough user memory\n");
-                return SG_LIB_FILE_ERROR;
-            }
-            wrkPos = wrkBuff;
+        wrkPos = (uint8_t *)sg_memalign(blk_sz * bpt, 0, &wrkBuff,
+                                        verbose > 3);
+        if (NULL == wrkPos) {
+            pr2serr("Not enough user memory\n");
+            return sg_convert_errno(ENOMEM);
         }
     }
 
     blocks_per = bpt;
-#ifdef SG_DEBUG
+#ifdef DEBUG
     pr2serr("Start of loop, count=%" PRId64 ", blocks_per=%d\n", dd_count,
             blocks_per);
 #endif
+    if (dry_run > 0)
+        goto fini;
+
     if (do_time) {
         start_tm.tv_sec = 0;
         start_tm.tv_usec = 0;
@@ -1247,7 +1344,8 @@ main(int argc, char * argv[])
         }
         else {
             while (((res = read(infd, wrkPos, blocks * blk_sz)) < 0) &&
-                   ((EINTR == errno) || (EAGAIN == errno)))
+                   ((EINTR == errno) || (EAGAIN == errno) ||
+                    (EBUSY == errno)))
                 ;
             if (verbose > 2)
                 pr2serr("read(unix): count=%d, res=%d\n", blocks * blk_sz,
@@ -1301,7 +1399,8 @@ main(int argc, char * argv[])
             out_full += blocks; /* act as if written out without error */
         else {
             while (((res = write(outfd, wrkPos, blocks * blk_sz)) < 0) &&
-                   ((EINTR == errno) || (EAGAIN == errno)))
+                   ((EINTR == errno) || (EAGAIN == errno) ||
+                    (EBUSY == errno)))
                 ;
             if (verbose > 2)
                 pr2serr("write(unix): count=%d, res=%d\n", blocks * blk_sz,
@@ -1346,13 +1445,16 @@ main(int argc, char * argv[])
         }
     }
 
+fini:
     if (wrkBuff)
         free(wrkBuff);
-    if (STDIN_FILENO != infd)
+    if ((STDIN_FILENO != infd) && (infd >= 0))
         close(infd);
-    if ((STDOUT_FILENO != outfd) && (FT_DEV_NULL != out_type))
-        close(outfd);
-    if (0 != dd_count) {
+    if ((STDOUT_FILENO != outfd) && (FT_DEV_NULL != out_type)) {
+        if (outfd >= 0)
+            close(outfd);
+    }
+    if ((0 != dd_count) && (0 == dry_run)) {
         pr2serr("Some error occurred,");
         if (0 == ret)
             ret = SG_LIB_CAT_OTHER;
