@@ -1,8 +1,10 @@
 /*
- * Copyright (c) 2013-2018 Douglas Gilbert.
+ * Copyright (c) 2013-2020 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 #include <unistd.h>
@@ -19,7 +21,7 @@
 
 #include <time.h>
 
-#ifdef __GNUC__
+#if defined(__GNUC__) && ! defined(SG_LIB_FREEBSD)
 #include <byteswap.h>
 #endif
 
@@ -28,6 +30,7 @@
 #endif
 
 #include "sg_lib.h"
+#include "sg_pr2serr.h"
 
 /* Uncomment the next two undefs to force use of the generic (i.e. shifting)
  * unaligned functions (i.e. sg_get_* and sg_put_*). Use "-b 16|32|64
@@ -41,7 +44,7 @@
  * related to snprintf().
  */
 
-static const char * version_str = "1.10 20180311";
+static const char * version_str = "1.14 20200901";
 
 
 #define MAX_LINE_LEN 1024
@@ -159,63 +162,45 @@ usage()
             "[--printf]\n"
             "                  [--sense] [--unaligned] [--verbose] "
             "[--version]\n"
-#ifdef __GNUC__
-            "  where: --byteswap=B|-b B    B is 16, 32 or 64; tests "
-            "NUM byteswaps\n"
-            "                              compared to sg_unaligned "
+            "  where:\n"
+#if defined(__GNUC__) && ! defined(SG_LIB_FREEBSD)
+            "    --byteswap=B|-b B    B is 16, 32 or 64; tests NUM "
+            "byteswaps\n"
+            "                         compared to sg_unaligned "
             "equivalent\n"
-            "         --exit|-e          test exit status strings\n"
+            "    --exit|-e          test exit status strings\n"
 #else
-            "  where: --exit|-e          test exit status strings\n"
+            "    --exit|-e          test exit status strings\n"
 #endif
-            "         --help|-h          print out usage message\n"
-            "         --hex2|-H          test hex2* variants\n"
-            "         --leadin=STR|-l STR    every line output by --sense "
+            "    --help|-h          print out usage message\n"
+            "    --hex2|-H          test hex2* variants\n"
+            "    --leadin=STR|-l STR    every line output by --sense "
             "should\n"
-            "                                be prefixed by STR\n"
-            "         --num=NUM|-n NUM    number of iterations (def=1)\n"
-            "         --printf|-p        test library printf variants\n"
-            "         --sense|-s         test sense data handling\n"
-            "         --unaligned|-u     test unaligned data handling\n"
-            "         --verbose|-v       increase verbosity\n"
-            "         --version|-V       print version string and exit\n\n"
+            "                           be prefixed by STR\n"
+            "    --num=NUM|-n NUM    number of iterations (def=1)\n"
+            "    --printf|-p        test library printf variants\n"
+            "    --sense|-s         test sense data handling\n"
+            "    --unaligned|-u     test unaligned data handling\n"
+            "    --verbose|-v       increase verbosity\n"
+            "    --version|-V       print version string and exit\n\n"
             "Test various parts of sg_lib, see options. Sense data tests "
             "overlap\nsomewhat with examples/sg_sense_test .\n"
            );
 
 }
 
-/* Want safe, 'n += snprintf(b + n ...)' like function. If cp_max_len is 1
- * then assume cp is pointing to a null char and do nothing. Returns number
- * number of chars placed in cp excluding the trailing null char. So for
- * cp_max_len > 0 the return value is always < cp_max_len; for cp_max_len
- * <= 0 the return value is 0 (and no chars are written to cp). */
-static int
-my_snprintf(char * cp, int cp_max_len, const char * fmt, ...)
-{
-    va_list args;
-    int n;
-
-    if (cp_max_len < 2)
-        return 0;
-    va_start(args, fmt);
-    n = vsnprintf(cp, cp_max_len, fmt, args);
-    va_end(args);
-    return (n < cp_max_len) ? n : (cp_max_len - 1);
-}
-
-char *
+static char *
 get_exit_status_str(int exit_status, bool longer, int b_len, char * b)
 {
     int n;
 
-    n = my_snprintf(b, b_len, "  ES=%d: ", exit_status);
+    n = sg_scnpr(b, b_len, "  ES=%d: ", exit_status);
     if (n >= (b_len - 1))
         return b;
     if (sg_exit2str(exit_status, longer, b_len - n, b + n)) {
         n = (int)strlen(b);
         if (n < (b_len - 1))
-            my_snprintf(b + n, b_len - n, " [ok=true]");
+            sg_scnpr(b + n, b_len - n, " [ok=true]");
         return b;
     } else
         snprintf(b, b_len, "  No ES string for %d%s", exit_status,
@@ -225,7 +210,7 @@ get_exit_status_str(int exit_status, bool longer, int b_len, char * b)
 
 static uint8_t arr[64];
 
-#define OFF 7	/* in byteswap mode, can test different alignments (def: 8) */
+#define OFF 7   /* in byteswap mode, can test different alignments (def: 8) */
 
 int
 main(int argc, char * argv[])
@@ -315,6 +300,8 @@ main(int argc, char * argv[])
     }
 
     if (do_exit_status) {
+        ++did_something;
+
         printf("Test Exit Status strings (add -v for long version):\n");
         printf("  No error (es=0): %s\n",
                sg_get_category_sense_str(0, b_len, b, vb));
@@ -333,6 +320,7 @@ main(int argc, char * argv[])
         printf("%s\n", get_exit_status_str(8, (vb > 0), b_len, b));
         printf("%s\n", get_exit_status_str(25, (vb > 0), b_len, b));
         printf("%s\n", get_exit_status_str(33, (vb > 0), b_len, b));
+        printf("%s\n", get_exit_status_str(36, (vb > 0), b_len, b));
         printf("%s\n", get_exit_status_str(48, (vb > 0), b_len, b));
         printf("%s\n", get_exit_status_str(50, (vb > 0), b_len, b));
         printf("%s\n", get_exit_status_str(51, (vb > 0), b_len, b));
@@ -375,7 +363,7 @@ main(int argc, char * argv[])
         sg_print_sense(leadin, desc_sense_data5,
                        (int)sizeof(desc_sense_data5), vb);
         printf("\n");
-        printf("desc_sense UA subsidiary binfing changed test6\n");
+        printf("desc_sense UA subsidiary binding changed test6\n");
         sg_print_sense(leadin, desc_sense_data6,
                        (int)sizeof(desc_sense_data6), vb);
         printf("\n");
@@ -384,83 +372,83 @@ main(int argc, char * argv[])
 
     if (do_printf) {
         ++did_something;
-        printf("Testing my_snprintf():\n");
+        printf("Testing sg_scnpr():\n");
         b[0] = '\0';
         len = b_len;
-        n = my_snprintf(b, len, "%s", "test");
-        printf("my_snprintf(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
+        n = sg_scnpr(b, len, "%s", "test");
+        printf("sg_scnpr(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
                len, n, (uint32_t)strlen(b));
         if (strlen(b) > 0)
             printf("Resulting string: %s\n", b);
 
         b[0] = '\0';
         len = -1;
-        n = my_snprintf(b, len, "%s", "test");
-        printf("my_snprintf(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
+        n = sg_scnpr(b, len, "%s", "test");
+        printf("sg_scnpr(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
                len, n, (uint32_t)strlen(b));
         if (strlen(b) > 0)
             printf("Resulting string: %s\n", b);
 
         b[0] = '\0';
         len = 0;
-        n = my_snprintf(b, len, "%s", "test");
-        printf("my_snprintf(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
+        n = sg_scnpr(b, len, "%s", "test");
+        printf("sg_scnpr(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
                len, n, (uint32_t)strlen(b));
         if (strlen(b) > 0)
             printf("Resulting string: %s\n", b);
 
         b[0] = '\0';
         len = 1;
-        n = my_snprintf(b, len, "%s", "test");
-        printf("my_snprintf(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
+        n = sg_scnpr(b, len, "%s", "test");
+        printf("sg_scnpr(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
                len, n, (uint32_t)strlen(b));
         if (strlen(b) > 0)
             printf("Resulting string: %s\n", b);
 
         b[0] = '\0';
         len = 2;
-        n = my_snprintf(b, len, "%s", "test");
-        printf("my_snprintf(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
+        n = sg_scnpr(b, len, "%s", "test");
+        printf("sg_scnpr(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
                len, n, (uint32_t)strlen(b));
         if (strlen(b) > 0)
             printf("Resulting string: %s\n", b);
 
         b[0] = '\0';
         len = 3;
-        n = my_snprintf(b, len, "%s", "test");
-        printf("my_snprintf(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
+        n = sg_scnpr(b, len, "%s", "test");
+        printf("sg_scnpr(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
                len, n, (uint32_t)strlen(b));
         if (strlen(b) > 0)
             printf("Resulting string: %s\n", b);
 
         b[0] = '\0';
         len = 4;
-        n = my_snprintf(b, len, "%s", "test");
-        printf("my_snprintf(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
+        n = sg_scnpr(b, len, "%s", "test");
+        printf("sg_scnpr(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
                len, n, (uint32_t)strlen(b));
         if (strlen(b) > 0)
             printf("Resulting string: %s\n", b);
 
         b[0] = '\0';
         len = 5;
-        n = my_snprintf(b, len, "%s", "test");
-        printf("my_snprintf(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
+        n = sg_scnpr(b, len, "%s", "test");
+        printf("sg_scnpr(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
                len, n, (uint32_t)strlen(b));
         if (strlen(b) > 0)
             printf("Resulting string: %s\n", b);
 
         b[0] = '\0';
         len = 6;
-        n = my_snprintf(b, len, "%s", "test");
-        printf("my_snprintf(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
+        n = sg_scnpr(b, len, "%s", "test");
+        printf("sg_scnpr(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
                len, n, (uint32_t)strlen(b));
         if (strlen(b) > 0)
             printf("Resulting string: %s\n", b);
 
         b[0] = '\0';
         len = 7;
-        n = my_snprintf(b, len, "%s", "test");
-        printf("my_snprintf(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
+        n = sg_scnpr(b, len, "%s", "test");
+        printf("sg_scnpr(,%d,,\"test\") -> %d; strlen(b) -> %u\n",
                len, n, (uint32_t)strlen(b));
         if (strlen(b) > 0)
             printf("Resulting string: %s\n", b);
@@ -569,7 +557,7 @@ main(int argc, char * argv[])
         printf("  u64r[v=8 bytes]=0x%" PRIx64 "\n\n", u64r);
     }
 
-#ifdef __GNUC__
+#if defined(__GNUC__) && ! defined(SG_LIB_FREEBSD)
     if (byteswap_sz > 0) {
         uint32_t elapsed_msecs;
         uint16_t count16 = 0;
@@ -607,11 +595,11 @@ main(int argc, char * argv[])
         elapsed_msecs = (end_tm.tv_sec - start_tm.tv_sec) * 1000;
         elapsed_msecs += (end_tm.tv_nsec - start_tm.tv_nsec) / 1000000;
         if (16 == byteswap_sz)
-            printf("  count16=%u\n", count16);
+            printf("  last k=%d, last count16=%u\n", k, count16);
         else if (32 == byteswap_sz)
-            printf("  count32=%u\n", count32);
+            printf("  last k=%d, last count32=%u\n", k, count32);
         else
-            printf("  count64=%" PRIu64 "\n", count64);
+            printf("  last k=%d, last count64=%" PRIu64 "\n", k, count64);
         printf("Unaligned elapsed milliseconds: %u\n", elapsed_msecs);
         count16 = 0;
         count32 = 0;
@@ -652,11 +640,11 @@ main(int argc, char * argv[])
         elapsed_msecs = (end_tm.tv_sec - start_tm.tv_sec) * 1000;
         elapsed_msecs += (end_tm.tv_nsec - start_tm.tv_nsec) / 1000000;
         if (16 == byteswap_sz)
-            printf("  count16=%u\n", count16);
+            printf("  last k=%d, last count16=%u\n", k, count16);
         else if (32 == byteswap_sz)
-            printf("  count32=%u\n", count32);
+            printf("  last k=%d, last count32=%u\n", k, count32);
         else
-            printf("  count64=%" PRIu64 "\n", count64);
+            printf("  last k=%d, last count64=%" PRIu64 "\n", k, count64);
         printf("Byteswap/memcpy elapsed milliseconds: %u\n", elapsed_msecs);
         count16 = 0;
         count32 = 0;
